/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.internal.backend.impl;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;

import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.httpclient.DefaultHttpMethodRetryHandler;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.httpclient.params.HttpMethodParams;
import org.apache.xmlrpc.client.XmlRpcClient;
import org.apache.xmlrpc.client.XmlRpcClientConfigImpl;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.conf.VanishConfiguration.TimeUnit;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.backend.TooManySharesLostVanishException;
import edu.washington.cs.vanish.internal.backend.VanishBackendInterface;
import edu.washington.cs.vanish.internal.backend.VanishBackendException;
import edu.washington.cs.vanish.logging.LogStat;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.util.Constants;
import edu.washington.cs.vanish.util.RefreshableConfig;
import edu.washington.cs.vanish.util.RefreshableParams;

/**
 * OpenDHT-based implementation of the Vanish backend.
 * 
 * IMPORTANT CONCEPT:
 * 
 * The future for self-destructing data is to use a HYBRID approach that
 * combines the two extreme solutions that people have thought of in the
 * context of self-destructing data: centralized services (a la Ephemerizer)
 * and DHTs, the Web, or other solutions (a la original Vanish prototype,
 * v 0.1). Such hybrid approaches combine the best of the two worlds and
 * increase security. For details, read our paper at
 * http://vanish.cs.washington.edu/pubs/dhts.pdf. 
 * 
 * For this reason, we have implemented starting in Vanish Version 0.2 an
 * OpenDHT client that is an example of a more centralized solution than
 * Vuze. 
 * 
 *  
 * 
 * Implementation details:
 * TODO(roxana): For now, this is mostly a hack. Some small redesign is
 * needed to the Vanish internal interfaces to support the hybrid approach,
 * and in particular OpenDHT's flexible timeouts. 
 * TODO(roxana): Performance's BAD b/c of sequential accesses.
 * TODO(roxana): Blacklisting for slow OpenDHT servers.
 * 
 * @author roxana
 */

class DHTParams implements RefreshableParams {	
	final int  kDhtIndexLen;
	final int kOperationTimeout_s;
	final String kServerListURL;
	final long kRefreshServerListInterval_ms;
	
	public DHTParams(int dhtIndexLen, int operationTimeout_s,
			         String serverListURL,
			         long refreshServerListInterval_ms) {
		this.kDhtIndexLen   = dhtIndexLen;
		this.kOperationTimeout_s = operationTimeout_s;
		this.kServerListURL = serverListURL;
		this.kRefreshServerListInterval_ms = refreshServerListInterval_ms;
	}
}

public class OpenDHTVanishBackendImpl 
implements VanishBackendInterface, RefreshableConfig<DHTParams>, Constants {
	// TODO(roxana): This is a hack, till we port to new Vuze, which
	// supports > 8h timeouts. Then, we can make the timeout be a param
	// up to 3 days.
	private static final int OPENDHT_TIMEOUT_S = (int)(8.5 * 3600);
	
	// Vanish backend params, which are reloadable from the Configuration.
	private DHTParams params;

	// State:

	private final VanishLogger LOG;
	private final VanishConfiguration CONF;
	
	private final Object servers_list_lock = new Object();
	private Vector<String> servers_list;
	
	private OpenDHTServer current_server;
	
	private Random random = new Random();
	
	private Timer server_list_refresher = new Timer();

	public OpenDHTVanishBackendImpl(VanishConfiguration config,
			                        VanishLogger LOG)
	throws InvalidArgumentVanishException {
		this.LOG = LOG;
		this.CONF = loadConfig(config);
		
		// Load the parameters from the configuration:
		params = this.getRenewedParamsFromConfig(CONF);
		setParams(params);
		
		server_list_refresher.schedule(
				new RefreshServerListTask(), 0, 
				params.kRefreshServerListInterval_ms);
	}
	
	private VanishConfiguration loadConfig(VanishConfiguration config) {
		VanishConfiguration new_config = config;
		try {
			new_config = new VanishConfiguration(
					OpenDHTVanishBackendDefaults.OPENDHT_BACKEND_CONFIGURATION_FILE);
		} catch (ConfigurationException e) {  // Ignore errors.
		}
		return new_config;
	}
	
    // Initializable interface:

	/**
	 * Thread-safe.
	 */
	// @Override
	public void init() throws VanishBackendException {
		final LogStat log_stat = LOG.logStat("OpenDHTVanishBackend.init");
		
		// Wait till we have we'll all have to wait for the server list to
		// come.
		// TODO(roxana): VERY, VERY ugly now:
		while (! isInitialized()) {
			try { Thread.sleep(50); }
			catch (InterruptedException e) { }
		}
		
		// Choose a server for starters.
		chooseCurrentServer();
		
		log_stat.end();
	}

	/**
	 * Not thread-safe!
	 */
	// @Override
	public boolean isInitialized() {
		return servers_list != null;
	}

	// @Override
	public void stop() throws VanishException {
	}
	
	// VanishBackend interface:

	// @Override
	public byte[][] getShares(byte[][] locations, int needed_keys)
	throws VanishBackendException, TooManySharesLostVanishException {
		if (! isInitialized()) throw new VanishBackendException("uninit");
		final LogStat log_stat = LOG.logStat(
				"OpenDHTVanishBackend.getShares");

		int num_results_successful = 0;
		byte[][] results = new byte[locations.length][];
		for (int i = 0;
		     i < locations.length && num_results_successful < needed_keys;
		     ++i) {
			results[i] = getOneShare(locations[i]);
			if (results[i] != null) ++num_results_successful;
		}

		if (num_results_successful < needed_keys) {
			// Oops, lost too many keys...
			throw new TooManySharesLostVanishException("Got " +
					num_results_successful + "/" + needed_keys +
					" shares (opendht)");
		}
		log_stat.end();
		return results;
	}
		
	// @Override
	public void pushShares(byte[][] keys, byte[][] locations)
	throws VanishBackendException {
		if (! isInitialized()) throw new VanishBackendException("Not init");

		final LogStat log_stat = LOG.logStat(
				"OpenDHTVanishBackend.pushShares");

		assert (keys.length == locations.length);
		
		for (int i = 0; i < keys.length; ++i) {
			VanishBackendException exception = null;
			for (int step = 0; step < 3; ++step) {
				try {
					putOneShare(locations[i], keys[i]);
					exception = null;
					break;
				} catch (VanishBackendException e) {
					exception = e;
				}
			}
			if (exception != null) throw exception;
		}
		
		log_stat.end();
	}
	
	// @Override
	public byte[] generateLocation(Random prg) {
		byte[] index = new byte[params.kDhtIndexLen];
		prg.nextBytes(index);
		return index;
	}
	
	// Helpers:
	
	@SuppressWarnings("unchecked")
	private byte[] getOneShare(byte[] location)
	throws VanishBackendException {
		OpenDHTServer client = chooseServer();
		if (client == null) {
			throw new VanishBackendException(
					"Failed to choose OpenDHT server");
		}
		
		Vector op_params = new Vector();
		op_params.add(location);  // key
		op_params.add(new Integer(1));  // maxvals
		op_params.add(new byte[0]);   // placemark_bytes
		op_params.add("vanish");  // application.
		
		// Execute get:
		Object[] get_result = null;
		try {
			synchronized(client) {
				get_result = (Object[]) client.execute("get", op_params);
			}
		} catch (Exception e) {
			e.printStackTrace();
			reportError(client);
			return null;
		}
		if (get_result == null || get_result.length == 0) return null;
		Object[] values = (Object[])get_result[0];
		if (values == null || values.length == 0) return null;
		return (byte[])values[0];
	}
	
	@SuppressWarnings("unchecked")
	private boolean putOneShare(byte[] location, byte[] value)
	throws VanishBackendException {
		OpenDHTServer server = chooseServer();
		if (server == null) {
			throw new VanishBackendException(
					"Failed to choose OpenDHT server");
		}
		
		Vector op_params = new Vector();
		op_params.add(location);  // key
		op_params.add(value);  // value
		op_params.add(new Integer(OPENDHT_TIMEOUT_S));  // object ttl
		op_params.add("vanish");  // application.
		
	    // Execute put:
		Integer put_result = null;
		try {
			synchronized(server) {
				put_result = (Integer) server.execute("put", op_params);
			}
		} catch (Exception e) {
			e.printStackTrace();
			reportError(server);
		}
		if (put_result != null && put_result == 0) {
			return true;
		} else {
			throw new VanishBackendException("Failed to put.");
		}
	}
	
	private OpenDHTServer chooseServer() {
		return current_server;
	}
	
	private void reportError(OpenDHTServer server) {
		servers_list.removeElement(server.getHost());
		
		chooseCurrentServer();
	}
	
	private void chooseCurrentServer() {
		String dht_server_name = null;
		synchronized(servers_list_lock) {
			if (servers_list.size() > 0) {
				dht_server_name =
					servers_list.get(random.nextInt(servers_list.size()));
			}
		}
		if (dht_server_name != null) {
			LOG.info("OpenDHTVanishBackend.chooseCurrentServer: " +
					 "Switching to server " + dht_server_name);
            try { 
            	// Switch happens atomically.
	            current_server = new OpenDHTServer(dht_server_name,
					OpenDHTVanishBackendDefaults.OPENDHT_GATEWAY_PORT);
			} catch (MalformedURLException e) {
				e.printStackTrace();
			}
		}
	}
	
	private class RefreshServerListTask extends TimerTask {
		private final HttpClient server = new HttpClient();
		
		@Override
		public void run() {
			final HttpMethod get_method = new GetMethod(params.kServerListURL);
			get_method.getParams().setParameter(
					HttpMethodParams.RETRY_HANDLER,
					new DefaultHttpMethodRetryHandler(3, false));
			
			try {
				int status_code = server.executeMethod(get_method);
				if (status_code != HttpStatus.SC_OK) {
					LOG.error("RefreshServerListTask.run: " +
							  "Error downloading OpenDHT server list");
				}
				String server_list_blob =
					new String(get_method.getResponseBody());
				updateServerList(server_list_blob);				
			} catch (Exception e) {
				LOG.error("RefreshServerListTask: Failed to get the servers " +
						  "list", e);
			} finally {
				get_method.releaseConnection();
			}
		}
	}

	private void updateServerList(String new_list_blob) {
		String[] new_servers_list = new_list_blob.split("\n");
		Vector<String> new_servers = new Vector<String>(); 
		for (int i = 1; i < new_servers_list.length; ++i) {
			try {
				String new_server = new_servers_list[i];
				if (new_server.toLowerCase().endsWith(".edu")) {
					new_servers.add(new_server);
				}
			} catch (Exception e) { }
		}
		synchronized(servers_list_lock) {
			servers_list = new_servers;
		}
	}
	
	// RefreshableConfig interface:
	
	// @Override
	public DHTParams getRenewedParamsFromConfig(VanishConfiguration config) {		
		final int kDhtIndexLen = config.getInt(
				OpenDHTVanishBackendDefaults.CONF_VANISH_BACKEND_OPENDHTIMPL_DHTINDEXLEN,
				OpenDHTVanishBackendDefaults.CONF_VANISH_BACKEND_OPENDHTIMPL_DHTINDEXLEN_DEFAULT);
		final int kOperationTimeout_s = config.getInt(
				OpenDHTVanishBackendDefaults.CONF_VANISH_BACKEND_OPENDHTIMPL_OPERATION_TIMEOUT,
				OpenDHTVanishBackendDefaults.CONF_VANISH_BACKEND_OPENDHTIMPL_OPERATION_TIMEOUT_DEFAULT);
		final String kServerListURL = config.getString(
				OpenDHTVanishBackendDefaults.CONF_VANISH_BACKEND_OPENDHTIMPL_SERVER_URL,
				OpenDHTVanishBackendDefaults.CONF_VANISH_BACKEND_OPENDHTIMPL_SERVER_URL_DEFAULT);
		final long kRefreshServerListInterval = config.getTimeInMillis(
				OpenDHTVanishBackendDefaults.CONF_VANISH_BACKEND_OPENDHTIMPL_REFRESH_LIST_INTERVAL,
				OpenDHTVanishBackendDefaults.CONF_VANISH_BACKEND_OPENDHTIMPL_REFRESH_LIST_INTERVAL_DEFAULT,
				TimeUnit.MINUTES);
		
		DHTParams params = new DHTParams(kDhtIndexLen, kOperationTimeout_s,
				kServerListURL, kRefreshServerListInterval);
		
		// validate(params);
		return params;
	}

	// @Override
	public void setParams(DHTParams params) {
		this.params = params;
	}

	public String getIdentifier() {
		return "opendht";
	}
}

class OpenDHTServer extends XmlRpcClient {
  private String hostname;

  public OpenDHTServer(String hostname, int port) throws MalformedURLException {
      super();
      this.hostname = hostname;

      XmlRpcClientConfigImpl config = new XmlRpcClientConfigImpl();
      config.setServerURL(new URL("http://" + hostname + ":" + port + "/RPC2"));
      setConfig(config);
  }

  public String getHost() {
	return hostname;
  }
}

