/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.internal.backend.rpcimpl;

import java.net.MalformedURLException;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.Random;

import org.apache.xmlrpc.XmlRpcException;
import org.apache.xmlrpc.client.XmlRpcClient;
import org.apache.xmlrpc.client.XmlRpcClientConfigImpl;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.backend.TooManySharesLostVanishException;
import edu.washington.cs.vanish.internal.backend.VanishBackendException;
import edu.washington.cs.vanish.internal.backend.VanishBackendInterface;
import edu.washington.cs.vanish.logging.LogStat;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.util.Constants;
import edu.washington.cs.vanish.util.Defaults;
/**
 * 
 * @author roxana
 *
 */
public class RPCVanishBackendImpl implements VanishBackendInterface {
	private static final int  kIndexLen = 20;  // TODO(roxana): mk configurable
	private static final long kTimeToSleepBetweenBackendInitChecks =
		5 * Constants.SECONDS;  // TODO(roxana): mk configurable.
	
	private final URL kVanishBackendURL;
	private final String kVanishBackendServiceName;
	private final int kRPCTimeout;

	private final VanishLogger LOG;
	private XmlRpcClient client;
	
	public RPCVanishBackendImpl(VanishConfiguration config, VanishLogger LOG)
	throws InvalidArgumentVanishException {
		this.LOG = LOG;
		
		String default_url = "http://localhost:" +
			(config.getInt(Defaults.CONF_VANISH_PORT,
					      Defaults.CONF_VANISH_PORT_DEFAULT) + 1) +
			"/xmlrpc";
		try {
			kVanishBackendURL = new URL(config.getString(
					RPCVanishBackendDefaults.CONF_VANISH_BACKEND_URL, default_url));
		} catch (MalformedURLException e) {
			throw new InvalidArgumentVanishException("");
		}
		kRPCTimeout = (int)config.getTimeInMillis(
				RPCVanishBackendDefaults.CONF_VANISH_BACKEND_TIMEOUT,
				RPCVanishBackendDefaults.CONF_VANISH_BACKEND_TIMEOUT_DEFAULT,
				VanishConfiguration.TimeUnit.SECONDS);
		kVanishBackendServiceName = config.getString(
				RPCVanishBackendDefaults.CONF_VANISH_BACKEND_SERVICE_NAME,
				RPCVanishBackendDefaults.CONF_VANISH_BACKEND_SERVICE_NAME_DEFAULT);
	}

    // @Override
	public byte[] generateLocation(Random prg) {
		byte[] index = new byte[kIndexLen];
		prg.nextBytes(index);
		return index;
	}

	// The implementations below are extremely hacky!!!
	
    // @Override
	public byte[][] getShares(byte[][] locations, int threshold)
	throws VanishBackendException, TooManySharesLostVanishException {
		if (!isInitialized()) {
			throw new VanishBackendException("Invalid uninitialized state");
		}
		LogStat log_stat = LOG.logStat("RPCVanishBackendImpl.getShares");
		VanishException exception = null;
		try {
			return fromByteArrayToByteArray((Object[])client.execute(
					kVanishBackendServiceName + ".getShares",
					new Object[] {locations, threshold}));
		} catch (XmlRpcException e) {
			Throwable t = e.getCause();
			if (t != null && t instanceof TooManySharesLostVanishException) {
				exception = (TooManySharesLostVanishException)t;
				throw (TooManySharesLostVanishException)t;
			} else if (t != null && t instanceof VanishBackendException) {
				exception = (VanishBackendException)t;
				throw (VanishBackendException)t;
			} else {
				exception = new VanishBackendException(
						"Failed to execute getShares", t);
				throw (VanishBackendException)exception;
			}
		} finally {
			log_stat.end(exception);
		}
	}

    // @Override
	public void pushShares(byte[][] shares, byte[][] locations)
	throws VanishBackendException {
		if (!isInitialized()) {
			throw new VanishBackendException("Invalid uninitialized state");
		}
		LogStat log_stat = LOG.logStat("RPCVanishBackendImpl.pushShares");
		VanishException exception = null;
		try {
			client.execute(kVanishBackendServiceName + ".pushShares",
					       new Object[] {shares, locations});
		} catch (XmlRpcException e) {
			Throwable t = e.getCause();
			if (t != null && t instanceof VanishBackendException) {
				exception = (VanishBackendException)t;
				throw (VanishBackendException)t;
			} else {
				exception = new VanishBackendException(
						"Failed to execute getShares", t);
				throw (VanishBackendException)exception;
			}
		} finally {
			log_stat.end(exception);
		}
	}
	
	/**
	 * Horrible hack that handles the apache-XMLRPC problem about 
	 * supporting params or return types byte[][].  
	 * @param objects
	 * @return
	 */
	public static byte[][] fromByteArrayToByteArray(Object[] objects) {
		byte[][] ba = new byte[objects.length][];
		for (int i = 0 ; i < ba.length; ++i) {
			ba[i] = (byte[])objects[i];
		}
		return ba;
	}

    // Initializable interface:

	// @Override
	public void init() throws InvalidArgumentVanishException {
		assert (!isInitialized());

		LogStat log_stat = LOG.logStat("RPCVanishBackendImpl.init");
		// Perform some special configs.
		System.setProperty("sun.net.client.defaultConnectTimeout",
				"" + kRPCTimeout);
		System.setProperty("sun.net.client.defaultReadTimeout",
				"" + kRPCTimeout);

		VanishException exception = null;
		try {
			try {
				initConnectionToVanishServer();
			} catch (UnknownHostException e) {
				throw new InvalidArgumentVanishException(
						"Failed to init vanish server connection", e);
			}

			waitUntilVuzeVanishBackendProcessIsInitialized();
		} catch (InvalidArgumentVanishException e) {
			exception = e;
			throw e;
		} finally {
			log_stat.end(exception);
		}
	}

	// @Override
	public boolean isInitialized() { return client != null; }

	// @Override
	public void stop() {
		try {
			client.execute(kVanishBackendServiceName + ".stop",
					       new Object[] { });
		} catch (XmlRpcException e) { }
	}

	private void initConnectionToVanishServer() throws UnknownHostException {
		XmlRpcClientConfigImpl config = new XmlRpcClientConfigImpl();
		config.setEnabledForExtensions(true);
		config.setEnabledForExceptions(true);
		config.setServerURL(kVanishBackendURL);
		if (kRPCTimeout > 0) {
			config.setReplyTimeout(kRPCTimeout);
			config.setConnectionTimeout(kRPCTimeout);
		}
		client = new XmlRpcClient();
		client.setConfig(config);
	}
	
	private void waitUntilVuzeVanishBackendProcessIsInitialized() {
		boolean initialized = false;
		while (! initialized) {
			try {
				initialized = (Boolean)client.execute(
						kVanishBackendServiceName + ".isInitialized",
						new Object[]{});
			} catch (XmlRpcException e) { }
			if (! initialized) {
				try { Thread.sleep(kTimeToSleepBetweenBackendInitChecks); }
				catch (InterruptedException e) { }
			}
		}
	}

	public String getIdentifier() {
		return "vuze";
	}
}
