/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.internal.crypto;

/**
 * The Shamir Secret Sharing interface. Allows splitting of secrets into
 * shares and composing those shares back into the original secrets.
 * <br>
 * Shamir's scheme is described in the following paper:
 * http://portal.acm.org/citation.cfm?id=359176
 * <br>
 * A C implementation of the Secret Sharing scheme can be found at:
 * http://point-at-infinity.org/ssss/
 * <br>
 * Other parameters of the SecretSharing scheme (e.g., the share size, a
 * pseudo-random generator) can be taken in by implementing functions via
 * their constructor.
 *  
 * @author roxana
 */
public interface SecretSharingInterface {
	/**
	 * Splits a secret into a set of shares using the Shamir Secret Sharing
	 * scheme.
	 * 
	 * @param secret  the secret to split.
	 * @param num_total_shares  the number of shares in which to split the
	 *        secret.
	 * @param threshold the threshold required to reconstruct the secret.
	 * @return the generated shares, any threshold of which are needed to
	 *        reconstruct the original secret. 
	 * @throws CryptoVanishException if an error occurs.
	 */
	public byte[][] split(byte[] secret, int num_total_shares,
			              int threshold)
	throws CryptoVanishException;
	
	/**
	 * Combines a set of shares into a secret using the Shamir Secret Sharing
	 * scheme. These shares are assumed to have been obtained by applying
	 * split on a set of shares.
	 * 
	 * @param shares  the shares obtained from a split. These shares should be
	 *        combined back into the original secret.
	 * @param threshold the threshold to use. Can assume that
	 *        shares.length >= share_threshold. 
	 * @return the secret reconstructed from the shares.
	 * @throws CryptoVanishException if an error occurs.
	 */
	public byte[] compose(byte[][] shares, int threshold)
	throws CryptoVanishException;
}
