/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.internal.crypto.impl;

import java.security.SecureRandom;
import java.util.*;
import java.math.*;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.crypto.SecretSharingInterface;
import edu.washington.cs.vanish.internal.crypto.CryptoVanishException;
import edu.washington.cs.vanish.logging.LogStat;
import edu.washington.cs.vanish.logging.VanishLogger;

/**
 * KNOWN ISSUE: Mod-prime math not used in this implementation of the
 * Shamir Secret Sharing algorithm, which can lead to security issues.
 * 
 * TODO(roxana): Fix this.
 */

/**
 * An implementation of the SecretSharing interface that uses Shamir's
 * algorithm to encrypt and decrypt secrets by creating keys such that a
 * given number of the keys are needed to recreate a secret.
 * 
 * @author Andrew Morgan
 */
public class ShamirSecretSharing implements SecretSharingInterface {
	@SuppressWarnings("unused") private final VanishConfiguration CONF;
	private final VanishLogger LOG;
	
	private final Random random;

	public ShamirSecretSharing(SecureRandom random,
			                   VanishConfiguration conf,
			                   VanishLogger log)
	throws InvalidArgumentVanishException {
		this.random = random;
		CONF = conf;
		LOG = log;
	}
	
	/**
	 * Splits a secret (byte array) into keys (byte arrays) by creating a
	 * polynomial with the secret as its constant term and creating keys
	 * based on the values of the polynomial at various points.
	 * Post: the array at index (n) in the returned array represents
	 * key (n + 1), or the polynomial evaluated at x = n + 1.
	 */
	// @Override
    public byte[][] split(byte[] secret, int num_total_shares, int threshold)
	throws CryptoVanishException {
    	final LogStat log_stat = LOG.logStat("ShamirSecretSharing.split");
    	if (num_total_shares == 1) {
    		assert(threshold == 1);
    		log_stat.end();
    		return new byte[][] { secret };
    	}
    	if (threshold == 1) {
    		byte[][] shares = new byte[num_total_shares][];
    		for (int i = 0; i < num_total_shares; ++i) {
    			shares[i] = secret;
    		}
    		log_stat.end();
    		return shares;
    	}
		
        if(secret == null || secret[0] == 0 || secret[0] == -1) {
            throw new CryptoVanishException("Invalid secret.");
        } else if(num_total_shares < threshold || threshold <= 0){
            throw new CryptoVanishException("Illegal parameters.");
        }
        BigInteger[] coef = new BigInteger[threshold]; // degree = threshold.
        coef[0] = new BigInteger(secret); // secret is constant term.
        for(int i = 1; i < threshold; i++){
            coef[i] = new BigInteger(secret.length * 8, random); // random coefs
        }
        byte[][] res = new byte[num_total_shares][];
        for(int i = 1; i <= num_total_shares; i++){
            res[i - 1] = polyEval(coef, i).toByteArray(); // evaluate polyn.
        }
        
        log_stat.end();
        return res;
    }
    
    /**
     * Evaluates the given polynomial at the given x point.
     * @param coef
     * @param n
     * @return
     */
    private BigInteger polyEval(BigInteger[] coef, int n){
        BigInteger res = BigInteger.ZERO;
        BigInteger k = new BigInteger("" + n);
        for(int i = 0; i < coef.length; i++){
            res = res.add(coef[i].multiply(k.pow(i)));
        }
        return res;
    }
    
    /**
     * Composes a secret (byte array) from keys (byte arrays) by determining
     * the constant term of the polynomial by which the keys were created using
     * Lagrange polynomials.
     * Pre: the array at any index (n) in the given array of keys must be the
     * correct key (the key generated by evaluating the polynomial at
     * x = (n + 1)) or null (representing that the correct key was not given
     * or available).
     */
    // @Override
    public byte[] compose(byte[][] shares, int threshold)
    throws CryptoVanishException {
    	LogStat log_stat = LOG.logStat("ShamirSecretSharing.compose");
    	if (shares.length == 1) {
    		assert(threshold == 1);
    		log_stat.end();
    		return shares[0];
    	}
    	if (threshold == 1) {
    		for (byte[] share: shares) {
    			if (share != null) return share;
    		}
    		log_stat.end();
    		return null;
    	}
    	
        if (threshold < 1) {
        	throw new CryptoVanishException("Illegal argument.");
        }
        int[] xVar = new int[threshold]; // the x-values
        BigInteger[] yVar = new BigInteger[threshold]; // the corresp. y-values
        int index = 0;
        for (int i = 0; index < threshold; i++) {
        	// Place enough given keys into x/y arrays.
            if(i == shares.length) {
                throw new CryptoVanishException("Not enough shares.");
            }
            if(shares[i] != null) {
                xVar[index] = i + 1;
                yVar[index] = new BigInteger(shares[i]);
                index++;
            }
        }
        
        /*
         * The following process determines the constant term of the generating 
         * polynomial using part of the Lagrange interpolation, found at
         * http://en.wikipedia.org/wiki/Lagrange_polynomial. The given method
         * is used with x = 0 to find the constant term of the polynomial.
         * Also, decimals are avoided in the computation because as great a
         * degree of precision as possible is necessary.
         */
        BigInteger[] aVar = new BigInteger[threshold];
        BigInteger[] bVar = new BigInteger[threshold];
        for(int i = 0; i < threshold; i++) {
            BigInteger a = BigInteger.ONE;
            BigInteger b = BigInteger.ONE;
            for(int j = 0; j < threshold; j++) {
                if(j != i){
                    a = a.multiply(new BigInteger("" + xVar[j]).negate());
                    b = b.multiply(new BigInteger("" + xVar[i])
                                   .subtract(new BigInteger("" + xVar[j])));
                }
            }
            aVar[i] = a;
            bVar[i] = b;  
        }
        BigInteger denom = BigInteger.ONE;
        for(int i = 0; i < threshold; i++) {
            denom = lcm(denom, bVar[i]);
        }
        BigInteger result = BigInteger.ZERO;
        for(int i = 0; i < threshold; i++) {
            result = result.add(aVar[i].multiply(yVar[i])
            		            .multiply(denom.divide(bVar[i])));
        }
        
        log_stat.end();
        return result.divide(denom).toByteArray(); //constant term is returned.
    }
    
    /**
     * Returns the least common multiple of two BigIntegers.
     */
    private static BigInteger lcm(BigInteger a, BigInteger b) {
        return a.divide(a.gcd(b)).multiply(b);
    }
}
