/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.internal.crypto.impl;

import javax.crypto.Cipher;
import javax.crypto.CipherInputStream;
import javax.crypto.KeyGenerator;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.crypto.CryptoVanishException;
import edu.washington.cs.vanish.internal.crypto.SymmetricEncDec;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.logging.LogStat;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.util.Constants;
import edu.washington.cs.vanish.util.Defaults;
import edu.washington.cs.vanish.util.RefreshableConfig;
import edu.washington.cs.vanish.util.RefreshableParams;
import edu.washington.cs.vanish.util.FileUtil;
import edu.washington.cs.vanish.util.IOVanishException;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.GeneralSecurityException;
import java.security.Key;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;

/**
 * 
 * @author roxana
 */
class CryptoParams implements RefreshableParams {
	final String kEncryptionAlgorithm;
	final String kEncryptionMode;
	final int kEncryptionKeyLength_bits;
	
	public CryptoParams(String encryptionAlgorithm, String encryptionMode,
			int encryptionKeyLength_bits) {
		this.kEncryptionAlgorithm = encryptionAlgorithm;
		this.kEncryptionMode = encryptionMode;
		this.kEncryptionKeyLength_bits = encryptionKeyLength_bits;
	}
}

public class SymmetricEncDecImpl
implements SymmetricEncDec, RefreshableConfig<CryptoParams>, Constants {
	private final VanishLogger LOG;
	
	// Configuration parameters:
	private final VanishConfiguration CONF;
	private CryptoParams params;
	
	private static final byte[] kIvParameterBaseValue =
		new byte[] { 12, 34, 56, 78, 90, 87, 65, 43 };

	// State:
	private final KeyGenerator default_settings_keygen;

	private final SecureRandom random;

    public SymmetricEncDecImpl(SecureRandom random, VanishConfiguration conf,
    		                   VanishLogger log)
    throws InvalidArgumentVanishException {    	
    	this.LOG = log;
    	
    	this.CONF = conf;
    	CryptoParams params = getRenewedParamsFromConfig(CONF);
    	setParams(params);
    	
    	this.random = random;
    	
    	// Init crypto services operating with default params.
    	
    	default_settings_keygen = getNewKeyGenerator(params.kEncryptionAlgorithm,
    			params.kEncryptionKeyLength_bits);
    }

    // Initializable interface:

    // @Override
	public void init() {
		// keygen.init(params.kEncryptionKeyLength_bits, random);
	}

	// @Override
	public boolean isInitialized() { return default_settings_keygen != null; }

	// @Override
	public void stop() throws CryptoVanishException { }
	
	// SymmetricEncDec interface:

	// @Override
    public InputStream encrypt(InputStream clear_is, byte[] key_in,
    		VDOParams vdo_params) 
    throws CryptoVanishException, InvalidArgumentVanishException {
    	return performCryptOperation(clear_is, key_in, Cipher.ENCRYPT_MODE,
    			vdo_params);
  	}

    // @Override
    public InputStream decrypt(InputStream enc_is, byte[] key_in,
    		VDOParams vdo_params)
    throws CryptoVanishException, InvalidArgumentVanishException {
    	return performCryptOperation(enc_is, key_in, Cipher.DECRYPT_MODE,
    			                     vdo_params);
  	}
    
    private InputStream performCryptOperation(InputStream is, byte[] key_in,
    		int mode, VDOParams vdo_params)
    throws CryptoVanishException, InvalidArgumentVanishException {
    	String op_name = (mode == Cipher.ENCRYPT_MODE ? "encrypt" : "decrypt");
    	final LogStat logstat = LOG.logStat("SymmetricEncDecImpl." + op_name);
    	Cipher cipher = initCipher(key_in, mode, vdo_params);
   		InputStream ret = new CipherInputStream(is, cipher);
   		logstat.end();
   		return ret;
    }

    // @Override
    public byte[] generateKey(VDOParams params)
    throws CryptoVanishException, InvalidArgumentVanishException {
    	Key k = null;
    	KeyGenerator kg = getCachedOrNewKeyGenerator(
    			params.getEncryptionAlgorithm(),
    			params.getEncryptionKeyLengthBits());
    	// assert(keygen != null); 
    	synchronized (kg) { k = kg.generateKey(); }
    	return k.getEncoded();
    }
 
    // @Override
	public void encryptToOutputStream(InputStream clear_is,
			                          OutputStream enc_os,
			                          byte[] key,
			                          VDOParams vdo_params)
	throws CryptoVanishException, InvalidArgumentVanishException {
		InputStream enc_is = encrypt(clear_is, key, vdo_params);

		try {
			FileUtil.writeInputOutput(enc_is, enc_os);
		} catch (IOVanishException e) {
			throw new CryptoVanishException(
					"Failed to encrypt output stream", e);
		} finally {
			try { enc_is.close(); }
			catch (IOException e) { e.printStackTrace(); }
		}
	}
	
	// Helper functions:
	
	private Cipher initCipher(byte[] key_in, int mode, VDOParams vdo_params)
    throws CryptoVanishException, InvalidArgumentVanishException {
    	Cipher cipher = null;
    	final String algorithm = vdo_params.getEncryptionAlgorithm();
    	final String enc_mode  = vdo_params.getEncryptionMode();
    	final int enc_key_bits = vdo_params.getEncryptionKeyLengthBits();
    	final int enc_key_bytes = enc_key_bits / 8;
    	try {
    		cipher = Cipher.getInstance(getEncryptionAlgorithmName(
    				algorithm, enc_mode));
    	} catch (Exception e1) {
    		throw new InvalidArgumentVanishException(
    				"Invalid algorithm " + algorithm); 
    	}

    	byte[] key = key_in;
    	if (key.length > enc_key_bytes) {
    		key = new byte[enc_key_bytes];
    		for (int i = 0; i < enc_key_bytes; i++) key[i] = key_in[i];
    	}

    	Key k = new SecretKeySpec(key, algorithm);
        try {
        	if (enc_mode.equals("CBC")) {
        		cipher.init(mode, k, getNewInitializationVectorForCBC(
        				       	cipher.getBlockSize()),
        			        random);
        	} else {
        		cipher.init(mode, k, random);
        	}
        } catch (GeneralSecurityException e) {
        	throw new CryptoVanishException("Failed to initialize cipher", e);
        }

        return cipher;
    }
	
	private IvParameterSpec getNewInitializationVectorForCBC(
			 int key_length_bytes) {
		int num_repetitions_iv = key_length_bytes /
		                         kIvParameterBaseValue.length;
	    byte[] iv = null;
	    if (num_repetitions_iv == 1) {
	    	iv = kIvParameterBaseValue;
	    } else {  // extend the existing initialization vector further.
	    	iv = new byte[key_length_bytes];
	    	for (int i = 0; i < num_repetitions_iv; ++i) {
	    		for (int j = 0; j < kIvParameterBaseValue.length; ++j) {
	    			iv[i * kIvParameterBaseValue.length + j] =
	    				kIvParameterBaseValue[j];
	    		}
	    	 }
	     }
	     return new IvParameterSpec(iv);
	}
	
	private KeyGenerator getCachedOrNewKeyGenerator(String algorithm,
			int key_len_bits)
	throws InvalidArgumentVanishException {
		if (algorithm.equals(params.kEncryptionAlgorithm) &&
			key_len_bits == params.kEncryptionKeyLength_bits) {
			return default_settings_keygen;
		} else {
			return getNewKeyGenerator(algorithm, key_len_bits);
		}
	}
	
	private KeyGenerator getNewKeyGenerator(String algorithm, int key_len_bits)
	throws InvalidArgumentVanishException {
		KeyGenerator kg = null;
		try {
			kg = KeyGenerator.getInstance(algorithm);
			kg.init(key_len_bits);
    	} catch (NoSuchAlgorithmException e) {
			throw new InvalidArgumentVanishException(
					"Failed to create keygen", e);
		}
    	return kg;
	}
	
	public static void validate(CryptoParams params)
	throws InvalidArgumentVanishException {
		if (params.kEncryptionKeyLength_bits < 128 ||
			params.kEncryptionKeyLength_bits > 256 ||
			params.kEncryptionKeyLength_bits % 64 != 0) {
			throw new InvalidArgumentVanishException("Encryption key is " +
					"too short (vanish.internal.crypto.keylength)");
		}
	}
	
	private String getEncryptionAlgorithmName(String enc_algo,
			                                  String enc_mode) {
		return enc_algo + "/" + enc_mode + "/PKCS5Padding";
	}

	public CryptoParams getRenewedParamsFromConfig(VanishConfiguration config)
			throws InvalidArgumentVanishException {
		final String kEncryptionAlgorithm = CONF.getString(
				Defaults.CONF_VANISH_CRYPTO_ENCALGORITHM,
				Defaults.CONF_VANISH_CRYPTO_ENCALGORITHM_DEFAULT);
		final String kEncryptionMode = CONF.getString(
				Defaults.CONF_VANISH_CRYPTO_ENCMODE,
				Defaults.CONF_VANISH_CRYPTO_ENCMODE_DEFAULT);
		final int kEncryptionKeyLength_bits = CONF.getInt(
				Defaults.CONF_VANISH_CRYPTO_ENCKEYLENGTH_BITS,
				Defaults.CONF_VANISH_CRYPTO_ENCKEYLENGTH_BITS_DEFAULT);

		CryptoParams params =
			new CryptoParams(kEncryptionAlgorithm, kEncryptionMode,
				kEncryptionKeyLength_bits);
		validate(params);
		return params;
	}

	public void setParams(CryptoParams params) { this.params = params; }
}
