/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.internal.impl;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Map;
import java.util.Random;
import java.util.concurrent.Callable;

import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.VanishInternalInterface;
import edu.washington.cs.vanish.internal.backend.TooManySharesLostVanishException;
import edu.washington.cs.vanish.internal.backend.VanishBackendException;
import edu.washington.cs.vanish.internal.backend.VanishBackendInterface;
import edu.washington.cs.vanish.internal.crypto.CryptoVanishException;
import edu.washington.cs.vanish.internal.crypto.SecretSharingInterface;
import edu.washington.cs.vanish.internal.crypto.SymmetricEncDec;
import edu.washington.cs.vanish.internal.metadata.Metadata;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.logging.LogStat;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.util.IOVanishException;
import edu.washington.cs.vanish.util.Immutable;
import edu.washington.cs.vanish.util.FileUtil;

/**
 * 
 * @author roxana
 */
public class BasicVanishImpl implements VanishInternalInterface {
	private final VanishLogger LOG;
	
	private final Map<String, VanishBackendInterface> backend_modules;
	
	private final SymmetricEncDec crypto_module;  // TODO(roxana): combine w/ sss.
	
	private final SecretSharingInterface sss_module;

	public BasicVanishImpl(Map<String, VanishBackendInterface> backend_modules,
                           SymmetricEncDec crypto_module,
                           SecretSharingInterface sss_module,
                 		   VanishLogger log) {
		assert(backend_modules != null  && ! backend_modules.isEmpty() &&
			   crypto_module != null &&
			   sss_module != null);
		this.crypto_module = crypto_module;
		this.sss_module = sss_module;
		this.backend_modules = backend_modules;
		this.LOG = log;
	}

	// Initializable interface:

	// @Override
	public void init() throws VanishException {
		for (VanishBackendInterface backend: backend_modules.values()) {
			backend.init();
		}
		crypto_module.init();  // must be last for isInitialized to work.
	}

	// @Override
	public boolean isInitialized() {
		return crypto_module.isInitialized();
	}

	// @Override
	public void stop() throws VanishException {
		for (VanishBackendInterface backend: backend_modules.values()) {
			backend.stop();
		}
	}

	// Vanish interface:

	// @Override
	public void encapsulate(InputStream clear_data_is, 
                            OutputStreamWrapper enc_data_os,
                            Metadata metadata)
	throws VanishException {
		final LogStat logstat = LOG.logStat("BasicVanishImpl.encapsulate");
		
		validateParamsEncapsulate(clear_data_is, enc_data_os, metadata);
		
		// Create and push a new encryption key.
		byte[] encryption_key = null;
		try {
			// Generate an encryption key.
			encryption_key = crypto_module.generateKey(
					metadata.getVDOParams());
			
			// Push it into the backend.
			pushEncryptionKeyIntoBackend(encryption_key, metadata);
			
			// Encrypt using the key.
			encapsulateUsingPushedKey(clear_data_is, enc_data_os, metadata,
					                  encryption_key);
		} finally {  // Clean up in any case.
			VanishUtil.cleanSensitiveData(encryption_key);
		}

		logstat.end();
	}

	// @Override
	public void encapsulateUsingPrepushedMetadata(InputStream clear_data_is,
            OutputStreamWrapper enc_data_os,
            @Immutable Metadata metadata)
	throws VanishException {
		final LogStat logstat = LOG.logStat(
				"BasicVanishImpl.encapsulateUsingPrepushedMetadata");

		validateParamsEncapsulate(clear_data_is, enc_data_os, metadata);
		
		byte[] key = null;
		try {
			// Get the decryption key.
			key = getDecryptionKeyFromBackend(metadata);
			
			// Encrypt using the key.
			encapsulateUsingPushedKey(clear_data_is, enc_data_os, metadata,
					                  key);
		} finally {  // Clean up in any case.
			VanishUtil.cleanSensitiveData(key);
		}
		
		logstat.end();
	}
	
	private void encapsulateUsingPushedKey(
			InputStream clear_data_is,
            OutputStreamWrapper enc_data_os,
            Metadata metadata,
            byte[] encryption_key)
	throws VanishException {
		if (clear_data_is == null || enc_data_os == null) return;
		
		// Encrypt the data using the encryption key.
		InputStream encrypted_data_is = crypto_module.encrypt(clear_data_is,
				encryption_key, metadata.getVDOParams());

		if (encrypted_data_is == null) {
			throw new CryptoVanishException("Decrypt returned null.");
		}
		
		// Write the encrypted data to the OutputStream.
		try {
			writeEncryptedDataToOutputStream(encrypted_data_is, enc_data_os,
					                         metadata);
		} finally {
			try { encrypted_data_is.close(); }
			catch (IOException e) { e.printStackTrace(); }
		}
	}
	
	/**
	 * Override if you want to write other things (e.g., the metadata) in the
	 * OutputStream.
	 * @param encrypted_data_is
	 * @param encrypted_data_os
	 * @param metadata
	 * @throws IOVanishException 
	 * @throws VanishException
	 */
	protected void writeEncryptedDataToOutputStream(
			InputStream encrypted_data_is,
			OutputStreamWrapper encrypted_data_os,
			@SuppressWarnings("unused") @Immutable Metadata metadata)
 	throws VanishException {
		if (encrypted_data_is == null || encrypted_data_os == null) return;
		FileUtil.writeInputOutput(encrypted_data_is, encrypted_data_os);
	}
	
	// @Override
	public void decapsulate(InputStreamWrapper enc_data_is,
			                OutputStream clear_data_os,
			                Metadata metadata)
	throws VanishException {
		final LogStat logstat = LOG.logStat("BasicVanishImpl.decapsulate"); 
		
		validateParamsDecapsulate(enc_data_is, clear_data_os, metadata);
    	byte[] decryption_key = null;
    	try {
    		// Get the decryption key from the backend.
        	decryption_key = getDecryptionKeyFromBackend(metadata);
        	
        	// And decrypt using the key.
        	decapsulateUsingKey(enc_data_is, clear_data_os, metadata,
        			            decryption_key);
    	} finally {  // Clean up in any case.
    		VanishUtil.cleanSensitiveData(decryption_key);
    	}

    	logstat.end();
	}
	
	private void decapsulateUsingKey(InputStream enc_data_is,
                                     OutputStream clear_data_os,
                                     Metadata metadata,
                                     byte[] decryption_key)
	throws CryptoVanishException, IOVanishException,
	       InvalidArgumentVanishException {
		if (enc_data_is == null || clear_data_os == null) return;
		InputStream clear_data_is = null;
		try {
			// Decrypt the data using the decryption key.
			clear_data_is = crypto_module.decrypt(enc_data_is, decryption_key,
					metadata.getVDOParams());
			if (clear_data_is == null) {
				throw new CryptoVanishException("Decrypt returned null IS.");
			}
    	
			// Write the clear-text to the OutputStream.
			writeDecryptedDataToOutputStream(clear_data_is, clear_data_os,
					                         metadata, decryption_key);
		} finally {
			VanishUtil.cleanSensitiveData(clear_data_is);  // also closes it.
		}
	}
	
	/**
	 * Override if you want to write something else into the output stream
	 * upon decryption.
	 * @param clear_data_is
	 * @param clear_data_os
	 * @param metadata
	 * @param decryption_key
	 * @throws IOVanishException 
	 * @throws VanishException
	 */
	protected void writeDecryptedDataToOutputStream(InputStream clear_data_is,
			OutputStream clear_data_os,
			@SuppressWarnings("unused") @Immutable Metadata metadata,
			byte[] decryption_key)
	throws IOVanishException {
		if (clear_data_is == null || clear_data_os == null) return;
		FileUtil.writeInputOutput(clear_data_is, clear_data_os);
	}

	// @Override
	public void refreshReadOnly(@Immutable Metadata old_metadata)
	throws VanishException {
		final LogStat logstat = LOG.logStat("BasicVanishImpl.refreshReadOnly");

		validateParamsRefreshReadOnly(old_metadata);
		byte[] encryption_key = null;
		try {
			encryption_key = internalGetKeyFromBackend(old_metadata);
			if (encryption_key == null) {
				throw new VanishException("Null encryption/decryption key");
			}
			internalPushKeyIntoBackend(encryption_key, old_metadata);
		} finally {
			VanishUtil.cleanSensitiveData(encryption_key);
		}
		
		logstat.end();
	}

	// @Override
	public void refreshDeeply(
			final InputStreamWrapper enc_data_is,
			final OutputStreamWrapper refreshed_enc_data_os,
            @Immutable final Metadata old_metadata,
            final Metadata new_metadata)
	throws VanishException {
		final LogStat logstat = LOG.logStat("BasicVanishImpl.refreshDeeply");
		validateParamsRefreshDeeply(enc_data_is, refreshed_enc_data_os,
				old_metadata,
				new_metadata);
		
		RefreshThread<byte[]> get_thread =
        	new RefreshThread<byte[]>(new Callable<byte[]>() {
				public byte[] call() throws Exception {
					// Get the decryption key from the backend.
		        	return getDecryptionKeyFromBackend(old_metadata);
				}
        	});
		
		RefreshThread<byte[]> put_thread =
			new RefreshThread<byte[]>(new Callable<byte[]>() {
				public byte[] call() throws VanishException {
					byte[] new_encryption_key =
						crypto_module.generateKey(new_metadata.getVDOParams());
					pushEncryptionKeyIntoBackend(new_encryption_key,
							                     new_metadata);
					return new_encryption_key;
				}
        	});
                
        get_thread.start();
        put_thread.start();
        
        try {
        	get_thread.join();
            put_thread.join();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		
		byte[] decryption_key = null;
		byte[] new_encryption_key = null;
		InputStream clear_data_is = null;
		try {
			try {
				decryption_key = get_thread.getResult();
				new_encryption_key = put_thread.getResult();
			} catch (Exception e) {
				if (e instanceof VanishException) throw (VanishException)e;
				else throw new VanishException("Failed to get result", e);
			}
			if (decryption_key == null || new_encryption_key == null) {
				throw new VanishException("Refresh threads have failed.");
			}
		
			// Decrypt the data using the old key.
			clear_data_is = crypto_module.decrypt(enc_data_is,
					decryption_key, old_metadata.getVDOParams());
			
			// Encrypt the data back, using the new key.
			this.encapsulateUsingPushedKey(clear_data_is,
					                       refreshed_enc_data_os,
					                       new_metadata,
					                       new_encryption_key);
		} finally {  // Always clean up.
			VanishUtil.cleanSensitiveData(decryption_key);
			VanishUtil.cleanSensitiveData(new_encryption_key);
			VanishUtil.cleanSensitiveData(clear_data_is);
		}
		
		logstat.end();
	}

	protected void pushEncryptionKeyIntoBackend(byte[] encryption_key,
			                                    Metadata object_metadata)
	throws VanishException {
    	this.internalPushKeyIntoBackend(encryption_key, object_metadata);
    }
    
	/**
	 * Stores an ecryption key into the backend.
	 * 
	 * @param encryption_key
	 * @param object_metadata  the Vanish object's metadata.
	 * @return
	 * @throws CryptoVanishException 
	 * @throws VanishBackendException 
	 * @throws VanishException
	 */
	private void internalPushKeyIntoBackend(byte[] encryption_key,
			                                Metadata object_metadata)
	throws CryptoVanishException, VanishBackendException,
		   InvalidArgumentVanishException {
		VDOParams object_params = object_metadata.getVDOParams();
		
		VanishBackendInterface[] selected_backends = 
			selectBackends(object_params);
		
		// Split the master key into |backends| pieces, all of which are
		// required to reconstruct.
		byte[][][] key_shares = new byte[selected_backends.length][][];
		
		// TODO(roxana): Horrible hack because of the secret-rejection in
		// the SSS. Ask Amit why secret rejection happens...
		boolean ok = false;
		for (int s = 0; s < 10; ++s) {
			byte[][] master_key_shares = sss_module.split(encryption_key,
				selected_backends.length, selected_backends.length);
			boolean error = false;
			for (int j = 0; j < selected_backends.length; ++j) {
				byte[] master_key_share = master_key_shares[j];
                // Split key slare into many keys.
				try {
					key_shares[j] = splitForBackend(selected_backends[j],
							object_params, master_key_share);
				} catch (CryptoVanishException e) { 
					// TODO(roxana): Ask Amit WHY THIS EXC. IS NECESSARY!
					error = true;
					break;
				}
			}
			if (!error) {
				ok = true;
				break;
			}
		}
		
		if (! ok) throw new CryptoVanishException("Invalid secret....");
		
		if (selected_backends.length == 1) {
			byte[][] shares = key_shares[0];
			VanishBackendInterface backend = selected_backends[0];
			if (shares == null) {
				throw new CryptoVanishException("SSS returned null shares.");
			}
			// Generate the locations where to place the objects.
			byte[][] locations = generateLocations(backend,
					object_metadata.getLocationSeed(),
					shares.length);
			pushShares(backend, shares, locations, object_params);
		} else {
			Random random = new Random(object_metadata.getLocationSeed());
			for (int i = 0; i < selected_backends.length; ++i) {
				byte[][] shares = key_shares[i];
				VanishBackendInterface backend = selected_backends[i];
				if (shares == null) {
					throw new CryptoVanishException(
							"SSS returned null shares.");
				}
				// Generate the locations where to place the objects.
				byte[][] locations = generateLocations(backend,
						random.nextLong(),
						shares.length);
				pushShares(backend, shares, locations, object_params);
			}
		}
	}
	
	private byte[][] splitForBackend(VanishBackendInterface backend,
			VDOParams object_params, byte[] master_key_share)
	throws CryptoVanishException {
		int num_shares = object_params.getNumShares();
		int threshold = object_params.getThreshold();
		
		// TODO(roxana): Replace hack with per-backend VDOParam parameter!
		if (backend.getIdentifier().equals("opendht")) {
			num_shares = 3;
			threshold = 1;
		}
		
		return sss_module.split(master_key_share,
				num_shares, threshold);
	}
	
	private VanishBackendInterface[] selectBackends(VDOParams object_params)
	throws InvalidArgumentVanishException {
		String[] backend_ids = object_params.getBackendIdentifiers();
		
		if (backend_ids == null) {
			// TODO(roxana): REALLY need a version number in VDOParams!!!
			LOG.warn("VDOParams doesn't specify a network; entering backward " +
					"compat: assuming vuze");
			backend_ids = new String[] { "vuze" };
		}
		
		assert(backend_ids != null && backend_ids.length > 0);

		VanishBackendInterface[] selected_backends =
			new VanishBackendInterface[backend_ids.length];
		for (int i = 0; i < backend_ids.length; ++i) {
			if (! backend_modules.containsKey(backend_ids[i])) {
				throw new InvalidArgumentVanishException("Unknown DHT type " +
						backend_ids[i]);
			}
			selected_backends[i] = backend_modules.get(backend_ids[i]);
		}
		return selected_backends;
	}
	
	private void pushShares(VanishBackendInterface backend,
			                byte[][] shares, byte[][] locations,
			                VDOParams params)
	throws VanishBackendException {		
        // Push the shares in those locations.
		backend.pushShares(shares, locations);
	}
	
    protected byte[] getDecryptionKeyFromBackend(Metadata object_metadata)
    throws VanishException {
    	return this.internalGetKeyFromBackend(object_metadata);
    }
	
    /**
     * Reconstructs a decryption key for a Vanish object from the backend.
     * <br>
     * Goes through all of the object's possible location seeds and try to get
     * the shares at locations generated using each seed. Stops as soon as the
     * decryption key is successfully deconstructed.
     * <br>
     * 
     * TODO(roxana): How to ensure that the shares we are looking at are the
     * right ones and not some other object's? Now, we assume it can't happen
     * (low probability).
     * 
     * @param object_metadata  the vanish object's metadata.
     * @return the decryption key.<br>
     *         This should be later destroyed when no longer needed.
     * @throws VanishException 
     * @throws VanishException if the key could not be reconstructed.
     */
    private byte[] internalGetKeyFromBackend(Metadata object_metadata)
    throws VanishException {
        byte[] decryption_key = null;
        VanishException last_exception = null;
        
        VDOParams object_params = object_metadata.getVDOParams();
        long[] all_possible_location_seeds =
        	object_metadata.getAllPossibleLocationSeeds();
        LOG.info("BasicVanishImpl.internalGetKeyFromBackend: trying " +
        		 all_possible_location_seeds.length + " possible locations");
        for (long possible_loc_rand: all_possible_location_seeds) {
        	try {
        		decryption_key = reconstructKeyFromFixedLocations(
        				object_params, possible_loc_rand);
        	} catch (VanishException e) {
        		last_exception = e;
        		continue;  // try again w/ the next possible location seed.
        	}
        	
        	// If we've gotten a decryption key, stop.
        	if (decryption_key != null) break;
        }
        
        if (decryption_key == null) {
        	if (last_exception != null) throw last_exception;
        	else throw new VanishException("Failed to decrypt");  // not happen
        }
        
        return decryption_key;
    }
    
    private byte[] reconstructKeyFromFixedLocations(VDOParams params,
    		                                        long location_seed)
    throws CryptoVanishException, VanishBackendException,
           TooManySharesLostVanishException, InvalidArgumentVanishException {
    	VanishBackendInterface[] backends = selectBackends(params);
    	
    	byte[][] master_key_pieces = new byte[backends.length][];
    	if (backends.length == 1) {
            // Generate the locations where to place the objects.
    		byte[][] locations = generateLocations(backends[0],
    				location_seed, params.getNumShares());

    		int share_threshold = params.getThreshold();
    		// Push the shares in those locations.
    		byte[][] shares = getShares(backends[0], locations,
    				                    share_threshold);
    		assert(shares != null && shares.length >= share_threshold);

    		// Found a set of shares; try to reconstruct the decryption key from
    		// them using SSS.
    		master_key_pieces[0] = sss_module.compose(shares,
    				share_threshold);
    	} else {
    		Random random = new Random(location_seed); 
    		for (int i = 0; i < backends.length; ++i) {
    			master_key_pieces[i] = composeForBackend(backends[i],
    					params, random.nextLong());
    		}
    	}
    	return sss_module.compose(master_key_pieces, master_key_pieces.length);
    }
    
    private byte[] composeForBackend(VanishBackendInterface backend,
    		VDOParams params, long location_seed)
    throws VanishBackendException, TooManySharesLostVanishException,
           CryptoVanishException {
    	int num_shares = params.getNumShares();
    	int share_threshold = params.getThreshold();
    	
    	if (backend.getIdentifier().equals("opendht")) {
    		num_shares = 3;
    		share_threshold = 1;
    	}

        // Generate the locations where to place the objects.
		byte[][] locations = generateLocations(backend,
				location_seed, num_shares);
		
		// Push the shares in those locations.
		byte[][] shares = getShares(backend, locations, share_threshold);
		assert(shares != null && shares.length >= share_threshold);

		// Found a set of shares; try to reconstruct the decryption key from
		// them using SSS.
		return sss_module.compose(shares, share_threshold);
    }
    
    protected byte[][] getShares(VanishBackendInterface backend,
    		byte[][] locations, int threshold)
    throws VanishBackendException, TooManySharesLostVanishException {
    	return backend.getShares(locations, threshold);
    }
	
	private byte[][] generateLocations(VanishBackendInterface backend,
			                           long location_seed, int total_keys) {
		// The pseudo-random generator will generate the same sequence of
		// numbers given the same seed.
		Random prg = new Random(location_seed);
		
		// Generate indexes:
		byte[][] indexes = new byte[total_keys][];
		for(int i = 0; i < total_keys; i++) {
			indexes[i] = backend.generateLocation(prg);
		}
		
		return indexes;
	}
	
	// Param validators:
	
	protected void validateParamsEncapsulate(
			InputStream clear_data_is,
            OutputStreamWrapper enc_data_os,
            Metadata metadata)
	throws InvalidArgumentVanishException { 
		if (metadata == null) {
			throw new InvalidArgumentVanishException(
					"Invalid metadata for encapsulate");
		}
		if (clear_data_is == null && enc_data_os != null ||
			clear_data_is != null && enc_data_os == null) {
			throw new InvalidArgumentVanishException(
					"Invalid I/O streams for encapsulate");
		}
	}
	
	protected void validateParamsDecapsulate(
			InputStreamWrapper enc_data_is,
            OutputStream clear_data_os,
            Metadata metadata)
	throws InvalidArgumentVanishException {
		if (metadata == null) {
			throw new InvalidArgumentVanishException(
					"Invalid metadata for decapsulate");
		}
		if (enc_data_is == null && clear_data_os != null ||
				enc_data_is != null && clear_data_os == null) {
			throw new InvalidArgumentVanishException(
					"Invalid I/O streams for decapsulate");
		}
	}

	protected void validateParamsRefreshDeeply(
			InputStreamWrapper enc_data_is,
			OutputStreamWrapper refreshed_enc_data_os,
            Metadata old_metadata,
            Metadata new_metadata)
	throws InvalidArgumentVanishException {
		validateParamsRefreshReadOnly(old_metadata);
		if (enc_data_is == null || refreshed_enc_data_os == null ||
			new_metadata == null) {
			throw new InvalidArgumentVanishException(
					"Invalid input for refreshDeeply");
		}
	}

	protected void validateParamsRefreshReadOnly(Metadata metadata)
	throws InvalidArgumentVanishException {
		if (metadata == null) {
			throw new InvalidArgumentVanishException(
					"Invalid metadata for refreshReadOnly");
		}
	}
	
	protected SymmetricEncDec getCryptoModule() { return crypto_module; }
	
    // Class used for refresh:
	
	// TODO(roxana): Replace this with ThreadPoolExecutor.
	private class RefreshThread<RET> extends Thread {
		private Callable<RET> callable;
		
		// exception != null => result == null.
		public RET result = null;
		public Exception exception = null;
		
		public RefreshThread(Callable<RET> callable) {
			this.callable = callable;
		}
		
		public void run() {
			clear();
			try { result = callable.call(); }
			catch (Exception e) { exception = e; }
		}
		
		private void clear() {
			result = null;
			exception = null;
		}
		
		public RET getResult() throws Exception {
			if (exception != null) throw exception;
			return result;
		}
	}
}
