/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.internal.impl;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;

import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.backend.VanishBackendInterface;
import edu.washington.cs.vanish.internal.crypto.SecretSharingInterface;
import edu.washington.cs.vanish.internal.crypto.SymmetricEncDec;
import edu.washington.cs.vanish.internal.metadata.Metadata;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.internal.metadata.impl.IndirectKeyMetadataImpl;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.util.IOVanishException;

/**
 * 
 * @author roxana
 */
public class IndirectKeyVanishImpl extends BasicVanishImpl {
	public IndirectKeyVanishImpl(
			Map<String, VanishBackendInterface> backend_modules,
			SymmetricEncDec crypto_module,
			SecretSharingInterface sss_module,
			VanishLogger log)
	throws VanishException {
		super(backend_modules, crypto_module, sss_module, log);
	}

	@Override
	protected void pushEncryptionKeyIntoBackend(byte[] data_key,
			                                    Metadata metadata)
	throws VanishException {
		if (! (metadata instanceof IndirectKeyMetadataImpl)) {
			super.pushEncryptionKeyIntoBackend(data_key, metadata);
			return;  // done.
		}
		
		IndirectKeyMetadataImpl object_metadata =
			(IndirectKeyMetadataImpl)metadata; 
		 
		byte[] encryption_key_for_data_key = null;
		ByteArrayInputStream clear_key_is = null;
		ByteArrayOutputStream enc_key_os = null;
		
		try {
			clear_key_is = new ByteArrayInputStream (data_key);
			enc_key_os = new ByteArrayOutputStream();  // data_key.length
		
			VDOParams vdo_params = metadata.getVDOParams();
			
			encryption_key_for_data_key = getCryptoModule().generateKey(
					vdo_params);
			super.pushEncryptionKeyIntoBackend(encryption_key_for_data_key,
					                           object_metadata);
			getCryptoModule().encryptToOutputStream(clear_key_is, enc_key_os,
					encryption_key_for_data_key, vdo_params);
			
            // Update the object's metadata to include the encrypted data key.
			byte[] encrypted_data_key = enc_key_os.toByteArray();
			object_metadata.setEncryptedDataKey(encrypted_data_key);
		} finally {
            // Time to clean up the encryption key for the data key.
			VanishUtil.cleanSensitiveData(encryption_key_for_data_key);
			VanishUtil.cleanSensitiveData(clear_key_is);
			try { enc_key_os.close(); }
			catch (IOException e) { e.printStackTrace(); }
		}
	}

	@Override
	public byte[] getDecryptionKeyFromBackend(Metadata metadata)
	throws VanishException {
		if (! (metadata instanceof IndirectKeyMetadataImpl)) {
			return super.getDecryptionKeyFromBackend(metadata);
		}
		
		IndirectKeyMetadataImpl object_metadata =
			(IndirectKeyMetadataImpl)metadata;

		InputStream enc_data_key_is = null;
		InputStream clear_data_key_is = null;
		byte[] encryption_key_for_data_key = null;
		byte[] data_key = null;
		
		try {
			byte[] encrypted_data_key = object_metadata.getEncryptedDataKey();
			if (encrypted_data_key == null) {
				throw new InvalidArgumentVanishException(
						"Invalid metadata with null encrypted data key");
			}
			enc_data_key_is = new ByteArrayInputStream(encrypted_data_key);
			
			// Get the data key from the backend.
			encryption_key_for_data_key = super.getDecryptionKeyFromBackend(
					object_metadata);
			clear_data_key_is = getCryptoModule().decrypt(enc_data_key_is,
					encryption_key_for_data_key,
					metadata.getVDOParams());
			
			data_key = readKey(clear_data_key_is, encrypted_data_key.length);
		} catch (VanishException e) {
			// In case of any exception, clean the data key. Otherwise don't.
			VanishUtil.cleanSensitiveData(data_key);
			throw e;  // throw the exception further.
		} finally {
			VanishUtil.cleanSensitiveData(clear_data_key_is);
			VanishUtil.cleanSensitiveData(encryption_key_for_data_key);
			if (enc_data_key_is != null) {
				try { enc_data_key_is.close(); }
				catch (IOException e) { e.printStackTrace(); }
			}
		}
		
		return data_key;
	}
	
	// Helper functions:
	
	// TODO(roxana): Ugly code.
	private byte[] readKey(InputStream dec_key_is, int keylgth)
	throws IOVanishException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
        byte[] k = new byte[keylgth];
        try {
        	int n;
        	while ((n = dec_key_is.read(k)) > 0) baos.write(k, 0, n);
		} catch (IOException e) {
			throw new IOVanishException("Failed to read key from IS", e);
		}
		
		return baos.toByteArray();
	}
}
