/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.internal.impl;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.util.Map;

import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.MetadataEncapsulatedVanishException;
import edu.washington.cs.vanish.internal.MetadataEncapsulatedVanishInternalInterface;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.backend.VanishBackendInterface;
import edu.washington.cs.vanish.internal.crypto.SecretSharingInterface;
import edu.washington.cs.vanish.internal.crypto.SymmetricEncDec;
import edu.washington.cs.vanish.internal.metadata.Metadata;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.util.IOVanishException;
import edu.washington.cs.vanish.util.Immutable;

/**
 * @author roxana
 *
 */
public class MetadataEncapsulatedIndirectKeyVanishImpl
extends IndirectKeyVanishImpl
implements MetadataEncapsulatedVanishInternalInterface {
	public MetadataEncapsulatedIndirectKeyVanishImpl(
			Map<String, VanishBackendInterface> backend_module,
			SymmetricEncDec crypto_module,
			SecretSharingInterface sss_module,
			VanishLogger log)
	throws VanishException {
		super(backend_module, crypto_module, sss_module, log);
	}
	
	@Override
	protected void writeEncryptedDataToOutputStream(
			InputStream encrypted_data_is,
			OutputStreamWrapper encrypted_data_os,
			@Immutable Metadata metadata)
	throws VanishException {
		if (encrypted_data_os == null || encrypted_data_os == null) {
			throw new InvalidArgumentVanishException(
					"Invalid null I/O stream for decapsulate");
		}
		try {
			ObjectOutputStream oos = encrypted_data_os.getObjectOutputStream();
			if (oos == null) {
				throw new InvalidArgumentVanishException(
						"Invalid null OutputStreamWrapper.");
			}
			oos.writeObject(metadata);
		} catch (IOException e) {
			throw new IOVanishException("Failed to write metadata", e);
		}
		
		if (metadata == null) {
			throw new VanishException("Invalid metadata");
		}
		super.writeEncryptedDataToOutputStream(encrypted_data_is, 
					                           encrypted_data_os,
					                           metadata);
	}
	
	// TODO(roxana): Problematic if different clients use different of types
	// of Vanish cores.
	@Override
	public void decapsulate(InputStreamWrapper enc_data_is,
                            OutputStream clear_data_os,
                            Metadata metadata)
	throws VanishException {
		this.doDecapsulate(enc_data_is, clear_data_os, metadata);
	}
	
	private Metadata doDecapsulate(InputStreamWrapper enc_data_is,
			                       OutputStream clear_data_os,
			                       Metadata metadata)
	throws MetadataEncapsulatedVanishException, VanishException {
		if (enc_data_is == null || clear_data_os == null) {
			throw new InvalidArgumentVanishException(
					"Invalid null I/O streams for decapsulate");
		}
		Metadata inner_metadata = getMetadataFromInputStream(enc_data_is);
		if (inner_metadata != null) metadata = inner_metadata;
		
		if (metadata == null) {
			throw new InvalidArgumentVanishException(
					"Invalid null metadata for decapsulate");
		}
		try {
			super.decapsulate(enc_data_is, clear_data_os, metadata);
		} catch (VanishException e) {
			throw new MetadataEncapsulatedVanishException(e, metadata);
		}
		
		return metadata;
	} 
	
	private Metadata getMetadataFromInputStream(InputStreamWrapper enc_data_is)
	throws IOVanishException {
		Metadata metadata = null;
		try {
			ObjectInputStream ois = enc_data_is.getObjectInputStream();
			if (ois == null) {
				throw new InvalidArgumentVanishException(
						"Invalid null OutputStreamWrapper.");
			}
			metadata = (Metadata)ois.readObject();
		} catch (Exception e) {
			try {
				enc_data_is.reset();
				metadata = null;
			} catch (IOException e1) {
				throw new IOVanishException(
						"Failed to reset IS after failed read of meta", e1);
			}
		}
		
		return metadata;
	}

	@Override
	public void refreshDeeply(final InputStreamWrapper enc_data_is,
			final OutputStreamWrapper refreshed_enc_data_os,
            @Immutable Metadata old_metadata,
            final Metadata new_metadata)
	throws VanishException {
		this.doRefreshDeeply(enc_data_is, refreshed_enc_data_os,
				             old_metadata, new_metadata);
	}
	
	private Metadata doRefreshDeeply(final InputStreamWrapper enc_data_is,
			final OutputStreamWrapper refreshed_enc_data_os,
            @Immutable Metadata old_metadata,
            final Metadata new_metadata)
	throws MetadataEncapsulatedVanishException, VanishException {
		if (enc_data_is == null || refreshed_enc_data_os == null) {
			throw new InvalidArgumentVanishException(
					"Invalid null I/O streams for decapsulate");
		}
		Metadata inner_metadata = getMetadataFromInputStream(enc_data_is);
		if (inner_metadata != null) old_metadata = inner_metadata;
		
		if (old_metadata == null) {
			throw new InvalidArgumentVanishException(
					"Invalid null metadata for decapsulate");
		}
		
		try {
			super.refreshDeeply(enc_data_is, refreshed_enc_data_os,
					            old_metadata, new_metadata);
		} catch(VanishException e) {
			throw new MetadataEncapsulatedVanishException(e, old_metadata);
		}
		return old_metadata;
	}
	
	// MetadataEncapsulatedVanish interface:

	// @Override
	public Metadata decapsulate(InputStreamWrapper enc_data_is,
			                    OutputStream clear_data_os)
	throws MetadataEncapsulatedVanishException, VanishException {
		return this.doDecapsulate(enc_data_is, clear_data_os, null);
	}

	// @Override
	public Metadata refreshDeeply(InputStreamWrapper enc_data_is,
			                      OutputStreamWrapper refreshed_enc_data_os,
			                      Metadata new_metadata)
	throws MetadataEncapsulatedVanishException, VanishException {
		return this.doRefreshDeeply(enc_data_is, refreshed_enc_data_os, null,
				                    new_metadata);
	}
}
