/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.internal.metadata;

import java.io.Serializable;
import java.util.Map;

import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.util.Defaults;

/**
 * Basic class for VDO creation parameters.
 * 
 * Subclasses can add new parameters, which must be understood by clients and
 * servers.
 * 
 * @author roxana
 *
 */
public class VDOParams extends Mappable implements Serializable {
	private static final long serialVersionUID = -4774421669366587764L;
	
	@SuppressWarnings("deprecation")
	public static final int kDefaultNumShares = Defaults.CONF.getInt(
			Defaults.CONF_VANISH_VDO_DEFAULTNUMSHARES,
			Defaults.CONF_VANISH_VDO_DEFAULTNUMSHARES_DEFAULT);
	@SuppressWarnings("deprecation")
	public static final int kDefaultTheshold = Defaults.CONF.getInt(
			Defaults.CONF_VANISH_VDO_DEFAULTTHRESHOLD,
			Defaults.CONF_VANISH_VDO_DEFAULTTHRESHOLD_DEFAULT);
	@SuppressWarnings("deprecation")
	public static final int kDefaultTimeout_h = Defaults.CONF.getInt(
			Defaults.CONF_VANISH_VDO_DEFAULTTIMEOUT_H,
			Defaults.CONF_VANISH_VDO_DEFAULTTIMEOUT_H_DEFAULT);
	
	@SuppressWarnings("deprecation")
	public static final String[] kDefaultNetworks =
		Defaults.CONF.getStringArray(
			Defaults.CONF_VANISH_VDO_DEFAULTNETWORKS,
			Defaults.CONF_VANISH_VDO_DEFAULTNETWORKS_DEFAULT);
	
	@SuppressWarnings("deprecation")
	public static final String kEncryptionAlgo =
		Defaults.CONF.getString(
			Defaults.CONF_VANISH_CRYPTO_ENCALGORITHM,
			Defaults.CONF_VANISH_CRYPTO_ENCALGORITHM_DEFAULT);
	@SuppressWarnings("deprecation")
	public static final int kEncryptionKeyLength =
		Defaults.CONF.getInt(
			Defaults.CONF_VANISH_CRYPTO_ENCKEYLENGTH_BITS,
			Defaults.CONF_VANISH_CRYPTO_ENCKEYLENGTH_BITS_DEFAULT);	
	@SuppressWarnings("deprecation")
	public static final String kEncryptionMode =
		Defaults.CONF.getString(
			Defaults.CONF_VANISH_CRYPTO_ENCMODE,
			Defaults.CONF_VANISH_CRYPTO_ENCMODE_DEFAULT);
	
	@Deprecated
	public static final VDOParams kDefaultParameters = newDefaultVDOParams();
	
	/** The number of shares needed to restore the key. */
	private int  num_shares;
	
	/** The total number of shares to break the key into. */
	private int  threshold;
	
	/** The desired timeout in hours. */
	private int  timeout_h;
	
	/** Encryption algorithm */
	private String encryption_algorithm;
	
	/** Encryption key length */
	private int encryption_key_length;
	
	/** Encryption mode. */
	private String encryption_mode;
	
	/** Creation timestamp. **/
	private long creation_ts = 0;
	
	/** Stats elements. */
	private long vdo_id = 0;
	
	private String[] networks;
	
	public VDOParams() throws InvalidArgumentVanishException {
		this(kDefaultNumShares, kDefaultTheshold, kDefaultTimeout_h,
			 kEncryptionAlgo, kEncryptionKeyLength, kEncryptionMode,
			 kDefaultNetworks);
	}
	
	public VDOParams(Map<String, Object> map)
	throws InvalidArgumentVanishException {
		this();
		long creation_ts = this.creation_ts; // TODO(roxana): Hacky! Make 
				// setFromMap check for "no-override" annotation (Amit's idea).
		this.setFromMap(map);
		this.creation_ts = creation_ts;
		validate();
	}
	
	public VDOParams(int num_shares, int threshold, int timeout_h)
	throws InvalidArgumentVanishException {
		this(num_shares, threshold, timeout_h,
			 kEncryptionAlgo, kEncryptionKeyLength, kEncryptionMode,
			 kDefaultNetworks);
	}
	
	public VDOParams(int num_shares, int threshold, int timeout_h,
			         String encryption_algorithm, int encryption_key_length,
			         String encryption_mode, String[] networks)
	throws InvalidArgumentVanishException {
		setNumShares(num_shares);
		setThreshold(threshold);
		setTimeoutInHours(timeout_h);
		
		this.encryption_algorithm = encryption_algorithm;
		this.encryption_key_length = encryption_key_length;
		this.encryption_mode = encryption_mode;
		
		this.networks = networks;

		validate();

		this.creation_ts = System.currentTimeMillis();
	}
	
	/**
	 * Sets all non-fixed params (i.e., those that do not change the behavior
	 * of a VDO). 
	 * This is used by the prepusher to inherit some non-fixed params when it
	 * encapsulates using a prepushed metadata.
	 * @param other
	 */
	public void inheritNonFixedParamsFromOtherVDOParams(VDOParams other) {
		if (other == null) return;  // nothing to do.
		//this.creation_ts = other.creation_ts;
		this.vdo_id = other.vdo_id;
	}
	
	public void validate() throws InvalidArgumentVanishException {
		if (0 >= num_shares || num_shares <= threshold) {
			throw new InvalidArgumentVanishException(
					"The number of shares must be greater than the threshold");
		}
		if (threshold <= 0) {
			throw new InvalidArgumentVanishException(
					"The threshold must be positive " + threshold);
		}
		if (timeout_h <= 0) {
			throw new InvalidArgumentVanishException(
					"The timeout must be positive");
		}
		
		if (networks == null || networks.length == 0) {
			throw new InvalidArgumentVanishException(
					"Backend identifiers must not be null");
		}
		
		// TODO(roxana): Validate other parameters?
	}
	
	// Accessors:
	
	public int getNumShares() { return num_shares; }
	public void setNumShares(int num_shares) {this.num_shares = num_shares; }
	
	public int getThreshold() { return threshold; }
	public void setThreshold(int threshold) { this.threshold = threshold; }
	
	public int getTimeout() { return timeout_h; }
	public void setTimeoutInHours(int timeout_h) {
		this.timeout_h = timeout_h;
	}
	
	public String getEncryptionAlgorithm() { return encryption_algorithm; }
	public int getEncryptionKeyLengthBits() { return encryption_key_length; }
	public String getEncryptionMode() { return encryption_mode; }
	
	public String[] getBackendIdentifiers() { return networks; }
	
	// Used by stats.
	public long getVDOId() { return vdo_id; }
	public void setVDOId(long id) { this.vdo_id = id; }

	public long getCreationDate() { return creation_ts; }
	
	public static boolean isDefaultVDOParams(VDOParams params) {
		return params != null && params.equals(kDefaultParameters);
	} 
	
	// Object methods:
	
	@Override
	public boolean equals(Object other) {
		if (other == null) return false;
		if (other == this) return true;
		try {
			VDOParams o = (VDOParams)other;
			return (num_shares == o.num_shares && threshold == o.threshold &&
			        timeout_h == o.timeout_h &&
			        encryption_algorithm.equals(o.encryption_algorithm) &&
			        encryption_key_length == o.encryption_key_length &&
			        encryption_mode.equals(o.encryption_mode));
		} catch (ClassCastException e) {
			return false;
		}
	}

	@Override
	public int hashCode() {
		int hash = 7;
		hash = 31 * hash + num_shares;
		hash = 31 * hash + threshold;
		hash = 31 * hash + timeout_h;
		hash = 31 * hash + encryption_algorithm.hashCode();
		hash = 31 * hash + encryption_key_length;
		hash = 31 * hash + encryption_mode.hashCode();
		return hash;
	}
	
	@Override
	public String toString() {
		return "N=" + num_shares + ";M=" + threshold + "T=" + timeout_h;
	}
	
	public static VDOParams newDefaultVDOParams() {
		VDOParams params = null;
		try {
			params = new VDOParams();
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace(System.err);
		}
		
		return params;
	}
}
