/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.internal.metadata.impl;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.metadata.Metadata;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.util.Constants;
import edu.washington.cs.vanish.util.Defaults;
import edu.washington.cs.vanish.util.Pair;
/**
 * 
 * @author roxana
 *
 */
public class EpochAwareMetadataImpl implements Metadata {
	private static final long serialVersionUID = 5580672788044873115L;

	@SuppressWarnings("deprecation")
	private static final long kDefaultRefreshPeriod =
		Defaults.CONF.getTimeInMillis(
    		    Defaults.CONF_VANISH_SYSTEM_REFRESH_PERIOD_H,
    			Defaults.CONF_VANISH_SYSTEM_REFRESH_PERIOD_H_DEFAULT,
    			VanishConfiguration.TimeUnit.HOURS);
	
	private final Metadata metadata;
	private final long epoch_length;

	public EpochAwareMetadataImpl(Metadata metadata, long epoch_length) {
		if (metadata == null) {
			throw new NullPointerException("Metadata cannot be null"); 
		}
		if (metadata instanceof EpochAwareMetadataImpl) {  // Copy meta.
			EpochAwareMetadataImpl other = (EpochAwareMetadataImpl)metadata;
			this.metadata = other.metadata;
			this.epoch_length = other.epoch_length;
		} else {  // Encapsulate meta.
			this.metadata = metadata;
			this.epoch_length = epoch_length;
		}
	}

	// Metadata interface:
	
	// @Override
	public VDOParams getVDOParams() {
		return metadata.getVDOParams();
	}

	// @Override
	public long[] getAllPossibleLocationSeeds() {
		Pair<Long, Boolean> p = this.getTimeSeed();
		long time_seed = p.o1;
		boolean is_time_seed_fixed = p.o2;
		if (is_time_seed_fixed) {  // the time seed is fixed for this vdo.
			// Return the single possible location where it might be.
			return new long[] { locationFunction(time_seed) };
		} else {  // the time seed varies with real time for this vdo.
			// Try out multiple locations, to cover loose clocks.
			return new long[] { locationFunction(time_seed),
					            locationFunction(time_seed - 1),
				                locationFunction(time_seed + 1),
				                locationFunction(time_seed - 2),
				                locationFunction(time_seed + 2) };
		}
	}

	// @Override
	public long getLocationSeed() { 
		return locationFunction(getTimeSeed().o1);
	}

	// TODO(roxana): Very, very, very hacky way of speeding up the normal
	// case (no-refresh one).
	//
	// NOTE: We assume (as the entire Vanish does) that everyone who gets this
	// VDO will adhere to the timeout set in this VDO and not change it or
	// refresh further.
	private Pair<Long, Boolean> getTimeSeed() {
		long time_seed;
		boolean is_time_seed_fixed;
		
		VDOParams params = metadata.getVDOParams();
		if (refreshesExpected(params.getTimeout(), epoch_length)) {
			long vdo_approximate_age_in_millis = System.currentTimeMillis() -
				metadata.getVDOParams().getCreationDate();
			long vdo_age_in_epochs = vdo_approximate_age_in_millis / 
		        (epoch_length > 0  // introduced for backward compatibility
		    		? epoch_length : kDefaultRefreshPeriod);
			
			time_seed = vdo_age_in_epochs;
			is_time_seed_fixed = false;
		} else {  // No refresh expected, which is the normal case.
			// Speed up the normal case by fixing its time seed to the
			// VDO creation time, which is fixed inside the VDO.
			time_seed = params.getCreationDate();
			is_time_seed_fixed = true;
		}
		return new Pair<Long, Boolean>(time_seed, is_time_seed_fixed);
	}
	
	private boolean refreshesExpected(long timeout_h, long epoch_length_ms) {
		return timeout_h * Constants.HOURS > epoch_length_ms;
	}
	
	public long getEpochLength() { return epoch_length; }

	private long locationFunction(long epoch_number) {  // XOR
        return metadata.getLocationSeed() ^ epoch_number;
	}
	
	// Object functions:
	
	@Override
	public String toString() { return metadata.toString(); }
	
	@Override 
	public boolean equals(Object o) {
		if (o == null || (! (o instanceof EpochAwareMetadataImpl))) {
			return false;
		}
		return metadata.equals(o);
	}
	
	@Override
	public int hashCode() { return metadata.hashCode(); }
}
