/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.service.impl;

import java.io.OutputStream;
import java.lang.reflect.Constructor;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.MetadataEncapsulatedVanishInternalInterface;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.backend.VanishBackendInterface;
import edu.washington.cs.vanish.internal.crypto.SecretSharingInterface;
import edu.washington.cs.vanish.internal.crypto.SymmetricEncDec;
import edu.washington.cs.vanish.internal.crypto.impl.ShamirSecretSharing;
import edu.washington.cs.vanish.internal.crypto.impl.SymmetricEncDecImpl;
import edu.washington.cs.vanish.internal.impl.MetadataEncapsulatedIndirectKeyVanishImpl;
import edu.washington.cs.vanish.internal.metadata.Metadata;
import edu.washington.cs.vanish.internal.metadata.MetadataFactory;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.internal.metadata.impl.EpochAwareMetadataImpl;
import edu.washington.cs.vanish.internal.metadata.impl.IndirectKeyMetadataImpl;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.system.VanishFactory;
import edu.washington.cs.vanish.system.VanishInterface;
import edu.washington.cs.vanish.system.VanishRefreshTask;
import edu.washington.cs.vanish.system.VanishRefreshTaskFactory;
import edu.washington.cs.vanish.system.impl.ReadOnlyVanishRefreshTask;
import edu.washington.cs.vanish.system.impl.VanishImpl;
import edu.washington.cs.vanish.util.Defaults;
import edu.washington.cs.vanish.util.VanishTime;

/**
 * 
 * @author roxana
 */
public class DefaultVanishFactory implements VanishFactory {
	private final VanishConfiguration CONF;
	
	/** Mandatory constructor. */
	public DefaultVanishFactory(VanishConfiguration conf) {
		this.CONF = conf;
	}
	
	public VanishInterface getInitializedVanishSystem()
	throws VanishException {
		SecureRandom random;
		try {
			final String kRandomGeneratorAlg = CONF.getString(
					Defaults.CONF_VANISH_CRYPTO_RANDOMGENERATOR,
					Defaults.CONF_VANISH_CRYPTO_RANDOMGENERATOR_DEFAULT);
			random = SecureRandom.getInstance(kRandomGeneratorAlg);
		} catch (NoSuchAlgorithmException e) {
			throw new InvalidArgumentVanishException(
					"Failed to create secure random", e);
		}
		VanishTime time = new VanishTime();
		
		VanishInterface vanish = instantiateVanish(
				instantiateVanishInternal(instantiateBackend(time, CONF),
						instantiateCryptoModule(random, CONF),
						instantiateSecretSharingModule(random, CONF),
						CONF),
				instantiateMetadataFactory(CONF),
				instantiateServerVanishRefreshTaskFactory(CONF),
				random,
				time,
				CONF);

		vanish.init();

		return vanish;
	}
	
	protected VanishInterface instantiateVanish(
			MetadataEncapsulatedVanishInternalInterface vanish_internal,
			MetadataFactory metadata_factory,
			VanishRefreshTaskFactory refresh_task_factory,
			Random random,
			VanishTime time,
			VanishConfiguration conf)
	throws VanishException {
		return new VanishImpl(vanish_internal, metadata_factory,
				              refresh_task_factory,
				              random,
				              time,
				              conf,
				              VanishLogger.getLogger("vanish.system"));
	}
		
	protected MetadataEncapsulatedVanishInternalInterface
	instantiateVanishInternal(
			Map<String, VanishBackendInterface> vanish_backend,
			SymmetricEncDec encryption_module,
			SecretSharingInterface sss_module,
			@SuppressWarnings("unused") VanishConfiguration conf)
	throws VanishException {
		return new MetadataEncapsulatedIndirectKeyVanishImpl(
				vanish_backend, encryption_module, sss_module,
				VanishLogger.getLogger("vanish.core"));
	}
	
	protected Map<String, VanishBackendInterface> instantiateBackend(
			VanishTime time, VanishConfiguration conf)
	throws VanishException {
		String[] backend_id_to_classnames =
			conf.getStringArray(Defaults.CONF_VANISH_BACKEND_IMPL_CLASS,
					Defaults.CONF_VANISH_BACKEND_IMPL_CLASS_DEFAULT);
		
		Map<String, VanishBackendInterface> backends =
			new HashMap<String, VanishBackendInterface>();
		for (String id_to_classname: backend_id_to_classnames) {
			String[] parts = id_to_classname.split(" *: *");
			if (parts.length != 2) {
				throw new InvalidArgumentVanishException(
						"Invalid vanish backend specification " +
						id_to_classname);
			}
			String backend_id = parts[0];
			String backend_classname = parts[1];
			try {
				Class<? extends VanishBackendInterface> backend_class =
					Class.forName(backend_classname)
					.asSubclass(VanishBackendInterface.class);
				Constructor<? extends VanishBackendInterface> constructor =
					backend_class.getConstructor(VanishConfiguration.class,
							VanishLogger.class);
				
				VanishBackendInterface backend = constructor.newInstance(conf,
						VanishLogger.getLogger("vanish.backend"));
				
				backends.put(backend_id, backend);
			} catch (Exception e) {
				throw new InvalidArgumentVanishException(
						"Invalid vanish backend " + backend_classname);
			}
		}
		
		return backends;
	}
	
	protected SymmetricEncDec instantiateCryptoModule(SecureRandom random,
			VanishConfiguration conf)
	throws InvalidArgumentVanishException {
		return new SymmetricEncDecImpl(random, conf,
				VanishLogger.getLogger("vanish.crypto"));
	}
	
	protected SecretSharingInterface instantiateSecretSharingModule(
			SecureRandom random,
			VanishConfiguration conf)
	throws InvalidArgumentVanishException {
		return new ShamirSecretSharing(random, conf,
				VanishLogger.getLogger("vanish.sss"));
	}
	
	protected MetadataFactory instantiateMetadataFactory(
			final VanishConfiguration conf) {
		final long kRefreshInterval = conf.getTimeInMillis(
				Defaults.CONF_VANISH_SYSTEM_REFRESH_PERIOD_H,
				Defaults.CONF_VANISH_SYSTEM_REFRESH_PERIOD_H_DEFAULT,
				VanishConfiguration.TimeUnit.HOURS);

		return new MetadataFactory() {
			public Metadata createMetadata(VDOParams params, long seed) {
				return new EpochAwareMetadataImpl(new IndirectKeyMetadataImpl(
						params, seed), kRefreshInterval);
			}
		};
	}
	
	protected VanishRefreshTaskFactory
	instantiateServerVanishRefreshTaskFactory(final VanishConfiguration conf) {
		return new VanishRefreshTaskFactory() {
			public VanishRefreshTask createRefreshTask(
					VanishInterface vanish_system_ptr,
					@SuppressWarnings("unused") OutputStream encrypted_data_os,
					Metadata metadata,
					long creation_time,
					long timeout_date,
					long refresh_interval) {
				return new ReadOnlyVanishRefreshTask(vanish_system_ptr,
						metadata,
						creation_time, timeout_date, refresh_interval);
			}
		};
	}
}
