/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.service.impl;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Map;
import java.util.Random;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.MetadataEncapsulatedVanishException;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.impl.InputStreamWrapper;
import edu.washington.cs.vanish.internal.impl.OutputStreamWrapper;
import edu.washington.cs.vanish.internal.metadata.Metadata;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.logging.LogStat;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.service.LocalVanishServiceInterface;
import edu.washington.cs.vanish.service.RemoteVanishServiceInterface;
import edu.washington.cs.vanish.stats.StatsCollector;
import edu.washington.cs.vanish.stats.StatsCollector.ReportLevel;
import edu.washington.cs.vanish.stats.impl.MultiValueStat;
import edu.washington.cs.vanish.system.VanishFactory;
import edu.washington.cs.vanish.system.VanishInterface;
import edu.washington.cs.vanish.util.Constants;
import edu.washington.cs.vanish.util.Defaults;
import edu.washington.cs.vanish.util.IOVanishException;

/**
 * 
 * @author roxana
 */
public class VanishServiceHandler
implements LocalVanishServiceInterface, RemoteVanishServiceInterface,
           Constants {
	@SuppressWarnings("unused")
	protected static VanishConfiguration CONF;
	protected static VanishLogger LOG;
	
	private static StatsCollector stats_collector;
	private static VanishFactory vanish_factory;
	private static Random vdo_id_generator;

	private static long kMaxFileLength;
	private static boolean kShouldZip;
	
	// Stats:
	
	public VanishServiceHandler() { }

	// TODO(roxana): Wrap the above static vars into another object, 
	// s.t. we have a simpler init function, which just sets the object.
	
	public static void init(VanishFactory _vanish_factory,
			                StatsCollector _stats_collector,
			                VanishConfiguration _conf,
			                VanishLogger _log)
	throws VanishException {
		checkNotInitialized();
		vanish_factory = _vanish_factory;
		stats_collector = _stats_collector;
		CONF = _conf;
		LOG = _log;
		vdo_id_generator = new Random();
		
		kMaxFileLength = CONF.getLong(
				Defaults.CONF_VANISH_SERVICE_REMOTE_MAXFILESIZE_KB,
				Defaults.CONF_VANISH_SERVICE_REMOTE_MAXFILESIZE_KB_DEFAULT)
				* KB;
		kShouldZip = CONF.getBoolean(
				Defaults.CONF_VANISH_SERVICE_SHOULD_GZIP,
				Defaults.CONF_VANISH_SERVICE_SHOULD_GZIP_DEFAULT);
	}

	public static void stop() throws VanishException {
		checkInitialized();
	}

	// LocalVanishService interface:

	// @Override
	public int decapsulate(Map<String, Object> caller_params,
			               String input, String output)
	throws VanishException {
		DecapsulateOperation op = new DecapsulateOperation(input, output);
		doWork(op, caller_params);
		return 0;  // only for compatibility w/ Apache's RPC server.
	}

	// @Override
	public int encapsulate(Map<String, Object> caller_params,
			               String input, String output,
			               Map<String, Object> vdo_params)
	throws VanishException {
		EncapsulateOperation op = new EncapsulateOperation(input, output,
				new VDOParams(vdo_params));
		doWork(op, caller_params);
		return 0;  // only for compatibility w/ Apache's RPC server.
	}

	// @Override
	public int refreshDeeply(Map<String, Object> caller_params,
			                 String input, String output,
			                 Map<String, Object> vdo_params)
	throws VanishException {
		RefreshDeeplyOperation op = new RefreshDeeplyOperation(input, output,
				new VDOParams(vdo_params));
		doWork(op, caller_params);
		return 0; // only for compatibility w/ Apache's RPC server.
	}
	
	// @Override
	public int refreshReadOnly(Map<String, Object> caller_params, String input)
	throws VanishException {
		RefreshReadOnlyOperation op = new RefreshReadOnlyOperation(input);
		doWork(op, caller_params);		
		return 0; // only for compatibility w/ Apache's RPC server.
	}

	// RemoteVanishService interface:

	// @Override
	public byte[] decapsulate(Map<String, Object> caller_params, byte[] input)
	throws VanishException {
		DecapsulateOperation op = new DecapsulateOperation(input);
		doWork(op, caller_params);	
		return op.output_data; // must have saved its output.
	}

	// @Override
	public byte[] encapsulate(Map<String, Object> caller_params, byte[] input,
			                  Map<String, Object> vdo_params)
	throws VanishException {
		EncapsulateOperation op = new EncapsulateOperation(input,
				new VDOParams(vdo_params));
		doWork(op, caller_params);
		return op.output_data;
	}

	// @Override
	public byte[] refreshDeeply(Map<String, Object> caller_params,
			                    byte[] input,
			                    Map<String, Object> vdo_params)
	throws VanishException {
		RefreshDeeplyOperation op = new RefreshDeeplyOperation(input, 
				new VDOParams(vdo_params));
		doWork(op, caller_params);
		return op.output_data;
	}

	// @Override
	public int refreshReadOnly(Map<String, Object> caller_params, byte[] input)
	throws VanishException {
		RefreshReadOnlyOperation op = new RefreshReadOnlyOperation(input);
		doWork(op, caller_params);
		return 0; // only for compatibility w/ Apache's RPC server.
	}

	// Helper functions:

	private static void checkNotInitialized()
	throws IllegalServerStateVanishException {
		try {
			checkInitialized();
			throw new IllegalServerStateVanishException(
					"VanishServiceHandler is initialized.");
		} catch (VanishException e) { }  // Exception expected. 
	}

	private static void checkInitialized()
	throws IllegalServerStateVanishException {
		if (vanish_factory == null) {
			throw new IllegalServerStateVanishException(
					"VanishServiceHandler is not initialized.");
		}
	}

	private void doWork(OperationCallback callback, 
			            Map<String, Object> caller_params)
	throws VanishException {
		final String op_name = callback.getOperationName();
		final LogStat logstat = LOG.logStat(
				"VanishServiceHandler." + op_name);

		VanishException exception = null;
		try { callback.performOperation(); }
		catch(VanishException e) { exception = e; }
		
		long time_op_took_ms = logstat.end(exception);
		stats_collector.report(new Stats(op_name, caller_params,
				                         callback.getVDOMetadata(),
				                         time_op_took_ms,
				                         exception)
		                       /*, exception */);
		if (exception != null) throw exception;
	}

	// OperationCallback callback class:
	
	private static final int STREAM_IS_NOT_GZIPPED = 0;
	private static final int STREAM_IS_GZIPPED = 1;

	private abstract class OperationCallback {
		String input_file;
		String output_file;
		
		byte[] input_data;
		byte[] output_data;

		InputStreamWrapper is;  // cannot be null.
		OutputStreamWrapper os;  // can be null.
		
		// Optional:
		VDOParams params;
		
		// Results:
		protected Metadata resulted_metadata;

		public OperationCallback(String input_file, String output_file,
                                 VDOParams params)
		throws InvalidArgumentVanishException {
			this.input_file = input_file;
			this.output_file = output_file;
			set(params);
		}

		public OperationCallback(byte[] input_data, VDOParams params)
		throws InvalidArgumentVanishException {
			if (input_data != null &&
				input_data.length > kMaxFileLength) {
				throw new InvalidArgumentVanishException(
						"Remote file max size exceeded");
			}
			this.input_data = input_data;
			set(params);
		}

		private void set(VDOParams params) {
			this.params = params;
		}
		
		private void checkParams() throws InvalidArgumentVanishException {
			if (params != null) params.validate();
			if (this.is == null) {
				if (this.os != null) {
					try { os.close(); }
					catch (IOException e) { }
				}
				throw new InvalidArgumentVanishException(
						"Invalid null input stream");
			}
		}

		public void performOperation() throws VanishException {			
			checkInitialized();

			openStreams();
			checkParams();
			
			VanishInterface vanish_system =
				vanish_factory.getInitializedVanishSystem();
			
			try { performSpecificOperation(vanish_system); }
			finally { closeStreams(); }
			
            // Save the result.
			if (os != null &&
				os.getBaseStream() instanceof ByteArrayOutputStream) {
					output_data = ((ByteArrayOutputStream)os
							.getBaseStream()).toByteArray();
			}
		}

		protected abstract void performSpecificOperation(
				VanishInterface vanish_system)
		throws VanishException;
		
		protected abstract String getOperationName();
		protected Metadata getVDOMetadata() { return resulted_metadata; }

		private void openStreams() throws IOVanishException {
			InputStream raw_is = null;
			OutputStream raw_os = null;
			if (input_file != null || output_file != null) {
				try {
					if (input_file != null) {
						raw_is = new FileInputStream(input_file);
					}
					if (output_file != null) {
						raw_os = new FileOutputStream(output_file);
					}
				} catch (FileNotFoundException e) {
					throw new IOVanishException("Failed to openStreams", e);
				}
			} else if (input_data != null) {
				raw_is = new ByteArrayInputStream(input_data);
				raw_os = new ByteArrayOutputStream();
			}

			if (raw_is != null) {
				this.is = new InputStreamWrapper(raw_is);
				wrap(this.is);
			}
			if (raw_os != null) {
				this.os = new OutputStreamWrapper(raw_os);
				wrap(this.os);
			}
		}
		
		private void closeStreams() {
			try {
				if (is != null) is.close();
				if (os != null) os.close();
			} catch (IOException e) { }
		}
		
		protected void wrap(InputStreamWrapper is) throws IOVanishException {
			if (is == null) return;
			try {
				int b = is.read();
				if (b == STREAM_IS_GZIPPED) {
					is.addStreamForClosing(new GZIPInputStream(
							is.getTopStream()));
				}
			} catch (IOException e) {
				throw new IOVanishException("Failed to gunzip");
			}
		}
		
		protected void wrap(OutputStreamWrapper os) throws IOVanishException {
			if (os == null) return;
			try {
				if (kShouldZip) {
					os.write(STREAM_IS_GZIPPED);
					os.addStreamForClosing(new GZIPOutputStream(
							os.getTopStream()));
				} else {
					os.write(STREAM_IS_NOT_GZIPPED);
				}
			} catch (IOException e) {
				throw new IOVanishException("Failed to gzip", e);
			}
		}
	}

	// Callback implementations:
	
	private class EncapsulateOperation extends OperationCallback {		
		public EncapsulateOperation(byte[] input, VDOParams params)
		throws VanishException {
			super(input, params);
			prepareParams(params);
		}

		public EncapsulateOperation(String input_file, String output_file,
				VDOParams params)
		throws VanishException {
			super(input_file, output_file, params);
			prepareParams(params);
		}
		
		private void prepareParams(VDOParams params) {
			if (stats_collector.getCurrentReportLevel().compareTo(
				ReportLevel.SENSITIVE) >= 0) {
				LOG.info("Generating random ID for VDO");
				params.setVDOId(vdo_id_generator.nextLong());
			}
		}

		@Override
		public void performSpecificOperation(VanishInterface vanish_system)
		throws VanishException {
			// Perform the operation:
			resulted_metadata = vanish_system.encapsulate(is, os, params);
		}
		
		@Override
		protected void wrap(InputStreamWrapper is) throws IOVanishException {
			// Don't wrap clear-text stream.
		}

		@Override
		protected String getOperationName() { return "encapsulate"; }
	}

	private class DecapsulateOperation extends OperationCallback {		
		public DecapsulateOperation(byte[] input) throws VanishException {
			super(input, null);
		}

		public DecapsulateOperation(String input_file, String output_file)
		throws VanishException {
			super(input_file, output_file, null);
		}

		@Override
		public void performSpecificOperation(VanishInterface vanish_system)
		throws VanishException {
			try {
				resulted_metadata = vanish_system.decapsulate(is, os);
			} catch (MetadataEncapsulatedVanishException e) {
				resulted_metadata = e.getMetadata();
				throw (VanishException)e.getCause();
			}
		}
		
		@Override
		protected void wrap(OutputStreamWrapper is) throws IOVanishException {
			// Don't wrap clear-text stream.
		}

		@Override
		protected String getOperationName() { return "decapsulate"; }
	}

	private class RefreshDeeplyOperation extends OperationCallback {
		public RefreshDeeplyOperation(byte[] input, VDOParams params)
		throws VanishException {
			super(input, params);
		}

		public RefreshDeeplyOperation(String input_file, String output_file,
                                      VDOParams params)
		throws VanishException {
			super(input_file, output_file, params);
		}

		@Override
		public void performSpecificOperation(VanishInterface vanish_system)
		throws VanishException {
			try {
				resulted_metadata = vanish_system.refreshDeeply(is, os,
						params);
			} catch (MetadataEncapsulatedVanishException e) {
				resulted_metadata = e.getMetadata();
				throw (VanishException)e.getCause();
			}
		}

		@Override
		protected String getOperationName() { return "refreshDeeply"; }
	}

	private class RefreshReadOnlyOperation extends OperationCallback {
		public RefreshReadOnlyOperation(byte[] input) throws VanishException {
			super(input, null);
		}

		public RefreshReadOnlyOperation(String input_file)
		throws VanishException {
			super(input_file, null, null);
		}

		@Override
		public void performSpecificOperation(VanishInterface vanish_system)
		throws VanishException {
			if (is == null) {
				throw new InvalidArgumentVanishException(
						"Invalid input stream for refresh");
			}

			try {
				resulted_metadata = (Metadata)is.getObjectInputStream()
				                    .readObject();
			} catch (Exception e) {
				throw new IOVanishException("Failed to read metadata");
			}

			vanish_system.refreshReadOnly(resulted_metadata);
		}

		@Override
		protected String getOperationName() { return "refreshRO"; }
	}
}

class Stats extends MultiValueStat {
	protected Stats(String operation,
			        Map<String, Object> caller_params, Metadata vdo_params,
			        long runtime,
			        Exception error_while_running) {
		super();		
		addStatPiece(ReportLevel.INSENSITIVE, "op", operation);
		loadCallerParams(caller_params);
		loadVDOParams(vdo_params);
		addStatPiece(ReportLevel.INSENSITIVE, "op_time_ms", runtime);
		if (error_while_running != null) {
			addStatPiece(ReportLevel.INSENSITIVE, "op_error_msg",
					     error_while_running.getMessage());
		}
	}

	protected void loadCallerParams(Map<String, Object> caller_params) {
		if (caller_params == null) return;  // nothing to do.
		Object vanish_id = caller_params.get("vid");
		if (vanish_id != null) {
			addStatPiece(ReportLevel.INSENSITIVE, "vid", vanish_id);
		}
		Object referrer = caller_params.get("ref");
		if (referrer != null) {
			addStatPiece(ReportLevel.SENSITIVE, "op_referrer", referrer);
		}
	}
	
	protected void loadVDOParams(Metadata metadata) {
		if (metadata == null) return;  // nothing to do.
		this.loadVDOParams(metadata.getVDOParams());
	}

	protected void loadVDOParams(VDOParams params) {
		if (params == null) return;  // nothing to do.
		long vdo_id = params.getVDOId();
		if (vdo_id != 0) addStatPiece(ReportLevel.SENSITIVE, "vdo_id", vdo_id);
		int age = (int)
			((System.currentTimeMillis() - params.getCreationDate()) /
			 Constants.MINUTES);
		addStatPiece(ReportLevel.INSENSITIVE, "vdo_age_min", age);
		addStatPiece(ReportLevel.INSENSITIVE, "vdo_shares",
				     params.getNumShares());
		addStatPiece(ReportLevel.INSENSITIVE, "vdo_threshold",
				     params.getThreshold());
		addStatPiece(ReportLevel.INSENSITIVE, "vdo_timeout_h",
				     params.getTimeout());
	}
}

