/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.service.impl;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.InetAddress;
import java.net.UnknownHostException;

import org.apache.xmlrpc.XmlRpcException;
import org.apache.xmlrpc.server.PropertyHandlerMapping;
import org.apache.xmlrpc.server.XmlRpcServer;
import org.apache.xmlrpc.server.XmlRpcServerConfigImpl;
import org.apache.xmlrpc.webserver.WebServer;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.logging.LogStat;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.service.LocalVanishServiceInterface;
import edu.washington.cs.vanish.service.RemoteVanishServiceInterface;
import edu.washington.cs.vanish.service.VanishServiceInterface;
import edu.washington.cs.vanish.stats.StatsCollector;
import edu.washington.cs.vanish.stats.impl.StatUploaderImpl;
import edu.washington.cs.vanish.stats.impl.StatsCollectorImpl;
import edu.washington.cs.vanish.system.VanishFactory;
import edu.washington.cs.vanish.system.VanishInterface;
import edu.washington.cs.vanish.util.Defaults;
import edu.washington.cs.vanish.util.RefreshableConfig;
import edu.washington.cs.vanish.util.RefreshableParams;
import edu.washington.cs.vanish.util.Initializable;
import edu.washington.cs.vanish.util.VanishTime;

/**
 * 
 * @author roxana
 */
class RPCServerParams implements RefreshableParams {
	final int kNumRPCWorkerThreads;
	
	public RPCServerParams(int numRPCWorkerThreads) {
		this.kNumRPCWorkerThreads = numRPCWorkerThreads;
	} 
}

public class VanishXMLRPCServer
implements Initializable, RefreshableConfig<RPCServerParams> {	
	private final VanishLogger LOG;
	private final VanishConfiguration CONF;
	
	// Non-refreshable params:
	private final int kVanishRPCServerPort;
	private final boolean kVanishRPCServerBindToLocalHost;
	private final Class<? extends VanishServiceInterface> kHandlerClass;
	
	// Refreshable params:
	private RPCServerParams params;
	
	private final SingletonVanishFactory vanish_factory;
	private final StatsCollector stats_collector;
	
	private final WebServer web_server;
	
	protected VanishXMLRPCServer(VanishFactory vanish_factory,
			StatsCollector stats_collector,
			VanishConfiguration conf, VanishLogger log)
	throws VanishException {
		this.vanish_factory = new SingletonVanishFactory(vanish_factory);
		this.stats_collector = stats_collector;
		
		this.CONF = conf;
		this.LOG  = log;
		
		// Non-refreshable params:
		kVanishRPCServerPort = CONF.getInt(Defaults.CONF_VANISH_PORT,
				Defaults.CONF_VANISH_PORT_DEFAULT);
		kVanishRPCServerBindToLocalHost = CONF.getBoolean(
				Defaults.CONF_VANISH_BIND_TO_LOCALHOST,
				Defaults.CONF_VANISH_BIND_TO_LOCALHOST_DEFAULT);
		
		try {
			kHandlerClass = Class.forName(CONF.getString(
					Defaults.CONF_VANISH_SERVICE_HANDLER_CLASS,
					Defaults.CONF_VANISH_SERVICE_HANDLER_CLASS_DEFAULT))
					.asSubclass(VanishServiceInterface.class);
		} catch (Exception e) {
			throw new InvalidArgumentVanishException("Invalid handler class.");
		}
		
		// Load params:
		RPCServerParams params = getRenewedParamsFromConfig(CONF);
		setParams(params);
		
		WebServer ws = null;
		
		try {
			ws = kVanishRPCServerBindToLocalHost
				? new WebServer(kVanishRPCServerPort,
				      InetAddress.getByAddress(new byte[]{ 127, 0, 0, 1 }))
			    : new WebServer(kVanishRPCServerPort);
		} catch (UnknownHostException e) { }  // exception can't happen.
		
		web_server = ws;
	}

	// Initializable interface:
	
	// @Override
	public void init() throws VanishException {
		final LogStat logstat = LOG.logStat("VanishXMLRPCServer.init");
		assert(! isInitialized());

		// Initialize the Vanish factory.
		vanish_factory.init();
		
		// Initialize the service handler class.
		try {
			Method m = kHandlerClass.getDeclaredMethod("init",
					VanishFactory.class, StatsCollector.class,
					VanishConfiguration.class, VanishLogger.class);
			if (! Modifier.isStatic(m.getModifiers())) {
				throw new IllegalAccessException("Method is not static.");
			}
			m.invoke(null, vanish_factory, stats_collector, CONF, LOG);
		} catch (Exception e) {
			LOG.warn("No appropriate init method for class " +
					kHandlerClass.getName() + " could be invoked. " +
					"This method must be static and have params: " +
					"VanishFactory, StatsCollector, VanishConfiguration, " +
					"VanishLogger. The error was: " + e.getMessage());
		}
		
		
		configureXmlRpcServer(kHandlerClass);
		
		// Start the SDS XML server -- should be the last thing we do in init.
		try { web_server.start(); }
	    catch (IOException e) {
	    	e.printStackTrace();
	    	throw new VanishException(e.getMessage());
	    }
	    logstat.end();
	}
	
	private void configureXmlRpcServer(
			Class<? extends VanishServiceInterface> handler_class)
	throws VanishException {
		XmlRpcServer xmlRpcServer = web_server.getXmlRpcServer();
		xmlRpcServer.setMaxThreads(this.params.kNumRPCWorkerThreads);
	    
	    // Add the handlers.
	    PropertyHandlerMapping phm = new PropertyHandlerMapping();
		try {
			phm.addHandler(LocalVanishServiceInterface.kVanishServiceName,
					       handler_class);
			phm.addHandler(RemoteVanishServiceInterface.kVanishServiceName,
					       handler_class);
		} catch (XmlRpcException e) {
			throw new VanishException(e.getMessage());
		}
	    xmlRpcServer.setHandlerMapping(phm);
	    
	    // Configure a bit more.
	    XmlRpcServerConfigImpl serverConfig =
	    	(XmlRpcServerConfigImpl)xmlRpcServer.getConfig();
	    serverConfig.setEnabledForExtensions(true);
	    serverConfig.setContentLengthOptional(false);
	}
	
	// @Override
	public void stop() throws VanishException {
		assert(isInitialized());
		web_server.shutdown();
		vanish_factory.stop();
	}
	
	// @Override
	public boolean isInitialized() { return vanish_factory.isInitialized(); }

	// Helper functions:
	
	protected static void doWork(VanishConfiguration conf)
	throws VanishException {
		VanishLogger log = VanishLogger.getLogger("vanish.rpcserver");
		VanishFactory vanish_factory = instantiateVanishFactory(conf);
		StatsCollector stats_collector = instantiateStatsCollector(conf, log);
		VanishXMLRPCServer vanish_server = new VanishXMLRPCServer(
				vanish_factory, stats_collector, conf, log);
		
		vanish_server.init();
		vanish_server.serve();
		vanish_server.stop();
	}
	
	// TODO(roxana): Do other things here: configuration refreshing, etc.
	private void serve() {
		assert(isInitialized());
		
		while (true) {
			try { Thread.sleep(3600000L); }
			catch (InterruptedException e) { e.printStackTrace(); }
		}
	}	
	
	// TODO(roxana): Move this into a factory!
	private static StatsCollector instantiateStatsCollector(
			VanishConfiguration conf, VanishLogger log)
	throws VanishException {
		return StatsCollectorImpl.getInitializedStatsCollector(
				new VanishTime(), new StatUploaderImpl(conf, log), conf, log);
	} 
	
	private static VanishFactory instantiateVanishFactory(
			VanishConfiguration conf)
	throws VanishException {		
		try {
			String vanish_factory_class_name = conf.getString(
					Defaults.CONF_VANISH_FACTORY_CLASS,
					Defaults.CONF_VANISH_FACTORY_CLASS_DEFAULT);
			Class<? extends VanishFactory> vanish_factory_class =
					Class.forName(vanish_factory_class_name).asSubclass(
							VanishFactory.class);
			
			VanishFactory vanish_factory = null;
			Constructor<?>[] constructors =
				vanish_factory_class.getConstructors();
			for (Constructor<?> constructor: constructors) {
				Class<?>[] p = constructor.getParameterTypes();
				if (p.length == 1 && p[0].isInstance(conf)) {
					vanish_factory = (VanishFactory)constructor.newInstance(
							conf);
					break;
				}
			}
			if (vanish_factory == null) {
				throw new InvalidArgumentVanishException(
						"Invalid VanishFactory class name or constructor " +
						"(vanish.service.vanishfactoryclass)");
			}
			return vanish_factory;
		} catch (Exception e) {
			throw new VanishException("Failed to instantiate VanishFactory",
					e);
		}
	}
	
    // RefreshableConfig interface:
	
	// @Override
	public RPCServerParams getRenewedParamsFromConfig(
			VanishConfiguration config)
	throws InvalidArgumentVanishException {
		final int kNumRPCWorkerThreads = CONF.getInt(
				Defaults.CONF_VANISH_SERVICE_NUM_THREADS,
				Defaults.CONF_VANISH_SERVICE_NUM_THREADS_DEFAULT);
		RPCServerParams params = new RPCServerParams(kNumRPCWorkerThreads);
		validate(params);
		return params;
	}

	// @Override
	public void setParams(RPCServerParams params) {
		this.params = params;
	}
	
	private void validate(RPCServerParams params)
	throws InvalidArgumentVanishException { }
	
	// Main function:
	
	@SuppressWarnings("deprecation")
	public static void main(String args[]) {
		VanishConfiguration conf = Defaults.CONF;
		
		VanishServerCmdLineParser parser = new VanishServerCmdLineParser(conf);
		if (! parser.parse(args)) System.exit(1);

		try { doWork(conf); }
		catch (Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
		
		System.exit(0);
	}
}

class SingletonVanishFactory implements VanishFactory, Initializable {
	private VanishFactory vanish_factory_delegate;
	private VanishInterface vanish;  // TODO(roxana): Maintain a set for expt.
	
	public SingletonVanishFactory(VanishFactory vanish_factory_delegate) {
		this.vanish_factory_delegate = vanish_factory_delegate;
		vanish = null;
	}

	public VanishInterface getInitializedVanishSystem()
	throws VanishException {
		assert(isInitialized());
		return vanish;
	}

	// @Override
	public void init() throws VanishException {
		if (isInitialized()) throw new VanishException("Already initialized");
		vanish = vanish_factory_delegate.getInitializedVanishSystem();
	}

	// @Override
	public boolean isInitialized() { return vanish != null; }

	// @Override
	public void stop() throws VanishException {
		assert(isInitialized());
		vanish.stop();
	}
}
