/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.stats.impl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.zip.GZIPOutputStream;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.methods.multipart.FilePart;
import org.apache.commons.httpclient.methods.multipart.MultipartRequestEntity;
import org.apache.commons.httpclient.methods.multipart.Part;
import org.apache.commons.httpclient.methods.multipart.StringPart;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.logging.LogStat;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.stats.StatUploader;
import edu.washington.cs.vanish.util.Defaults;
import edu.washington.cs.vanish.util.FileUtil;
import edu.washington.cs.vanish.util.IOVanishException;
import edu.washington.cs.vanish.util.RefreshableConfig;
import edu.washington.cs.vanish.util.RefreshableParams;

/**
 * 
 * @author roxana
 *
 */

class StatsUploaderParams implements RefreshableParams {	
	public final String kStatsCollectorURL;
	public final int kTimeout_ms;
	public final boolean kShouldZip;
	
	public StatsUploaderParams(String statsCollectorURL, int timeout_ms,
			boolean shouldZip) {
		this.kStatsCollectorURL = statsCollectorURL;
		this.kTimeout_ms = timeout_ms;
		this.kShouldZip = shouldZip;
	}
}

public class StatUploaderImpl
implements StatUploader, RefreshableConfig<StatsUploaderParams> {
	private final VanishLogger LOG;
	
	private StatsUploaderParams params;
	
	private final HttpClient http_client;
	
	public StatUploaderImpl(VanishConfiguration CONF, VanishLogger LOG)
	throws InvalidArgumentVanishException {
		this.LOG = LOG;
		
		setParams(getRenewedParamsFromConfig(CONF));
		
		http_client = new HttpClient();
		initHttpClient();
	}
	
	private void initHttpClient() {
		http_client.getHttpConnectionManager().
			getParams().setConnectionTimeout(params.kTimeout_ms);
	}
	
	// @Override
	public void upload(File file) throws IOVanishException {
		if (file == null || !file.exists() || file.length() == 0) return;

		LogStat log_stat = LOG.logStat("StatUploaderImpl.upload");

		IOVanishException exception = null;
		File new_file = null;
		try {
			new_file = prepareFileForUpload(file);
			if (new_file != null) performUpload(new_file);
		} catch(IOVanishException e) {
			exception = e;
		} finally {
			if (new_file != null && new_file != file) new_file.delete();
		}

		log_stat.end(exception);

		if (exception != null) throw exception;
	}
	
	/**
	 * Prepares a file for upload and returns a possibly new file that is the
	 * prepared version.
	 * @param file the file to prepare
	 * @return the file that should be uploaded. This file can have three types
	 * of values:
	 * <ul>
	 * <li> the returned file can be the input file object (returned == file),
	 *      meaning that the input object should be uploaded.
	 * <li> the returned file can be a brand new temporary file
	 *      (returned != file), which should be deleted by the caller after
	 *      upload.
	 * <li> the returned file should be null, in which case nothing should
	 *      be uploaded.
	 * </ul>
	 * @throws IOVanishException
	 */
	protected File prepareFileForUpload(File file) throws IOVanishException {
		if (params.kShouldZip) return zipFile(file);
		else return file;
	}
	
	/**
	 * Zips the input file to a tmp file and returnes the tmp file.
	 * @param file must not be null
	 * @return the tmp file containing the zipped version of the input file.
	 * @throws IOVanishException if any error has occurred.
	 * @throws NullPointerException if the input file is null.
	 */
	private File zipFile(File file) throws IOVanishException {
		final File tmp_file = FileUtil.generateTmpFile();
		
		InputStream file_is = null;
		OutputStream tmp_file_os = null;
		try {
			file_is = new FileInputStream(file);
			tmp_file_os = new GZIPOutputStream(
				new FileOutputStream(tmp_file, false));
			FileUtil.writeInputOutput(file_is, tmp_file_os);
		} catch (IOException e) {
			tmp_file.delete();
			throw new IOVanishException("Failed to write gzipped tmp file", e);
		} finally {
			if (tmp_file_os != null) {
				try { tmp_file_os.close(); }
				catch (IOException e) { }
			}
			if (file_is != null) {
				try { file_is.close(); }
				catch (IOException e) { }
			}
		}
		return tmp_file;
	}
	
	// The following HTTPClient-related code takes after the sample code at:
	// http://www.java-tips.org/other-api-tips/httpclient/\
	// how-to-use-multipart-post-method-for-uploading.html
	
	/**
	 * Performs an upload of the input file. 
	 * @param file
	 * @throws IOVanishException
	 */
	protected void performUpload(File file) throws IOVanishException {
		PostMethod post = new PostMethod(params.kStatsCollectorURL);
		int status = HttpStatus.SC_INTERNAL_SERVER_ERROR;
		try {            
			Part[] parts = constructPostParams(file);
			post.setRequestEntity(new MultipartRequestEntity(parts,
					post.getParams()));

			status = http_client.executeMethod(post); 
		} catch (Exception e) {
			throw new IOVanishException("Failed to submit stats", e);
		} finally {
			post.releaseConnection();
		}

        if (status != HttpStatus.SC_OK &&
			status != HttpStatus.SC_MOVED_TEMPORARILY) {
			throw new IOVanishException(
				"Failed to post the stats (errorcode " + status + ")");
		}
	}

	private Part[] constructPostParams(File file)
	throws FileNotFoundException {
		if (params.kShouldZip) {
			return new Part[] { new StringPart("zip", "gzip"),
                                new FilePart("content", file) };
		} else {
			return new Part[] { new FilePart("content", file) };
		}
	}
	
	public StatsUploaderParams getRenewedParamsFromConfig(
			VanishConfiguration config)
	throws InvalidArgumentVanishException {
		final String statsCollectorURL = config.getString(
				Defaults.CONF_VANISH_STATS_SAVEURL,
				Defaults.CONF_VANISH_STATS_SAVEURL_DEFAULT);
		final int timeout_ms = (int)config.getTimeInMillis(
				Defaults.CONF_VANISH_STATS_SAVETIMEOUT,
				Defaults.CONF_VANISH_STATS_SAVETIMEOUT_DEFAULT,
				VanishConfiguration.TimeUnit.SECONDS);
		final boolean should_gzip = config.getBoolean(
				Defaults.CONF_VANISH_STATS_SHOULDGZIP,
				Defaults.CONF_VANISH_STATS_SHOULDGZIP_DEFAULT);
		final StatsUploaderParams params =
			new StatsUploaderParams(statsCollectorURL, timeout_ms,
					                should_gzip);
		
		validate(params);
		
		return params;
	}
	
	private void validate(StatsUploaderParams params)
	throws InvalidArgumentVanishException { }  // nothing to do.

	public void setParams(StatsUploaderParams params) {
		this.params = params;
	}
}
