/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.stats.impl;

import java.io.File;
import java.io.IOException;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import org.apache.log4j.Layout;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.apache.log4j.RollingFileAppender;
import org.apache.log4j.WriterAppender;
import org.apache.log4j.spi.LoggingEvent;


import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.service.impl.IllegalServerStateVanishException;
import edu.washington.cs.vanish.stats.Stat;
import edu.washington.cs.vanish.stats.StatUploader;
import edu.washington.cs.vanish.stats.StatsCollector;
import edu.washington.cs.vanish.stats.StatsCollector.ReportLevel;
import edu.washington.cs.vanish.util.Defaults;
import edu.washington.cs.vanish.util.FileUtil;
import edu.washington.cs.vanish.util.IOVanishException;
import edu.washington.cs.vanish.util.RefreshableConfig;
import edu.washington.cs.vanish.util.RefreshableParams;
import edu.washington.cs.vanish.util.VanishTime;

/**
 * 
 * @author roxana
 */
class StatsCollectorParams implements RefreshableParams {	
	public final ReportLevel kStatReportingLevel;
	
	public StatsCollectorParams(ReportLevel statReportingLevel) {
		this.kStatReportingLevel = statReportingLevel;
	}
}

public class StatsCollectorImpl
implements StatsCollector, RefreshableConfig<StatsCollectorParams> {
	private static final String kStatsFilename =
		FileUtil.getPath(System.getProperty("vanish.log.dir"),
				         System.getProperty("vanish.log.file") + ".stat");
	
	private final VanishConfiguration CONF;
	private final VanishLogger LOG;
	private final VanishTime TIME;
	
	private StatsCollectorParams params;
	private final ReadWriteLock params_lock = new ReentrantReadWriteLock();
	
	// State:
	
	private final Logger stats_logger;
	private WriterAppender stats_appender;
	
	private final StatUploader uploader;
	
	private boolean is_initialized;
	
	protected StatsCollectorImpl(final VanishTime time,
			                     final StatUploader uploader,
			                     final VanishConfiguration conf,
			                     final VanishLogger log)
	throws InvalidArgumentVanishException {
		this.CONF = conf;
		this.LOG = log;
		this.TIME = time;
		this.uploader = uploader;
		
		setParams(getRenewedParamsFromConfig(CONF));

		stats_logger = Logger.getLogger("stats");
		configureStatsLogger();
	}
	
	/** Use only in testing. */
	protected Logger getStatsLogger() { return stats_logger; }

	private void configureStatsLogger() throws InvalidArgumentVanishException {
		stats_logger.setLevel(Level.INFO);
		stats_logger.setAdditivity(false);

		final Layout layout = new PatternLayout("%m%n") {
			@Override
			public String format(LoggingEvent event) {
				// Prints UTC time, not timezone-dependent time.
				return "ts=" + TIME.currentTimeMillis() + ";" +
				       super.format(event);
			}
		};
		
		if (stats_logger.getAppender("STAT") != null) {
			stats_logger.removeAppender("STAT");
		}

		stats_appender = createAppenderUsingLayout(layout);
		stats_appender.setName("STAT");
		stats_logger.addAppender(stats_appender);
	}
	
	protected WriterAppender createAppenderUsingLayout(Layout layout)
	throws InvalidArgumentVanishException {
		try {
			return new TimeAndFileSizeRollingAppender(layout, kStatsFilename);
		} catch (IOException e) {
			throw new InvalidArgumentVanishException(
					"Failed to create appender to file " + kStatsFilename);
		}
	}
	
	private class TimeAndFileSizeRollingAppender
	extends RollingFileAppender {
		private final Timer rollover_timer;
		private final File backup_file;

		public TimeAndFileSizeRollingAppender(Layout layout, String filename)
		throws IOException {
			super(layout, filename, false);
			
			setMaxBackupIndex(1);
			setMaximumFileSize(CONF.getInt(
					Defaults.CONF_VANISH_STATS_BULK_UPLOAD_SIZE,
					Defaults.CONF_VANISH_STATS_BULK_UPLOAD_SIZE_DEFAULT));
			backup_file = new File(filename + ".1");

			final long rollover_interval = CONF.getTimeInMillis(
					Defaults.CONF_VANISH_STATS_SAVEINTERVAL_M,
					Defaults.CONF_VANISH_STATS_SAVEINTERVAL_M_DEFAULT,
					VanishConfiguration.TimeUnit.MINUTES); 
			
			rollover_timer = new Timer("rollover", false);
			rollover_timer.schedule(
				new TimerTask() {
					@Override
					public void run() { rollOver(); }
				},
				rollover_interval,
				rollover_interval);
		}
		
		@Override
		public void rollOver() {
			// First, roll over to the other file.
			super.rollOver();

			// Then, submit the rolled over file to the server.
			try {
				uploadStats(backup_file);
			} catch (IOVanishException e) { 
				LOG.error("Failed to submit stats", e);
			}
		}
		
		@Override
		public void close() {
			rollover_timer.cancel();
			rollOver();  // do a final uploading.
			super.close();
		}
	}

	
	// StatsCollector interface:

	// @Override
	public void report(final Stat stat)
	throws IllegalServerStateVanishException {
		String stat_str = getToReport(stat);
		if (stat_str != null) stats_logger.info(stat_str);
	}
	
	private String getToReport(final Stat stat)
	throws IllegalServerStateVanishException {
		if (! isInitialized()) {
			throw new IllegalServerStateVanishException(
					"Stats collector not initialized");
		}
		
		ReportLevel level = getCurrentReportLevel();
		if (level == ReportLevel.NONE) return null;  // don't report anything.

		return stat.checkpoint(level);
	}
	
	// @Override
	public void report(final Stat stat, Throwable t)
	throws IllegalServerStateVanishException {
		if (t == null) {
			report(stat);
		} else {
			String stat_str = getToReport(stat);
			if (stat_str != null) stats_logger.error(stat_str, t);
		}
	}
	
	// @Override
	public ReportLevel getCurrentReportLevel() {
		params_lock.readLock().lock();
		ReportLevel report_level = params.kStatReportingLevel;
		params_lock.readLock().unlock();
		return report_level;
	}
	
	protected void uploadStats(File file) throws IOVanishException {
		if (getCurrentReportLevel().compareTo(ReportLevel.NONE) > 0) {
			uploader.upload(file);
		}
	}
	
	// Initializable interface:
	
	// @Override
	public void init() throws VanishException {	
		// This must be the last instruction in the function.
		is_initialized = true;
	}

	// @Override
	public boolean isInitialized() { return is_initialized; }

	// @Override
	public void stop() throws VanishException { stats_appender.close(); }
	
	// RefreshableConfig interface:
	
	// @Override
	public StatsCollectorParams getRenewedParamsFromConfig(
			VanishConfiguration config)
	throws InvalidArgumentVanishException {
		final String statsLevel_str = config.getString(
				Defaults.CONF_VANISH_STATS_ENABLED,
				Defaults.CONF_VANISH_STATS_ENABLED_DEFAULT)
				.toUpperCase();
		final ReportLevel kStatsLevel = ReportLevel.valueOf(statsLevel_str);
		
		StatsCollectorParams params = new StatsCollectorParams(kStatsLevel);
		
		validate(params);
		
		return params;
	}
	
	private void validate(StatsCollectorParams params)
	throws InvalidArgumentVanishException { }

	// @Override
	public void setParams(StatsCollectorParams params) {
		params_lock.writeLock().lock();
		this.params = params;
		params_lock.writeLock().unlock();
	}
	
	private static StatsCollector stats_collector;
	
	public static StatsCollector getInitializedStatsCollector(
            final VanishTime time,
            final StatUploader uploader,
            final VanishConfiguration conf,
            final VanishLogger log)
	throws VanishException {
		if (stats_collector == null) {
			stats_collector = new StatsCollectorImpl(time, uploader,
					conf, log);
			stats_collector.init();
		}
		return stats_collector;
	}
}
