/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.system;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.TimerTask;

import edu.washington.cs.vanish.logging.LogStat;
import edu.washington.cs.vanish.logging.VanishLogger;


/**
 * @author roxana
 */
public abstract class VanishRefreshTask extends TimerTask
implements Serializable {
	private static final VanishLogger LOG =
		VanishLogger.getLogger("vanish.refresh");
	
	// Unserialized state:
	private VanishInterface vanish_system_ptr;
	
	// Serialized state:
	private long timeout_date;
	private long next_scheduled_execution_time;
	private long ttl_after_refresh;

	public VanishRefreshTask(VanishInterface vanish_system_ptr,
							 long creation_date,
			                 long timeout_date,
			                 long ttl_after_refresh) {
		this.timeout_date        = timeout_date;
		this.next_scheduled_execution_time = creation_date + ttl_after_refresh;
		this.ttl_after_refresh   = ttl_after_refresh;
		
		init(vanish_system_ptr);
	}
	
	/**
	 * Explicitly used when object is obtained via de-serialization.
	 * @param vanish_system_ptr
	 */
	public void init(VanishInterface vanish_system_ptr) {
		this.vanish_system_ptr = vanish_system_ptr;
	}
	
	private boolean isInitialized() { return vanish_system_ptr != null; }

	// TimerTask functions:

	@Override
	public synchronized final void run() {
		if (!isInitialized()) { LOG.error("Task not initialized"); }
		final LogStat logstat = LOG.logStat("VanishRefreshTask.run");
		long now = getCurrentTime();
		if (shouldRefreshNow(now)) {  // Refresh.
			refresh(vanish_system_ptr);
			next_scheduled_execution_time = now + ttl_after_refresh;
		}
		
		// See if this task should continue or now.
		if (! shouldStillBeRefreshingAtTime(next_scheduled_execution_time)) {
			this.cancel();
		}

		logstat.end();
	}

	protected long getCurrentTime() { return scheduledExecutionTime(); }

	// Helper functions; can be overriden by subclasses.

	// None of this is synchronized.

	/**
	 * Thread-safe; it touches only members with final semantics. 
	 */
	public boolean shouldStillBeRefreshingAtTime(long time) {
		return (time + ttl_after_refresh <= timeout_date);
	}

	/**
	 * Not thread-safe. Use only when this task is not scheduled on a timer.
	 * 
	 * @param now
	 * @return
	 */
	public boolean shouldRefreshNow(long now) {
		return (next_scheduled_execution_time <= now &&
		        shouldStillBeRefreshingAtTime(now));
	}
	
	/**
	 * Thread-safe.
	 * @return
	 */
	public synchronized long getNextScheduledRefreshTime() { 
		return next_scheduled_execution_time;
	}
	
	@Deprecated
	public long getRefreshInterval() { return ttl_after_refresh; }

	protected abstract void refresh(VanishInterface vanish_system);
	
	// Serializable-related methods:
	
	/**
	 * Thread-safe.
	 */
	private void writeObject(ObjectOutputStream out)
    throws IOException {
		out.writeLong(timeout_date);
		out.writeLong(ttl_after_refresh);
		out.writeLong(getNextScheduledRefreshTime());
	}

	/**
	 * Not thread-safe (no need to be).
	 * @param in
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private void readObject(ObjectInputStream in)
    throws IOException, ClassNotFoundException {
		timeout_date = in.readLong();
		ttl_after_refresh = in.readLong();
		next_scheduled_execution_time = in.readLong();
	}
	
	// Object functions:
	
	/**
	 * Use only for testing.
	 */
	@Override
	public boolean equals(Object other) {
		if (this == other) return true;
		if (other == null) return false;
		if (! (other instanceof VanishRefreshTask)) return false;
		VanishRefreshTask o = (VanishRefreshTask)other;
		return (this.timeout_date == o.timeout_date &&
		        this.next_scheduled_execution_time ==
		        	o.next_scheduled_execution_time);
	}
	
	/**
	 * Use only for testing.
	 */
	@Override
	public int hashCode() {
		int hash = 7;
		hash = hash * 31 + (int)timeout_date;
		hash = hash * 31 + (int)next_scheduled_execution_time;
		return hash;
	}
}

