/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.system.impl;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.Date;
import java.util.Queue;
import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.ConcurrentLinkedQueue;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.MetadataEncapsulatedVanishException;
import edu.washington.cs.vanish.internal.MetadataEncapsulatedVanishInternalInterface;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.impl.InputStreamWrapper;
import edu.washington.cs.vanish.internal.impl.OutputStreamWrapper;
import edu.washington.cs.vanish.internal.metadata.Metadata;
import edu.washington.cs.vanish.internal.metadata.MetadataFactory;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.logging.LogStat;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.system.VanishInterface;
import edu.washington.cs.vanish.system.VanishRefreshTaskFactory;
import edu.washington.cs.vanish.util.Constants;
import edu.washington.cs.vanish.util.Defaults;
import edu.washington.cs.vanish.util.RefreshableConfig;
import edu.washington.cs.vanish.util.RefreshableParams;
import edu.washington.cs.vanish.util.VanishTime;

/**
 * 
 * @author roxana
 */
class VanishParams implements RefreshableParams {
	// Prepush task params.
	final int kNumPrepushVDOs;
	final int kPrepushThreshold;
	final long kPrepushInterval;
	final long kPrepushDelayBetweenPushes;
	
	public VanishParams(int numPrepushVDOs, int prepushThreshold,
			            long prepushInterval,
			            long prepushDelayBetweenConsecPushes) {		
		this.kNumPrepushVDOs = numPrepushVDOs;
		this.kPrepushThreshold = prepushThreshold;
		this.kPrepushInterval = prepushInterval;
		this.kPrepushDelayBetweenPushes =
			prepushDelayBetweenConsecPushes;
	}
}

public class VanishImpl
implements VanishInterface, RefreshableConfig<VanishParams>, Constants {	
	private final VanishLogger LOG;
	private final VanishConfiguration CONF;
	
	private final MetadataEncapsulatedVanishInternalInterface vanish_core;
	private final MetadataFactory metadata_factory;
	private final Random metadata_random_generator;
	
	private VanishParams params;
	
	// Management tasks:
	private final VanishRefreshTimer refresher;
	private final PrepushDaemonThread prepusher;

	private ExpiringMinimalCapacityConcurrentQueue<Metadata> prepushed_metadata_objects;
	
	public VanishImpl(MetadataEncapsulatedVanishInternalInterface vanish_core,
			          MetadataFactory metadata_factory,
			          VanishRefreshTaskFactory refresher_factory,
			          Random metadata_random_generator,
			          VanishTime time,
			          VanishConfiguration conf,
			          VanishLogger log)
	throws InvalidArgumentVanishException {
		this.vanish_core      = vanish_core;
		this.metadata_factory = metadata_factory;

		this.CONF = conf;
		this.LOG  = log;
		
		VanishParams params = getRenewedParamsFromConfig(CONF);
		setParams(params);
		
		final boolean kShouldUseOnlyPrepushedMetadata = CONF.getBoolean(
			Defaults.CONF_VANISH_SYSTEM_PREPUSH_SHOULD_ONLY_USE_PREPUSH,
			Defaults.CONF_VANISH_SYSTEM_PREPUSH_SHOULD_ONLY_USE_PREPUSH_DEFAULT);
		
		this.metadata_random_generator = metadata_random_generator;
		
		// Background management-related initializations:
		refresher = new VanishRefreshTimer(this, refresher_factory, time,
				                           CONF, LOG);

		if (params.kNumPrepushVDOs > 0) {
			prepusher = new PrepushDaemonThread();
			prepushed_metadata_objects =
				new ExpiringMinimalCapacityConcurrentQueue<Metadata>(
						params.kNumPrepushVDOs, params.kPrepushThreshold,
						kShouldUseOnlyPrepushedMetadata);
		} else {
			prepusher = null;
		}
	}
	
	// Initializable interface:

	// @Override
	public void init() throws VanishException {	
		vanish_core.init();

		refresher.init();

		// Background management processes:
		if (prepusher != null) { prepusher.start(); }
	}

	// @Override
	public boolean isInitialized() { return vanish_core.isInitialized(); }

	// @Override
	public void stop() throws VanishException {
		vanish_core.stop();
		refresher.stop();
		// if (prepusher != null) prepusher.cancel();  // nothing to do: daemon
	}
	
	// VanishInterface:
	
	// @Override
	public Metadata encapsulate(InputStream clear_data_is,
			                    OutputStreamWrapper enc_data_os,
				                VDOParams params)
	throws VanishException {
		checkState();
		
		Metadata metadata = tryToEncapsulateUsingPrepushedMetadata(
				clear_data_is, enc_data_os, params);
		if (metadata == null) {  // Push new metadata (will be slow).
			metadata = encapsulateUsingNewMetadata(clear_data_is, enc_data_os,
                                                   params);
		}
		return metadata;
	}

	// @Override
	@SuppressWarnings("unchecked")
	public Metadata decapsulate(InputStreamWrapper enc_data_is,
			                    OutputStream clear_data_os)
	throws MetadataEncapsulatedVanishException, VanishException {
		checkState();
		return vanish_core.decapsulate(enc_data_is, clear_data_os);
	}
	
	// @Override
	public void refreshReadOnly(Metadata metadata) throws VanishException {
		checkState();
		vanish_core.refreshReadOnly(metadata);
	}
	
	// @Override
	public Metadata refreshDeeply(InputStreamWrapper enc_data_is,
			                      OutputStreamWrapper refreshed_enc_data_os,
			                      VDOParams params)
	throws MetadataEncapsulatedVanishException, VanishException {
		checkState();
		
		Metadata new_metadata = generateMetadata(params);
		if (new_metadata == null) {
			throw new InvalidArgumentVanishException(
					"Failed to create metadata.");
		}
		return vanish_core.refreshDeeply(enc_data_is, refreshed_enc_data_os,
				                         new_metadata);
	}
	
	// Helper functions:
	
	private void checkState() throws VanishException {
		if (!isInitialized()) {
			throw new VanishException("Invalid Vanish state");
		}
	}
	
	private Metadata tryToEncapsulateUsingPrepushedMetadata(
			InputStream clear_data_is, OutputStreamWrapper enc_data_os,
            VDOParams params)
	throws VanishException {
		// Find an appropriate prepushed metadata.
		Metadata metadata = getSuitablePrepushedMetadata(params);		
		if (metadata == null) return null;  // Couldn't find any.
		
		// Inherit all parameters that we can from the user-given parameters.
		metadata.getVDOParams().inheritNonFixedParamsFromOtherVDOParams(
				params);

		// Was able to find, so let's encapsulate.
		vanish_core.encapsulateUsingPrepushedMetadata(clear_data_is,
				enc_data_os, metadata);
		return metadata;
	}
	
	private Metadata getSuitablePrepushedMetadata(final VDOParams params) {
		if (this.prepushed_metadata_objects != null &&
		    VDOParams.isDefaultVDOParams(params)) {
			return prepushed_metadata_objects.poll();  // doesn't block and
			                                           // may return null.
		} else {
			return null;  // nothing available
		}
	}
	
	private Metadata generateMetadata(VDOParams params) {
		long seed = this.metadata_random_generator.nextLong();
		return metadata_factory.createMetadata(params, seed);
	}
	
	private Metadata encapsulateUsingNewMetadata(InputStream clear_data_is,
			OutputStreamWrapper enc_data_os,
            VDOParams params)
	throws VanishException {
		Metadata metadata = generateMetadata(params);
		if (metadata == null) {
			throw new InvalidArgumentVanishException(
					"Failed to create metadata.");
		}

		vanish_core.encapsulate(clear_data_is, enc_data_os, metadata);
		
		refresher.scheduleRefreshesIfNeeded(metadata, enc_data_os);

		return metadata;
	}

	// Background management tasks:

	private class PrepushDaemonThread extends Thread {
		public PrepushDaemonThread() {
			super("PrepushThread");
			super.setDaemon(true);
		}

		@Override
		public void run() {
			while (true) {  // Run forever, as a daemon.
				synchronized(prepushed_metadata_objects) {
					while (! prepushed_metadata_objects.needMoreElements()) {
						try { prepushed_metadata_objects.wait();}
						catch (InterruptedException e) { }
					}
				}
				
				// Need to prepush more elements.
				// Lock on prepushed_metadata_objects not needed, as there's
				// only one prepusher, so needMoreElements() will keep 
				// returning true until this thread says otherwise.
				// TODO(roxana): If you ever add support for multiple
				// prepushers, this will need to change somehow.
				while (prepushed_metadata_objects.canFitMoreElements()) {
					long expiration_time = System.currentTimeMillis() +
                    		               params.kPrepushInterval;
					Metadata prepushed_object = prepushVDO();
					long time_to_sleep = 100;
					if (prepushed_object != null &&
						prepushed_metadata_objects.add(prepushed_object,
							                           expiration_time)) {
						time_to_sleep = params.kPrepushDelayBetweenPushes;
					}
					
					// Aggresivity is bad, so sleep between prepushes.
					if (time_to_sleep > 0) {
						try { Thread.sleep(time_to_sleep); }
						catch (InterruptedException e) { }
					}
				}
			}
		}
		
		private Metadata prepushVDO() {
			final LogStat logstat = LOG.logStat(
					"VanishImpl#PrepushTask.prepushOne");
			Metadata metadata = null;
			try {
				metadata = encapsulateUsingNewMetadata(null, null,
						VDOParams.newDefaultVDOParams());
			} catch (VanishException e) {
				LOG.error("Failed to prepush", e);
			}
			logstat.end();
			return metadata;
		}
	}

	// RefreshableConfig interface:
	
	// @Override
	public VanishParams getRenewedParamsFromConfig(VanishConfiguration config)
	throws InvalidArgumentVanishException {
		final int kNumPrepushVDOs = config.getInt(
				Defaults.CONF_VANISH_SYSTEM_PREPUSH_NUMVDOS,
				Defaults.CONF_VANISH_SYSTEM_PREPUSH_NUMVDOS_DEFAULT);
		final int kPrepushThreshold = config.getInt(
				Defaults.CONF_VANISH_SYSTEM_PREPUSH_THESHOLD,
				Defaults.CONF_VANISH_SYSTEM_PREPUSH_THESHOLD_DEFAULT);
		final long kPrepushInterval = config.getTimeInMillis(
				Defaults.CONF_VANISH_SYSTEM_PREPUSH_PERIOD_MIN,
				Defaults.CONF_VANISH_SYSTEM_PREPUSH_PERIOD_MIN_DEFAULT,
				VanishConfiguration.TimeUnit.MINUTES);
		
		final long kPrepushDelayBetweenConsecPushes = 500;
		
		VanishParams params = new VanishParams(
				kNumPrepushVDOs, kPrepushThreshold,
				kPrepushInterval, kPrepushDelayBetweenConsecPushes);
		validate(params);
		return params;
	}

	// @Override
	public void setParams(VanishParams params) {
		this.params = params;
	}
	
	private void validate(VanishParams params)
	throws InvalidArgumentVanishException {
		if (params.kPrepushInterval < 15) {
			throw new InvalidArgumentVanishException(
					"Prepush interval must be >= 15 minutes " +
					"(vanish.system.refresh.period_h)");
		}
		if (params.kNumPrepushVDOs < 0) {
			throw new InvalidArgumentVanishException(
				"Number of prepushed VDOs must be 0 or positive" +
				"(vanish.system.prepush.numvdos)");
		}
		if ((params.kPrepushThreshold < 0) ||
			(params.kNumPrepushVDOs > 0 &&        // prepush is enabled, but
			 (params.kPrepushThreshold == 0 ||    // threshold disables it.
			  params.kPrepushThreshold > params.kNumPrepushVDOs))) {
			// Second && condition above prevents disabling of prepushes by
			// setting kPrepushThreshold = 0. To disable prepushes, one must
			// set kNumPrepushVDOs to 0, and then kPrepushThreshold doesn't
			// matter anymore.
			throw new InvalidArgumentVanishException(
					"The prepushed VDO threshold must be a positive that's " +
					"<= vanish.system.prepush.numvdos " +
					"(vanish.system.prepush.numvdos)");
		}
	}
}

//TODO(roxana): Should move outside and test separately, but then we need
//to implement the entire BlockingQueue interface.

class ExpiringMinimalCapacityConcurrentQueue<T> {
	private int maximum_capacity;
	private int threshold;
	private boolean should_reuse_element_until_expiration;

	private Queue<ElementWrapper> backend;
	
	private Timer expiration_timer;
	
	public ExpiringMinimalCapacityConcurrentQueue(int maximum_capacity,
			int threshold,
			boolean should_reuse_element_until_expiration) {
		backend = new ConcurrentLinkedQueue<ElementWrapper>();
		expiration_timer = new Timer(true);  // make it a daemon
		
		this.maximum_capacity = maximum_capacity;
		this.threshold = threshold;
		this.should_reuse_element_until_expiration =
			should_reuse_element_until_expiration;
	}

	public boolean add(T object, long expiration_ts) {
		if (object == null || System.currentTimeMillis() > expiration_ts) {
			// Don't add the element, as it's already expired.
			return false;
		}

		// The object needs to be pushed.
		ElementWrapper e = new ElementWrapper(object);
		expiration_timer.schedule(e, new Date(expiration_ts));
		boolean added = backend.add(e);
		if (!added) e.cancel();
		return added;
	}
	
	public T poll() {  // serv.
		ElementWrapper e = backend.poll();
		if (e != null) {
			if (should_reuse_element_until_expiration) backend.add(e);
			else markElementAsRemoved(e);
			return e.get();
		} else {
			return null;
		}
	}
	
	private void remove(ElementWrapper e) {
		if (backend.remove(e)) markElementAsRemoved(e);
	}
	
	private void markElementAsRemoved(ElementWrapper e) {
		e.cancel();
		synchronized(this) {
			if (needMoreElements()) this.notify();  // wake prepusher up.
		}
	}
	
	public boolean needMoreElements() {
		return backend.size() < threshold;
	}
	
	public boolean canFitMoreElements() {
		return backend.size() < maximum_capacity;
	}
	
	private class ElementWrapper extends TimerTask {
		private T object;
		
		public ElementWrapper(T object) { this.object = object; }
		
		public T get() { return object; }

		// TimerTask functions:

		@Override
		public void run() { remove(this); }
		
		// Object functions:
		
		@SuppressWarnings("unchecked")
		@Override
		public boolean equals(Object other) {
			if (other == null) return false;
			if (other == this) return true;
			if (! (other.getClass().equals(ElementWrapper.class))) {
				return false;
			}
			ElementWrapper o = (ElementWrapper)other;
			if (this.object != null) return this.object.equals(o.object);
			else return o.object == null;
		}
		
		public int hashCode() {
			return object == null ? 0 : object.hashCode();
		}
	}
}