/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.system.impl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.util.Date;
import java.util.Iterator;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.metadata.Metadata;
import edu.washington.cs.vanish.internal.metadata.impl.EpochAwareMetadataImpl;
import edu.washington.cs.vanish.logging.LogStat;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.system.VanishInterface;
import edu.washington.cs.vanish.system.VanishRefreshTask;
import edu.washington.cs.vanish.system.VanishRefreshTaskFactory;
import edu.washington.cs.vanish.util.Constants;
import edu.washington.cs.vanish.util.Defaults;
import edu.washington.cs.vanish.util.FileUtil;
import edu.washington.cs.vanish.util.Initializable;
import edu.washington.cs.vanish.util.RefreshableConfig;
import edu.washington.cs.vanish.util.RefreshableParams;
import edu.washington.cs.vanish.util.VanishTime;

/**
 * 
 * @author roxana
 */
class VanishRefreshParams implements RefreshableParams {
	final long kRefreshInterval;
	final long kSaveStateInterval;
	final long kSlackTimeForRefresh;
	final File kSaveStateFile;
	
	public VanishRefreshParams(long refreshInterval, long saveStateInterval,
			long slackTimeForRefresh,
			File kSaveStateFile) {		
		this.kRefreshInterval = refreshInterval;
		this.kSaveStateInterval = saveStateInterval;
		this.kSlackTimeForRefresh = slackTimeForRefresh;
		this.kSaveStateFile = kSaveStateFile;
	}
}

public class VanishRefreshTimer
implements RefreshableConfig<VanishRefreshParams>, Initializable, Constants {
	private static final String SAVE_STATE_FILENAME = "refresh.vdo"; 
	
	private final VanishConfiguration CONF;
	private final VanishLogger LOG;
	private final VanishTime TIME;
	private final VanishInterface vanish;
	private final VanishRefreshTaskFactory refresher_factory;
	
	private final Timer timer;
	private final Timer save_state_timer;
	
	private VanishRefreshParams params;

	private Vector<VanishRefreshTask> tasks;
	
	private SaveStateToDiskTask state_serializer;

	public VanishRefreshTimer(VanishInterface vanish,
			VanishRefreshTaskFactory refresher_factory,
			VanishTime time,
			VanishConfiguration conf,
			VanishLogger log)
	throws InvalidArgumentVanishException {
		this.vanish = vanish;
		this.refresher_factory = refresher_factory;
		this.TIME = time;
		this.CONF = conf;
		this.LOG = log;
		
		timer = new Timer("refresh", true);
		
		save_state_timer = new Timer("save.state", true);
		
		state_serializer = new SaveStateToDiskTask();
		
		VanishRefreshParams params = getRenewedParamsFromConfig(conf);
		setParams(params);
	}
	
	/**
	 * Used for testing; please do not use elsewhere.
	 */
	protected VanishRefreshParams getParams() {
		return params;
	}
	
    // Initializable interface:

	// @Override
	@SuppressWarnings("deprecation")
	public void init() {
		Vector<VanishRefreshTask> tasks_set = readTasksFromFile(
				params.kSaveStateFile);
		if (tasks_set == null) tasks_set = new Vector<VanishRefreshTask>();
		
		long now = TIME.currentTimeMillis();
		Iterator<VanishRefreshTask> it = tasks_set.iterator();
		while (it.hasNext()) {
			VanishRefreshTask task = it.next();
			
			task.init(vanish);

			if (task.shouldStillBeRefreshingAtTime(now)) {  // schedule refresh
				timer.schedule(task,
						new Date(task.getNextScheduledRefreshTime()),
						    // schedules immediately if the date has passed.
						task.getRefreshInterval());
			} else {  // otherwise, just drop it, as it's expired
				System.out.println("init: REMOVING ONE");
				it.remove();
			}
		}
		
		// Do other initializations.
		if (isSaveStateEnabled()) {  // schedule save/cleanup task.
			if (LOG.isDebugEnabled()) {
				LOG.debug("Scheduling state-saving task at interval " +
						  params.kSaveStateInterval);
			}
			save_state_timer.schedule(state_serializer,
					                  params.kSaveStateInterval,
					                  params.kSaveStateInterval);
		}
		
		// As last operation, commit the initialization.
		this.tasks = tasks_set;
	}

	// @Override
	public boolean isInitialized() { return tasks != null; }

	// @Override
	public void stop() {
		timer.cancel();
		save_state_timer.cancel();
		
		// Force the state serializer.
		if (isSaveStateEnabled()) state_serializer.run();
	}
	
	private boolean isSaveStateEnabled() {
		return params.kSaveStateInterval > 0;
	}

	/**
	 * @param metadata
	 * @param enc_data_os  currently unused, but reserved for future use.
	 * @throws VanishException
	 */
	public void scheduleRefreshesIfNeeded(Metadata metadata,
			@SuppressWarnings("unused") OutputStream enc_data_os)
	throws VanishException {
		if (! isInitialized()) {
			throw new VanishException("Refresh timer is not init.");
		}
		
		long creation_date = TIME.currentTimeMillis();
		long refresh_interval = getRefreshInterval(metadata);
		long timeout_date = getTimeoutDate(creation_date,
				metadata.getVDOParams().getTimeout());
		
		// Create a task.
		VanishRefreshTask refresh_task =
			refresher_factory.createRefreshTask(vanish, enc_data_os, metadata,
					creation_date, timeout_date, refresh_interval);
		if (refresh_task.shouldStillBeRefreshingAtTime(
				creation_date + refresh_interval)) {
			tasks.add(refresh_task);
			timer.schedule(refresh_task, refresh_interval, refresh_interval);
		}  // else, no need to schedule refresh
	}
	
    // TODO(roxana): This check is pretty hacky: revise.
	private long getRefreshInterval(Metadata meta) {
		if (meta == null || !(meta instanceof EpochAwareMetadataImpl)) {
			return params.kRefreshInterval;
		} else {
			return ((EpochAwareMetadataImpl)meta).getEpochLength();
		}
	}

	private long getTimeoutDate(long creation_date, int timeout_h) {
		long timeout_ms = CONF.convertFromUnitToMillis(timeout_h,
				VanishConfiguration.TimeUnit.HOURS);
		return (creation_date + timeout_ms + params.kSlackTimeForRefresh);
	}

	@SuppressWarnings("unchecked")
	private static Vector<VanishRefreshTask> readTasksFromFile(File file) {
		Vector<VanishRefreshTask> read_tasks = null;

		if (file.exists()) {
			ObjectInputStream ois = null;
			try {
				ois = new ObjectInputStream(new FileInputStream(file));
				read_tasks = (Vector<VanishRefreshTask>)ois.readObject();
			} catch (Exception e) {
			} finally {
				if (ois != null) {
					try { ois.close(); }
					catch (IOException e) { }
				}
			}
		}
		
		return read_tasks;
	}
	
	private class SaveStateToDiskTask extends TimerTask {
		@Override
		public synchronized void run() {
			final LogStat logstat = LOG.logStat(
					"VanishRefreshTimer#SaveStateToDiskTask");
			ObjectOutputStream oos = null;
			try {
				oos = new ObjectOutputStream(new FileOutputStream(
						params.kSaveStateFile));
				
				// Create a snapshot of the tasks set.
				Vector<VanishRefreshTask> tasks_snapshot =
					getRefreshTaskSetSnapshot();

				// Write the snapshot to disk.
				oos.writeObject(tasks_snapshot);
			} catch (Exception e) {
			} finally {
				if (oos != null) {
					try { oos.close(); }
					catch (IOException e) { }
				}
				logstat.end();
			}
		}
	}
	
	@SuppressWarnings("unchecked")
	private Vector<VanishRefreshTask> getRefreshTaskSetSnapshot() {
		long now = TIME.currentTimeMillis();
		synchronized(tasks) {
			Iterator<VanishRefreshTask> it = tasks.iterator();
			while (it.hasNext()) {
				VanishRefreshTask t = it.next();
				if (! t.shouldStillBeRefreshingAtTime(now)) {  // cancel
					it.remove();
				}
			}
			return (Vector<VanishRefreshTask>)tasks.clone();
		}
	}
	
	// RefreshableConfig interface:
	
	// @Override
	public VanishRefreshParams getRenewedParamsFromConfig(
			VanishConfiguration config)
	throws InvalidArgumentVanishException {
		final long kRefreshInterval = config.getTimeInMillis(
				Defaults.CONF_VANISH_SYSTEM_REFRESH_PERIOD_H,
				Defaults.CONF_VANISH_SYSTEM_REFRESH_PERIOD_H_DEFAULT,
				VanishConfiguration.TimeUnit.HOURS);
		final long kSaveStateInterval = config.getTimeInMillis(
				Defaults.CONF_VANISH_SYSTEM_REFRESH_SAVESTATE_MIN,
				Defaults.CONF_VANISH_SYSTEM_REFRESH_SAVESTATE_MIN_DEFAULT,
				VanishConfiguration.TimeUnit.MINUTES);
		final long kSlackTimeForRefresh = config.getTimeInMillis(
				Defaults.CONF_VANISH_SYSTEM_REFRESH_SLACKTIME_MIN,
				Defaults.CONF_VANISH_SYSTEM_REFRESH_SLACKTIME_MIN_DEFAULT,
				VanishConfiguration.TimeUnit.MINUTES);
		final File kSaveStateFile = new File(
				FileUtil.getPath(config.getString(
						Defaults.CONF_VANISH_SYSTEM_STORAGE_DIR,
						Defaults.CONF_VANISH_SYSTEM_STORAGE_DIR_DEFAULT),
						SAVE_STATE_FILENAME));
		final File parent = kSaveStateFile.getParentFile();
		if (! parent.exists()) parent.mkdirs();
		
		VanishRefreshParams p = new VanishRefreshParams(
				kRefreshInterval, kSaveStateInterval, kSlackTimeForRefresh,
				kSaveStateFile);
		validate(p);
		
		return p;
	}
	
	private void validate(VanishRefreshParams p)
	throws InvalidArgumentVanishException {
		if (p.kSaveStateFile.exists() &&
			(! p.kSaveStateFile.isFile() || ! p.kSaveStateFile.canRead() ||
			 ! p.kSaveStateFile.canWrite())) {
			throw new InvalidArgumentVanishException(
				"State file exists but cannot be read/written: " +
				p.kSaveStateFile.getAbsolutePath());
		}
	}
	
	// @Override
	public void setParams(VanishRefreshParams params) {
		this.params = params;
	}
}
