/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.internal;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.VanishInternalInterface;
import edu.washington.cs.vanish.internal.impl.InputStreamWrapper;
import edu.washington.cs.vanish.internal.impl.OutputStreamWrapper;
import edu.washington.cs.vanish.internal.metadata.Metadata;


import junit.framework.TestCase;
/**
 * 
 * @author roxana
 *
 */
public abstract class
AbstractTestForVanishInternal<V extends VanishInternalInterface,
                              META extends Metadata>
extends TestCase {
	protected V vanish;  // protected to allow subclasses to use it directly.
	
	public AbstractTestForVanishInternal(V vanish) { this.vanish = vanish; }
	
    // TestCase functions:
	
	@Override
	protected void tearDown() { MockVanishBackend.getAllMappings().clear(); }
	
	// Helpful functions:
	
	protected abstract META instantiateMetadata(META other, long seed,
			                                    int num_shares,
			                                    int share_threshold)
	throws InvalidArgumentVanishException;
	
	protected META instantiateMetadata(long seed, int num_shares,
                                       int share_threshold)
	throws InvalidArgumentVanishException {
		return instantiateMetadata(null, seed, num_shares, share_threshold);
	}
	
	protected abstract void checkPutOutcome(META object_metadata,
			                                String value,
			                                byte[] enc_data);
	
	protected abstract void checkRefreshOutcome(META old_object_metadata,
			META new_object_metadata,
			byte[] refreshed_enc_data_can_be_null);
	
	protected abstract InputStreamWrapper wrapInputStream(InputStream is);
	protected abstract OutputStreamWrapper wrapOutputStream(OutputStream os);
	
	protected byte[] put(META object_metadata, String value) {
		ByteArrayInputStream is = new ByteArrayInputStream(value.getBytes());
		ByteArrayOutputStream os = new ByteArrayOutputStream();
		InputStreamWrapper clear_data_is = wrapInputStream(is);
		OutputStreamWrapper enc_data_os  = wrapOutputStream(os);

		try {
			vanish.encapsulate(clear_data_is, enc_data_os, object_metadata);
		} catch (VanishException e) {
			e.printStackTrace();
			fail("Failed to put into vanish.");
		}
		
		try {
			clear_data_is.close();
		} catch (IOException e) {
			e.printStackTrace();
			fail("Failed to close the data InputStream");
		}
		
		byte[] enc_data = os.toByteArray();
		
		try {
			enc_data_os.close();
		} catch (IOException e) {
			e.printStackTrace();
			fail("Failed to close the data OutputStream");
		}
		
		// Type-specific checks:
		checkPutOutcome(object_metadata, value, enc_data);
		
		return enc_data;
	}
	
	protected byte[] putUsingPrepushed(META object_metadata, String value) {
		ByteArrayInputStream is = new ByteArrayInputStream(value.getBytes());
		ByteArrayOutputStream os = new ByteArrayOutputStream();
		InputStreamWrapper clear_data_is = wrapInputStream(is);
		OutputStreamWrapper enc_data_os  = wrapOutputStream(os);

		int meta_hash = object_metadata.hashCode();
		try {
			vanish.encapsulateUsingPrepushedMetadata(clear_data_is,
					                                 enc_data_os,
					                                 object_metadata);
		} catch (VanishException e) {
			e.printStackTrace();
			fail("Failed to put into vanish.");
		}
		assertEquals(meta_hash, object_metadata.hashCode());
		
		try {
			clear_data_is.close();
		} catch (IOException e) {
			e.printStackTrace();
			fail("Failed to close the data InputStream");
		}
		
		byte[] enc_data = os.toByteArray();
		
		try {
			enc_data_os.close();
		} catch (IOException e) {
			e.printStackTrace();
			fail("Failed to close the data OutputStream");
		}
		
		// Type-specific checks:
		checkPutOutcome(object_metadata, value, enc_data);
		
		return enc_data;
	}
	
	protected byte[] refresh(byte[] encrypted_object,
			                 META old_object_metadata,
			                 META new_object_metadata,
			                 boolean exception_expected) {
		ByteArrayInputStream is = new ByteArrayInputStream(encrypted_object);
		ByteArrayOutputStream os = new ByteArrayOutputStream();
		InputStreamWrapper enc_data_is = wrapInputStream(is);
		OutputStreamWrapper refreshed_enc_data_os  = wrapOutputStream(os);		
		
		int old_meta_hash = old_object_metadata == null
		                    ? 0 : old_object_metadata.hashCode();
		try {
			vanish.refreshDeeply(enc_data_is, refreshed_enc_data_os,
					             old_object_metadata,
					             new_object_metadata);
			if (exception_expected) fail("Refresh succeeded when it " +
					                     "shouldn't have");
		} catch (VanishException e) {
			if (! exception_expected) {
				e.printStackTrace();
				fail("Failed to put into vanish.");
			}
		}
		
		// Check old_object_metadata was not mutated.
		if (old_object_metadata != null) {
			assertEquals(old_meta_hash, old_object_metadata.hashCode());
		}
		
		byte[] refreshed_encrypted_object = os.toByteArray();
		
		try {
			enc_data_is.close();
			refreshed_enc_data_os.close();
		} catch (IOException e) {
			e.printStackTrace();
			fail("Failed to close the data Input/OutputStream");
		}
		
        // Type-specific checks:
		checkRefreshOutcome(old_object_metadata, new_object_metadata,
				            refreshed_encrypted_object);
		
        // Lose the old shares forcibly.
		if (old_object_metadata.getLocationSeed() !=
			new_object_metadata.getLocationSeed()) {
			MockVanishBackend.loseShares(old_object_metadata.getLocationSeed(),
					old_object_metadata.getVDOParams().getNumShares());
		}
		
		return refreshed_encrypted_object;
	}
	
	protected void get(META object_metadata, byte[] encrypted_data,
			           String expected_return, boolean exception_expected) {
		ByteArrayInputStream is = new ByteArrayInputStream(encrypted_data);
		ByteArrayOutputStream os = new ByteArrayOutputStream();
		InputStreamWrapper enc_data_is = wrapInputStream(is);
		OutputStreamWrapper clear_data_os  = wrapOutputStream(os);

		int meta_hash = (object_metadata == null
		                ? 0 : object_metadata.hashCode());
		try {
			vanish.decapsulate(enc_data_is, clear_data_os, object_metadata);
			if (exception_expected) {
				fail("Succeeded in getting back value when we shouldn't have");
			}
			
			byte[] b = os.toByteArray();
			assertEquals(expected_return, new String(b));
		} catch (VanishException e) {
			if (!exception_expected) {  // error.
				e.printStackTrace();
				fail("Failed to get the object");
			} else {  // the exception was expected.
				assertEquals(expected_return, e.getMessage());
			}
		}
		
		if (object_metadata != null) {
			assertEquals(meta_hash, object_metadata.hashCode());  // const
		}
		
		try {
			enc_data_is.close();
			clear_data_os.close();
		} catch (IOException e) {
			e.printStackTrace();
			fail("Could not close Input/OutputStream");
		}
	}
	
	protected void refresh(META old_object_metadata,
			               String exception_message) {
		int old_meta_hash = old_object_metadata.hashCode();
		try {
			this.vanish.refreshReadOnly(old_object_metadata);
			if (exception_message != null) {
				fail("Refresh succeeded when it shouldn't have");
			}
		} catch (VanishException e) {
			if (exception_message == null) {
				e.printStackTrace();
				fail("Failed to refresh.");
			} else {
				assertEquals(exception_message, e.getMessage());
			}
		}
		assertEquals(old_meta_hash, old_object_metadata.hashCode());
		
		if (exception_message != null) return;  // finished checks for failed.
		
		// Type-specific checks:
		checkRefreshOutcome(old_object_metadata, null, null);
	}
	
	// Test cases:
	
	// Tests without share losses:
	
	public void testPutThenGetOriginalValue() {
		META object_metadata = null;
		try {
			object_metadata = instantiateMetadata(10L, 10, 7);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a put.
		byte[] encrypted_data = put(object_metadata, "my_value");
		assertNotNull(encrypted_data);
		
		// Do a get on the result.
		get(object_metadata, encrypted_data, "my_value", false);
	}
	
	public void testPutThenRefreshReadOnlySameMETAThenGet() {
		META object_metadata = null;
		try {
			object_metadata = instantiateMetadata(10L, 10, 7);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a put.
		byte[] encrypted_data = put(object_metadata, "my_value");
		assertNotNull(encrypted_data);
		
		// Do a refresh, using the same metadata params.
		refresh(object_metadata, null);
		
		// Do a get on the encrypted data and match its value to what we put.
		get(object_metadata, encrypted_data, "my_value", false);
	}
	
	public void testPutThenRefreshReadOnlyDifferentMETASameSeedThenGet() {
		META object_metadata = null;
		try {
			object_metadata = instantiateMetadata(10L, 10, 7);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a put.
		byte[] encrypted_data = put(object_metadata, "my_value");
		assertNotNull(encrypted_data);
		
		// Do a refresh.
		refresh(object_metadata, null);
		
		// Do a get on the encrypted data and match its value to what we put.
		get(object_metadata, encrypted_data, "my_value", false);
	}
	
	public void
	testPutThenRefreshReadOnlyDifferentMETADifferentSeedThenGet() {
		META object_metadata = null;
		try {
			object_metadata = instantiateMetadata(10L, 10, 7);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a put.
		byte[] encrypted_data = put(object_metadata, "my_value");
		assertNotNull(encrypted_data);
		
		// Do a refresh, using a different seed.
		refresh(object_metadata, null);
		
		// A get using the new metadata can get to them.
		get(object_metadata, encrypted_data, "my_value", false);
	}
	
	public void testPutThenRefreshUpdateThenGet() {
		META object_metadata = null;
		try {
			object_metadata = instantiateMetadata(10L, 10, 7);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a put.
		byte[] encrypted_data = put(object_metadata, "my_value");
		assertNotNull(encrypted_data);
		
		// Do a refresh.
		// Do not construct the new metadata from the old one, as we want to
		// check that the new one will be updated by refresh.
		META new_object_metadata = null;
		try {
			new_object_metadata = instantiateMetadata(11L, 10, 7);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		byte[] refreshed_encrypted_data = refresh(encrypted_data,
				                                  object_metadata,
				                                  new_object_metadata,
				                                  false);
		
		// Do a get on the encrypted data and match its value to what we put.
		get(new_object_metadata, refreshed_encrypted_data, "my_value", false);
	}
	
	public void testEncapsulateUsingPrepushedMetadata() {
		META object_metadata = null;
		try {
			object_metadata = instantiateMetadata(10L, 10, 7);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		put(object_metadata, "");
		
		byte[] encrypted_data = putUsingPrepushed(object_metadata, "my_value");
		
		// Do a get:
		get(object_metadata, encrypted_data, "my_value", false);
	}
	
	// Tests with share losses:
	
	public void testPutThenLoseKeysThenGetFails() {
		META object_metadata = null;
		try {
			object_metadata = instantiateMetadata(10L, 10, 9);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a put.
		byte[] encrypted_data = put(object_metadata, "my_value");
		assertNotNull(encrypted_data);
		
		// Lose some shares.
		MockVanishBackend.loseShares(10L, 2);
		
		// Do a get on the result.
		get(object_metadata, encrypted_data, "Got 8/9 shares", true);
	}
	
	public void testPutThenLoseKeysThenRefreshFails() {
		META object_metadata = null;
		try {
			object_metadata = instantiateMetadata(10L, 10, 9);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a put.
		byte[] encrypted_data = put(object_metadata, "my_value");
		assertNotNull(encrypted_data);
		
		// Lose some shares.
		MockVanishBackend.loseShares(10L, 2);
		
		// Do a refresh on the result.
		refresh(object_metadata, "Got 8/9 shares");
	}
	
	public void testPutThenLoseTooFewKeysThenGetSucceeds() {
		META object_metadata = null;
		try {
			object_metadata = instantiateMetadata(10L, 10, 8);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a put.
		byte[] encrypted_data = put(object_metadata, "my_value");
		assertNotNull(encrypted_data);
		
		// Lose some shares.
		MockVanishBackend.loseShares(10L, 2);
		
		// Do a get on the result.
		get(object_metadata, encrypted_data, "my_value", false);
	}
	
	public void testPutThenLoseTooFewKeysThenRefreshSucceeds() {
		META object_metadata = null;
		try {
			object_metadata = instantiateMetadata(10L, 10, 8);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a put.
		byte[] encrypted_data = put(object_metadata, "my_value");
		assertNotNull(encrypted_data);
		
        // Lose some shares.
		MockVanishBackend.loseShares(10L, 1);
		
		// Do a refresh.
		refresh(object_metadata, null);
		
		// Do a get on the encrypted data and match its value to what we put.
		get(object_metadata, encrypted_data, "my_value", false);
	}
	
	// Tests with corrupted metadata:
	
	public void testPutThenGetWithCorruptedMETABadSeed() {
		META object_metadata = null;
		try {
			object_metadata = instantiateMetadata(10L, 10, 9);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a put.
		byte[] encrypted_data = put(object_metadata, "my_value");
		assertNotNull(encrypted_data);

		META corrupted_metadata = null;
		try {
			corrupted_metadata = instantiateMetadata(object_metadata, 25L, 10, 9);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a get on the result.
		get(corrupted_metadata, encrypted_data, "Got 0/9 shares", true);
	}
	
	public void testPutThenGetWithCorruptedMETAWrongSSSParams() {
		META object_metadata = null;
		try {
			object_metadata = instantiateMetadata(10L, 10, 9);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a put.
		byte[] encrypted_data = put(object_metadata, "my_value");
		assertNotNull(encrypted_data);

		META corrupted_metadata = null;
		try {
			corrupted_metadata = instantiateMetadata(object_metadata, 10L, 9, 7);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		
		// Do a get on the result.
		ByteArrayInputStream bais = new ByteArrayInputStream(encrypted_data);
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		try {
			vanish.decapsulate(wrapInputStream(bais), wrapOutputStream(baos),
				corrupted_metadata);
		} catch (VanishException e) { }  // this is fine, as well.
		
		try {
			bais.close();
		} catch (IOException e) {
			e.printStackTrace();
			fail("Could not close InputStream.");
		}
		
		// TODO(my_value): Unfortunately, we don't get an error when we use
		// corrupt metadata. Rather, we get an empty InputStream. Should fix?
		String decrypted_value = new String(baos.toByteArray());
		assertFalse("my_value".equals(decrypted_value));  // check it's junk.
		assertFalse(decrypted_value.contains("my_value"));
	}
}
