/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.internal.crypto.impl;
/**
 * A class for testing the ShamirSecretSharing class by generating random
 * secrets, encoding them, removing some keys, decoding them, and comparing
 * the results to the original secrets.
 *  
 * @author  Andrew Morgan
 */

import java.security.SecureRandom;
import java.util.*;

import junit.framework.TestCase;

import org.junit.Test;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.crypto.impl.ShamirSecretSharing;
import edu.washington.cs.vanish.logging.VanishLogger;
/**
 * 
 * @author Andy Morgan
 *
 */
public class ShamirSecretSharingTest extends TestCase {
    private static final int NUM_TESTS = 100;
    private static final int MIN_CHARS = 500;
    private static final int MAX_CHARS = 2000;
    private static final int MIN_SHARES = 5;
    private static final int MAX_SHARES = 40;
    
    /**
     * Runs many tests (as described above).
     */
    @Test
    public void testMultiple() throws Exception {
        int[] res = new int[3];
        for(int i = 1; i <= NUM_TESTS; i++){
            res[testAll()]++;  // 0 = pass, 2 = fail.
        }
        System.out.println("\n::RESULTS::");
        System.out.println("Pass = " + res[0]);
        System.out.println("Fail = " + res[2]);
        if(res[2] > 0){
            fail();  // fail if any tests failed.
        } else {
            System.out.println("All tests passed!");
        }
    }
    
    /**
     * Runs a single test (as described above).
     */
    public int testAll() throws Exception {
        Random r = new Random();
        byte[] secret = new byte[r.nextInt(MAX_CHARS - MIN_CHARS + 1) +
                                 MIN_CHARS];
        r.nextBytes(secret);  // generate secret
        if(secret[0] == -1 || secret[0] == 0){  // remove invalid values
            secret[0] = 67;
        }
        int num_total_shares = r.nextInt(MAX_SHARES - MIN_SHARES + 1) +
                               MIN_SHARES;
        int threshold = r.nextInt(num_total_shares - 1) + 2;
        ShamirSecretSharing instance = new ShamirSecretSharing(
        		new SecureRandom(),
        		new VanishConfiguration(),
        		VanishLogger.getLogger("test"));
        byte[][] result = instance.split(secret, num_total_shares, threshold);
        for(int i = 0; i < num_total_shares - threshold; i++){
            int j = r.nextInt(result.length);
            result[j] = null;  // remove shares
        }
        byte[] done = instance.compose(result, threshold);  // recompose
        for(int i = 0; i < done.length; i++){  // check that result matches
            if(done[i] != secret[i]){
                // If no match, fail and print error message.
                double pct = (double) i / done.length * 100;
                System.out.print("\nERROR: % complete = " 
                        + pct + "(" + i + "/" + done.length + " chars)");
                for(int j = 0; j < secret.length; j++){
                    System.out.print(secret[j] + " ");
                }
                return 2;  // fail
            }
        }
        return 0;  // pass
    }
}