/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.internal.impl;

import java.security.SecureRandom;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Random;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.backend.TooManySharesLostVanishException;
import edu.washington.cs.vanish.internal.backend.VanishBackendException;
import edu.washington.cs.vanish.internal.backend.VanishBackendInterface;
import edu.washington.cs.vanish.internal.crypto.impl.ShamirSecretSharing;
import edu.washington.cs.vanish.internal.crypto.impl.SymmetricEncDecImpl;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.internal.metadata.impl.BasicMetadataImpl;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.stats.VanishStats;

/**
 * 
 * @author roxana
 */
public class BasicVanishHybridTest extends BasicVanishTest {
	private static final Map<String, VanishBackendInterface> kMockBackends =
		new HashMap<String, VanishBackendInterface>();
	
	private static final HybridMockVanishBackend backend1 =
		new HybridMockVanishBackend("test1");
	private static final HybridMockVanishBackend backend2 =
		new HybridMockVanishBackend("test2");
	
	static {
		kMockBackends.put("test1", backend1);
		kMockBackends.put("test2", backend2);
	}
	
	public BasicVanishHybridTest() throws Exception {
		super(new BasicVanishImpl(
				kMockBackends,
				new SymmetricEncDecImpl(new SecureRandom(),
						                new VanishConfiguration(),
						                VanishLogger.getLogger("test")),
				new ShamirSecretSharing(new SecureRandom(),
						new VanishConfiguration(),
		                VanishLogger.getLogger("test")),
				VanishLogger.getLogger("test")));
	}
	
	@Override
	public void setUp() {
		backend1.getAllMappings().clear();
		backend2.getAllMappings().clear();
	}
	
	@Override
	protected BasicMetadataImpl instantiateMetadata(
			@SuppressWarnings("unused") BasicMetadataImpl metadata,
			long location_seed,
			int num_shares,
			int share_threshold)
	throws InvalidArgumentVanishException {
		return new BasicMetadataImpl(
				new VDOParams(num_shares, share_threshold,
						      VDOParams.kDefaultTimeout_h,
						      VDOParams.kEncryptionAlgo,
						      VDOParams.kEncryptionKeyLength,
						      VDOParams.kEncryptionMode,
						      new String[] { "test1", "test2" }),
				location_seed);
	}
	
	@Override
	public void testPutThenLoseKeysThenGetFails() {
		BasicMetadataImpl object_metadata = null;
		try {
			object_metadata = instantiateMetadata(10L, 10, 9);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a put.
		byte[] encrypted_data = put(object_metadata, "my_value");
		assertNotNull(encrypted_data);
		
		// Lose some shares from one of them.
		backend1.loseShares(2);
		
		System.out.println(backend1.getAllMappings().size());
		
		// Do a get on the result.
		get(object_metadata, encrypted_data, "Got 8/9 shares", true);
	}
	
	public void testPutThenLoseKeysThenGetFailsBackend2() {
		BasicMetadataImpl object_metadata = null;
		try {
			object_metadata = instantiateMetadata(10L, 10, 9);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a put.
		byte[] encrypted_data = put(object_metadata, "my_value");
		assertNotNull(encrypted_data);
		
		// Lose some shares from one of them.
		backend2.loseShares(2);
		
		// Do a get on the result.
		get(object_metadata, encrypted_data, "Got 8/9 shares", true);
	}
	
	@Override
	public void testPutThenLoseTooFewKeysThenGetSucceeds() {
		BasicMetadataImpl object_metadata = null;
		try {
			object_metadata = instantiateMetadata(10L, 10, 8);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a put.
		byte[] encrypted_data = put(object_metadata, "my_value");
		assertNotNull(encrypted_data);
		
		// Lose some shares.
		backend1.loseShares(2);
		backend2.loseShares(2);
		
		// Do a get on the result.
		get(object_metadata, encrypted_data, "my_value", false);
	}
	
	@Override
	public void testPutThenLoseKeysThenRefreshFails() {
		// No need to check this.
		// (I think...)
	}
}

class HybridMockVanishBackend implements VanishBackendInterface {
	private Map<String, byte[]> storage;
	private boolean is_initialized;
	private String id; 
	
	public HybridMockVanishBackend(String name) {
		storage = Collections.synchronizedMap(new HashMap<String, byte[]>());
		this.id = name;
	}

	// VanishBackend interface:

	public byte[] getID() { return null; }

	public byte[][] getShares(byte[][] indexes, int needed_keys)
	throws TooManySharesLostVanishException, VanishBackendException {
		byte[][] shares = new byte[indexes.length][];
		int i = 0;
		int num_retrieved_shares = 0;
		for (byte[] index: indexes) {
			String index_str = new String(index);
			byte[] share =  storage.get(index_str);
			shares[i++] = share;
			if (share != null) ++num_retrieved_shares;
		}
		if (num_retrieved_shares < needed_keys) {
			throw new TooManySharesLostVanishException("Got " +
					num_retrieved_shares + "/" + needed_keys + " shares");
		}
		return shares;
	}

	public VanishStats getStats() { return null; }  // TODO(roxana): Impl mock.

	public void init() throws VanishBackendException {
		this.is_initialized = true;
	}

	public void pushShares(byte[][] shares, byte[][] locations)
	throws VanishBackendException {
		int i = 0;
		for (byte[] index: locations) {
			storage.put(new String(index), shares[i++]);
		}
	}

	public byte[] generateLocation(Random prg) {
		return generateLocationRandomly(prg);
	}
	
	private static byte[] generateLocationRandomly(Random prg) {
		String s = "k" + prg.nextInt();
		return s.getBytes();
	}
	
	public Map<String, byte[]> getAllMappings() { return storage; }
	
	public void loseShares(int how_many) {
		Iterator it = storage.entrySet().iterator();
		for (int i = 0; i < how_many; ++i) {
			it.next();
			it.remove();
		}
	}

	public boolean isInitialized() { return is_initialized; }

	public void stop() throws VanishException { }

	public String getIdentifier() {
		return id;
	}
}

