/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.internal.impl;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.OutputStream;
import java.security.SecureRandom;
import java.util.HashMap;
import java.util.Map;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.AbstractTestForVanishInternal;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.MockVanishBackend;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.backend.VanishBackendInterface;
import edu.washington.cs.vanish.internal.crypto.impl.ShamirSecretSharing;
import edu.washington.cs.vanish.internal.crypto.impl.SymmetricEncDecImpl;
import edu.washington.cs.vanish.internal.impl.InputStreamWrapper;
import edu.washington.cs.vanish.internal.impl.MetadataEncapsulatedIndirectKeyVanishImpl;
import edu.washington.cs.vanish.internal.impl.OutputStreamWrapper;
import edu.washington.cs.vanish.internal.metadata.Metadata;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.internal.metadata.impl.IndirectKeyMetadataImpl;
import edu.washington.cs.vanish.logging.VanishLogger;
/**
 * 
 * @author roxana
 *
 */
public class MetadataEncapsulatedIndirectKeyVanishImplTest
extends AbstractTestForVanishInternal
		<MetadataEncapsulatedIndirectKeyVanishImpl, IndirectKeyMetadataImpl> {
	private static final Map<String, VanishBackendInterface> kMockBackends =
		new HashMap<String, VanishBackendInterface>();
	static {
		kMockBackends.put("test", new MockVanishBackend());
	}
	
	public MetadataEncapsulatedIndirectKeyVanishImplTest()
	throws VanishException {
		super(new MetadataEncapsulatedIndirectKeyVanishImpl(
					kMockBackends,
					new SymmetricEncDecImpl(new SecureRandom(),
                		                new VanishConfiguration(),
                		                VanishLogger.getLogger("test")),
                    new ShamirSecretSharing(new SecureRandom(),
                	  					    new VanishConfiguration(),
                				            VanishLogger.getLogger("test")),
                    VanishLogger.getLogger("test")));
	}
	
	protected MetadataEncapsulatedIndirectKeyVanishImplTest(
			MetadataEncapsulatedIndirectKeyVanishImpl vanish)
	throws VanishException {
		super(vanish);
	}

	// AbstractTestForVanish abstract functions:
	
	@Override
	protected void get(IndirectKeyMetadataImpl object_metadata,
			byte[] encrypted_data, String expected_return,
			boolean exception_expected) {
		super.get(null, encrypted_data, expected_return, exception_expected);
	}
	
	@Override
	protected void checkPutOutcome(IndirectKeyMetadataImpl object_metadata,
			                       String value,
			                       byte[] enc_data) {
		try {
			ObjectInputStream ois = new ObjectInputStream(
					new ByteArrayInputStream(enc_data));
			try {
				Metadata m = (Metadata)ois.readObject();
				assertNotNull(m);
			} catch (ClassCastException e) {
				fail("Wrong metadata object type.");
			} catch (ClassNotFoundException e) {
				fail("Wrong metadata object type.");
			}
			ois.close();
		} catch (IOException e) {
			e.printStackTrace();
			fail("Unable to read the metadata from the encrypted object.");
		}
		
		assertNotNull(object_metadata.getEncryptedDataKey());
	}

	@Override
	protected void checkRefreshOutcome(
			IndirectKeyMetadataImpl old_object_metadata,
			IndirectKeyMetadataImpl new_object_metadata,
			byte[] refreshed_enc_data) {
		if (refreshed_enc_data != null) {
			try {
				ObjectInputStream ois = new ObjectInputStream(
						new ByteArrayInputStream(refreshed_enc_data));
				try {
					Metadata m = (Metadata)ois.readObject();
					assertNotNull(m);
				} catch (ClassCastException e) {
					fail("Wrong metadata object type.");
				} catch (ClassNotFoundException e) {
					fail("Wrong metadata object type.");
				}
				ois.close();
			} catch (IOException e) {
				e.printStackTrace();
				fail("Unable to read the metadata from the encrypted object.");
			}
		}
		
		assertNotNull(old_object_metadata.getEncryptedDataKey());
		if (new_object_metadata != null) {
			assertNotNull(new_object_metadata.getEncryptedDataKey());
		}
	}

	@Override
	protected IndirectKeyMetadataImpl instantiateMetadata(
			IndirectKeyMetadataImpl other,
			long location_seed,
			int num_shares,
			int share_threshold)
	throws InvalidArgumentVanishException {
		IndirectKeyMetadataImpl meta = new IndirectKeyMetadataImpl(
				new VDOParams(num_shares, share_threshold,
					      VDOParams.kDefaultTimeout_h,
					      VDOParams.kEncryptionAlgo,
					      VDOParams.kEncryptionKeyLength,
					      VDOParams.kEncryptionMode,
					      new String[] {"test"}),
				location_seed);
		if (other != null) {
			meta.setEncryptedDataKey(other.getEncryptedDataKey());
		}
		return meta;
	}

	@Override
	protected InputStreamWrapper wrapInputStream(InputStream is) {
		return new InputStreamWrapper(is);
	}

	@Override
	protected OutputStreamWrapper wrapOutputStream(OutputStream os) {
		return new OutputStreamWrapper(os);
	}

	// Replace some of the AbstractTestForVanish corrupted metadata cases
	// with successful tests.
	// Even if the given metadata object is corrupted, I should still be
	// able to get the object, as the given metadata is actually overriden
	// by the in-value serialized metadata object.

	@Override
	public void testPutThenGetWithCorruptedMETABadSeed() {
		IndirectKeyMetadataImpl object_metadata = null;
		try {
			object_metadata = instantiateMetadata(10L, 10, 9);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}

		// Do a put.
		byte[] encrypted_data = put(object_metadata, "my_value");
		assertNotNull(encrypted_data);

		IndirectKeyMetadataImpl corrupted_metadata = null;
		try {
			corrupted_metadata = instantiateMetadata(object_metadata,
					25L, 10, 9);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a get on the result.
		get(corrupted_metadata, encrypted_data, "my_value", false);
	}
	
	@Override
	public void testPutThenGetWithCorruptedMETAWrongSSSParams() {
		IndirectKeyMetadataImpl object_metadata = null;
		try {
			object_metadata = instantiateMetadata(10L, 10, 9);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a put.
		byte[] encrypted_data = put(object_metadata, "my_value");
		assertNotNull(encrypted_data);
		
		IndirectKeyMetadataImpl corrupted_metadata = null;
		try {
			corrupted_metadata = instantiateMetadata(object_metadata,
					10L, 9, 7);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("FAiled to instantiate metadata.");
		}
		
		// Do a get on the result.
		super.get(corrupted_metadata, encrypted_data, "my_value", false);
	}
}
