/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.service.impl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.security.SecureRandom;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.MockVanishBackend;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.backend.VanishBackendInterface;
import edu.washington.cs.vanish.internal.crypto.impl.ShamirSecretSharing;
import edu.washington.cs.vanish.internal.crypto.impl.SymmetricEncDecImpl;
import edu.washington.cs.vanish.internal.impl.MetadataEncapsulatedIndirectKeyVanishImpl;
import edu.washington.cs.vanish.internal.metadata.Metadata;
import edu.washington.cs.vanish.internal.metadata.MetadataFactory;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.internal.metadata.impl.IndirectKeyMetadataImpl;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.stats.Stat;
import edu.washington.cs.vanish.stats.StatsCollector;
import edu.washington.cs.vanish.system.VanishFactory;
import edu.washington.cs.vanish.system.VanishInterface;
import edu.washington.cs.vanish.system.VanishRefreshTask;
import edu.washington.cs.vanish.system.VanishRefreshTaskFactory;
import edu.washington.cs.vanish.system.impl.ReadOnlyVanishRefreshTask;
import edu.washington.cs.vanish.system.impl.VanishImpl;
import edu.washington.cs.vanish.util.Defaults;
import edu.washington.cs.vanish.util.VanishTime;
import junit.framework.TestCase;
/**
 * 
 * @author roxana
 */
public class VanishServiceHandlerTest extends TestCase {
	protected VanishServiceHandler vanish_rpc_handler;
	static {
		try {
			VanishConfiguration conf = new VanishConfiguration();
			conf.setProperty(Defaults.CONF_VANISH_SYSTEM_PREPUSH_NUMVDOS, 0);
			conf.setProperty(Defaults.CONF_VANISH_SERVICE_SHOULD_GZIP, true);
			VanishServiceHandler.init(new TestVanishFactory(conf),
					new TestStatsCollector(),
					conf, VanishLogger.getLogger("test"));
		} catch (VanishException e) {
			e.printStackTrace();
			fail("Failed to initialize.");
		}
	}
	
	public VanishServiceHandlerTest() {
		vanish_rpc_handler = new VanishServiceHandler();
	}
	
	@Override
	public void tearDown() {
		MockVanishBackend.getAllMappings().clear();
		File enc_file_fd = new File("/tmp/vanish.test.enc");
		enc_file_fd.setWritable(true);
	}
	
	private Map<String, Object> wrapVDOParams(int num_shares, int threshold,
			                                  int timeout) {
		VDOParams params = null;
		try {
			params = instantiateVDOParams(num_shares, threshold, timeout);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("Failed to create metadata");
		}
		return params.toMap();
	}
	
	// Remote interface test cases:

	public void testEncapsulateThenDecapsulate() {
		try {
			byte[] output = vanish_rpc_handler.encapsulate(null,
					"my value".getBytes(), wrapVDOParams(10, 7, 8));
			assertNotNull(output);
			
			byte[] original = vanish_rpc_handler.decapsulate(null, output);
			assertNotNull(original);
			assertEquals("my value", new String(original));
		} catch (VanishException e) {
			e.printStackTrace();
			fail("Failed.");
		}
	}
	
	public void testEncThenRefreshDeeply() {
		try {
			byte[] output = vanish_rpc_handler.encapsulate(null,
					"my value".getBytes(), wrapVDOParams(10, 7, 8));
			assertNotNull(output);

			Map<String, Object> w = wrapVDOParams(20, 10, 8);
			byte[] refreshed = vanish_rpc_handler.refreshDeeply(null, output,
					w);
			assertNotNull(refreshed);
			
			// The refresh has added 20 new shares.
			assertEquals(30, MockVanishBackend.getAllMappings().size());
			
			byte[] original = vanish_rpc_handler.decapsulate(null, output);
			assertNotNull(original);
			assertEquals("my value", new String(original));
		} catch (VanishException e) {
			e.printStackTrace();
			fail();
		}
	}
	
	public void testEncThenRefreshReadOnly() {
		try {
			byte[] output = vanish_rpc_handler.encapsulate(null,
					"my value".getBytes(), wrapVDOParams(10, 7, 8));
			assertNotNull(output);
			
			Iterator<String> it =
				MockVanishBackend.getAllMappings().keySet().iterator();
			it.next();
			it.remove();
			assertEquals(9, MockVanishBackend.getAllMappings().size());
			
			vanish_rpc_handler.refreshReadOnly(null, output);
			
			// We're back at 10 shares.
			assertEquals(10, MockVanishBackend.getAllMappings().size());
		} catch (VanishException e) {
			e.printStackTrace();
			fail();
		}
	}
	
	// Try out a few corner cases:
	
	public void testEncapsulateThenDecapsulateEmptyString() {
		try {
			byte[] output = vanish_rpc_handler.encapsulate(null, "".getBytes(),
					wrapVDOParams(10, 7, 8));
			assertNotNull(output);
			
			byte[] original = vanish_rpc_handler.decapsulate(null, output);
			assertNotNull(original);
			assertEquals("", new String(original));
		} catch (VanishException e) {
			e.printStackTrace();
			fail();
		}
	}
	
	public void testEncapsulateFailsForNullValue() {
		try {
			vanish_rpc_handler.encapsulate(null, null, wrapVDOParams(10, 7, 8));
			fail("Succeeded when it shouldn't have.");
		} catch (VanishException e) { }  // expected.
	}
	
	public void testEncThenDecThenEncThenDec() {
		try {
			byte[] output = vanish_rpc_handler.encapsulate(null,
					"my value".getBytes(), wrapVDOParams(10, 7, 8));
			assertNotNull(output);
			
			byte[] original = vanish_rpc_handler.decapsulate(null, output);
			assertNotNull(original);
			assertEquals("my value", new String(original));
			
			output = vanish_rpc_handler.encapsulate(null, original,
					wrapVDOParams(10, 7, 8));
			assertNotNull(output);
			
			original = vanish_rpc_handler.decapsulate(null, output);
			assertNotNull(original);
			assertEquals("my value", new String(original));
		} catch (VanishException e) {
			e.printStackTrace();
			fail("Failed.");
		}
	}
	
	// Local interface test cases:
	
	public void testEncapsulateThenDecapsulateFile() {
		// TODO(roxana): Modify for Win/Mac!
		String clear_file = "/tmp/vanish.test.dat";
		String enc_file   = "/tmp/vanish.test.enc";
		String dec_file   = "/tmp/vanish.test.dec";
		
		// Write something in the file.
		FileOutputStream fos;
		try {
			fos = new FileOutputStream(clear_file);
			fos.write("my value".getBytes());
			fos.close();
		} catch (Exception e1) {
			e1.printStackTrace();
			fail("Could not generate file");
		}
		
		try {
			vanish_rpc_handler.encapsulate(null, clear_file, enc_file,
					instantiateVDOParams(10, 7, 8).toMap());
			vanish_rpc_handler.decapsulate(null, enc_file, dec_file);
		} catch (VanishException e) {
			e.printStackTrace();
			fail("Failed to encapsulate");
		}
		
		FileInputStream fis = null;
		try {
			fis = new FileInputStream(dec_file);
		} catch (FileNotFoundException e) {
			fail("Failed to read back the data");
		}
		byte[] buf = new byte[8];
		try {
			assertEquals(8, fis.read(buf));
			assertEquals(-1, fis.read());
			fis.close();
		} catch (IOException e) {
			e.printStackTrace();
			fail("Failed to read file.");
		}
		
		assertEquals("my value", new String(buf));
	}
	
	private VDOParams instantiateVDOParams(int num_shares, int threshold,
			int timeout_h)
	throws InvalidArgumentVanishException {
		return new VDOParams(num_shares, threshold, timeout_h,
				VDOParams.kEncryptionAlgo, VDOParams.kEncryptionKeyLength,
				VDOParams.kEncryptionMode, new String[]{ "test" });
	} 
	
	public void testEncThenRefreshDeeplyFile() {
		String clear_file = "/tmp/vanish.test.dat";
		String enc_file   = "/tmp/vanish.test.enc";
		String dec_file   = "/tmp/vanish.test.dec";
		String ref_file  = "/tmp/vanish.test.refresh";
		File enc_file_fd = new File(enc_file);
		enc_file_fd.setWritable(true);
		
		// Write something in the file.
		FileOutputStream fos;
		try {
			fos = new FileOutputStream(clear_file);
			fos.write("my value".getBytes());
			fos.close();
		} catch (Exception e1) {
			e1.printStackTrace();
			fail("Could not generate file");
		}
		
		try {
			vanish_rpc_handler.encapsulate(null, clear_file, enc_file,
					instantiateVDOParams(10, 7, 8).toMap());
			vanish_rpc_handler.refreshDeeply(null, enc_file, ref_file,
					instantiateVDOParams(20, 15, 8).toMap());
			vanish_rpc_handler.decapsulate(null, ref_file, dec_file);
		} catch (VanishException e) {
			e.printStackTrace();
			fail("Failed to encapsulate/refresh/decapsulate");
		}
		
		// The refresh has added 20 new shares.
		assertEquals(30, MockVanishBackend.getAllMappings().size());
		
		FileInputStream fis = null;
		try {
			fis = new FileInputStream(dec_file);
		} catch (FileNotFoundException e) {
			fail("Failed to read back the data");
		}
		byte[] buf = new byte[8];
		try {
			assertEquals(8, fis.read(buf));
			assertEquals(-1, fis.read());
			fis.close();
		} catch (IOException e) {
			e.printStackTrace();
			fail("Failed to read file.");
		}
		
		assertEquals("my value", new String(buf));
	}
	
	public void testEncThenRefreshReadOnlyFile() {
		String clear_file = "/tmp/vanish.test.dat";
		String enc_file   = "/tmp/vanish.test.enc";
		String dec_file   = "/tmp/vanish.test.dec";
		
		// Write something in the file.
		FileOutputStream fos;
		try {
			fos = new FileOutputStream(clear_file);
			fos.write("my value".getBytes());
			fos.close();
		} catch (Exception e1) {
			e1.printStackTrace();
			fail("Could not generate file");
		}
		
		File enc_file_fd = new File(enc_file);
		try {
			vanish_rpc_handler.encapsulate(null, clear_file, enc_file,
					instantiateVDOParams(10, 7, 8).toMap());
			
			enc_file_fd.setWritable(false);  // mark it r/o s.t. we ensure
			    // we get an error if refresh modifies it.
			
			vanish_rpc_handler.refreshReadOnly(null, enc_file);
			
			vanish_rpc_handler.decapsulate(null, enc_file, dec_file);
		} catch (VanishException e) {
			e.printStackTrace();
			enc_file_fd.setWritable(true);
			fail("Failed to encapsulate");
		}

		assertEquals(10, MockVanishBackend.getAllMappings().size());
		
		FileInputStream fis = null;
		try {
			fis = new FileInputStream(dec_file);
		} catch (FileNotFoundException e) {
			fail("Failed to read back the data");
		}
		byte[] buf = new byte[8];
		try {
			assertEquals(8, fis.read(buf));
			assertEquals(-1, fis.read());
			fis.close();
		} catch (IOException e) {
			e.printStackTrace();
			fail("Failed to read file.");
		}
		
		assertEquals("my value", new String(buf));
	}
}

// Helper classes:

class TestVanishFactory implements VanishFactory {
	private VanishInterface vanish_interface;
	private VanishConfiguration conf;

	public TestVanishFactory(VanishConfiguration conf) {
		this.conf = conf;
	}
	
	public VanishInterface getInitializedVanishSystem()
	throws VanishException {
		if (vanish_interface == null) {
			final Map<String, VanishBackendInterface>
			kMockBackends = new HashMap<String, VanishBackendInterface>();
				kMockBackends.put("test", new MockVanishBackend());
			
			VanishLogger logger = VanishLogger.getLogger("test");
			SecureRandom random = new SecureRandom();
			vanish_interface = new VanishImpl(
				new MetadataEncapsulatedIndirectKeyVanishImpl(
						kMockBackends,
						new SymmetricEncDecImpl(random,
								conf, logger),
						new ShamirSecretSharing(new SecureRandom(),
								new VanishConfiguration(),
						        VanishLogger.getLogger("test")),
						logger),
				new TestMetadataFactory(), 
				new TestVanishRefreshTaskFactory(),
				random,
				new VanishTime(),
				conf, logger);
			vanish_interface.init();
		}
		return vanish_interface;
	}
}

class TestMetadataFactory implements MetadataFactory {
    // @Override
	public Metadata createMetadata(VDOParams params, long seed) {
		return new IndirectKeyMetadataImpl(params, seed);
	}
}

class TestVanishRefreshTaskFactory implements VanishRefreshTaskFactory {
	// @Override
	public VanishRefreshTask createRefreshTask(
			VanishInterface vanish_system_ptr,
			OutputStream encrypted_data_os,
			Metadata metadata,
			long creation_date,
			long timeout_date,
			long running_interval) {
		return new ReadOnlyVanishRefreshTask(vanish_system_ptr, metadata,
				creation_date, timeout_date, running_interval);
	}
}

class TestStatsCollector implements StatsCollector {
	StringBuffer stats;
	
	public TestStatsCollector() {
		stats = new StringBuffer();
	}
	
	public ReportLevel getCurrentReportLevel() {
		return ReportLevel.SENSITIVE;
	}

	public void report(String client_id, String stat_message, Stat stat,
			           ReportLevel level) {
		stats.append(client_id + ";" + stat_message + ";" + stat + "\n");
	}

	public void init() throws VanishException { }

	public boolean isInitialized() { return true; }

	public void stop() throws VanishException { }

	public void report(Stat stat) {
		stats.append(stat.checkpoint(getCurrentReportLevel()) + "\n");
	}

	public void report(Stat stat, Throwable t)
	throws IllegalServerStateVanishException {
		report(stat);
	}
}
