/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.stats.impl;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;

import org.apache.log4j.Layout;
import org.apache.log4j.WriterAppender;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.service.impl.IllegalServerStateVanishException;
import edu.washington.cs.vanish.stats.Stat;
import edu.washington.cs.vanish.stats.StatUploader;
import edu.washington.cs.vanish.stats.StatsCollector;
import edu.washington.cs.vanish.stats.StatsCollector.ReportLevel;
import edu.washington.cs.vanish.util.Defaults;
import edu.washington.cs.vanish.util.IOVanishException;
import edu.washington.cs.vanish.util.VanishTime;
import junit.framework.TestCase;

/**
 * 
 * @author roxana
 *
 */

class MockUploader implements StatUploader {
	protected File uploaded_file;
	
	// @Override
	public void upload(File file) throws IOVanishException {
		uploaded_file = file;
	}

	// @Override
	public void init() throws VanishException { }

	// @Override
	public boolean isInitialized() { return true; }

	// @Override
	public void stop() throws VanishException { }	
}

class MockStatsCollectorImpl extends StatsCollectorImpl {
	private TestAppender appender;
	
	public MockStatsCollectorImpl(StatUploader uploader, VanishTime time,
			                      VanishConfiguration conf)
	throws InvalidArgumentVanishException, IOException {
		super(time, uploader, conf, VanishLogger.getLogger("test"));
	}
	
	class TestAppender extends WriterAppender {
		private ByteArrayOutputStream log_os;
		
		public TestAppender(Layout layout, ByteArrayOutputStream os) {
			super(layout, os);
			this.log_os = os;
		}

		// @Override
		public InputStream getLogInputStream() {
			byte[] data_so_far = log_os.toByteArray();
			return new ByteArrayInputStream(data_so_far);
		}

		public void truncate() throws IOException {
			log_os.reset();
		}
	}
	
	@Override
	protected WriterAppender createAppenderUsingLayout(Layout layout) {
		appender = new TestAppender(layout, new ByteArrayOutputStream());
		return appender;
	}
	
	/*@Override
	protected void submitStatsToServer(InputStream submitted_is)
	throws IOVanishException {
		InputStream received_is = submitted_is;
		if (shouldZip()) {
			try {
				received_is = new GZIPInputStream(submitted_is);
			} catch (IOException e) {
				throw new IOVanishException("Failed to gunzip", e);
			}
		}
		
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		FileUtil.writeInputOutput(received_is, baos);
		
		last_submitted_stats = new String(baos.toByteArray());
		
		try {
			baos.close();
			received_is.close();
		} catch (IOException e) {
			throw new IOVanishException("Failed to close streams", e);
		}
	}*/
	
	public void restart() throws IOException {
		appender.truncate();
	}
	
	public String getOutput() {
		byte[] bytes = appender.log_os.toByteArray();
		return new String(bytes);
	}
	
	// Factory:
	
	public static StatsCollector stats_collector;

	public static StatsCollector getStatsCollector(StatUploader uploader,
			final VanishConfiguration conf, final VanishTime time)
	throws InvalidArgumentVanishException, IOException {
		if (stats_collector == null) {
			stats_collector = new MockStatsCollectorImpl(uploader, time, conf);
			try {
				stats_collector.init();
			} catch (VanishException e) {
				e.printStackTrace();
			}
		}
		return stats_collector;
	}
}

class TestStat implements Stat {
	protected String value;
	protected ReportLevel level;

	public TestStat(String value, ReportLevel level) {
		this.value = value;
		this.level = level;
	}

	// @Override
	public String checkpoint(ReportLevel reporting_level) {
		if (reporting_level.compareTo(this.level) >= 0) {
			return value;
		} else {
			return null;
		}
	}
}

class FakeTime extends VanishTime {
	@Override
	public long currentTimeMillis() { return 10; }
}

public class StatsCollectorImplTest extends TestCase {
	@SuppressWarnings("deprecation")
	static VanishConfiguration conf = Defaults.CONF;
	static FakeTime fake_time = new FakeTime();
	
	static MockUploader uploader;
	static MockStatsCollectorImpl collector;
	
	static {
		uploader = new MockUploader();
		try {
			collector = (MockStatsCollectorImpl)MockStatsCollectorImpl
			.getStatsCollector(uploader, conf, fake_time);
		} catch (Exception e) {
			e.printStackTrace();
			fail("Failed to construct collector");
		}
	}
	
	public void tearDown() {
		try {
			collector.restart();
		} catch (IOException e) {
			e.printStackTrace();
			fail("Failed to restart the collector.");
		}
	}

	@SuppressWarnings("deprecation")
	public void testOneInsensitiveOneSensitiveLogsOnlyInsensitive() {
		conf.setProperty(Defaults.CONF_VANISH_STATS_ENABLED,
		                 ReportLevel.INSENSITIVE.name());
		conf.setProperty(Defaults.CONF_VANISH_STATS_BULK_UPLOAD_SIZE, 1024);
		try {
			collector.setParams(collector.getRenewedParamsFromConfig(conf));
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("Failed to refresh configuration");
		}
		
		TestStat my_stat = new TestStat("insensitive_value",
				                        ReportLevel.INSENSITIVE);
		try {
			collector.report(my_stat);
		} catch (IllegalServerStateVanishException e) {
			e.printStackTrace();
			fail("Failed to report");
		}
		
		my_stat = new TestStat("sensitive_value", ReportLevel.SENSITIVE);
		try {
			collector.report(my_stat);
		} catch (IllegalServerStateVanishException e) {
			e.printStackTrace();
			fail("Failed to report");
		}

		String logged_entries = collector.getOutput();
		assertNotNull(logged_entries);

		assertEquals("ts=10;insensitive_value\n", logged_entries);
	}
	
	@SuppressWarnings("deprecation")
	public void testOneInsensitiveOneSensitiveLogsSensitive() {		
		conf.setProperty(Defaults.CONF_VANISH_STATS_ENABLED,
                ReportLevel.SENSITIVE.name());
		conf.setProperty(Defaults.CONF_VANISH_STATS_BULK_UPLOAD_SIZE, 1024);
		try {
			collector.setParams(collector.getRenewedParamsFromConfig(conf));
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("Failed to refresh configuration");
		}

		TestStat my_stat = new TestStat("insensitive_value",
				                        ReportLevel.INSENSITIVE);
		try {
			collector.report(my_stat);
		} catch (IllegalServerStateVanishException e) {
			e.printStackTrace();
			fail("Failed to report");
		}
		
		my_stat = new TestStat("sensitive_value", ReportLevel.SENSITIVE);
		try {
			collector.report(my_stat);
		} catch (IllegalServerStateVanishException e) {
			e.printStackTrace();
			fail("Failed to report");
		}
		
		// Check HERE:
		
		String logged_entries = collector.getOutput();
		assertNotNull(logged_entries);
		
		assertEquals("ts=10;insensitive_value\n" +
				     "ts=10;sensitive_value\n",
			         logged_entries);
	}

	@SuppressWarnings("deprecation")
	public void testOneInsensitiveOneSensitiveLogsNone() {
		conf.setProperty(Defaults.CONF_VANISH_STATS_ENABLED,
                         ReportLevel.NONE.name());
		conf.setProperty(Defaults.CONF_VANISH_STATS_BULK_UPLOAD_SIZE, 1024);
		try {
			collector.setParams(collector.getRenewedParamsFromConfig(conf));
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("Failed to refresh configuration");
		}

		TestStat my_stat = new TestStat("insensitive_value",
				                        ReportLevel.INSENSITIVE);
		try {
			collector.report(my_stat);
		} catch (IllegalServerStateVanishException e) {
			e.printStackTrace();
			fail("Failed to report");
		}

		my_stat = new TestStat("sensitive_value", ReportLevel.SENSITIVE);
		try {
			collector.report(my_stat);
		} catch (IllegalServerStateVanishException e) {
			e.printStackTrace();
			fail("Failed to report");
		}

		String logged_entries = collector.getOutput();
		assertNotNull(logged_entries);

		assertEquals("", logged_entries);
	}
	
	@SuppressWarnings("deprecation")
	public void
	testOneInsensitiveOneSensitiveChangesFromInsensitiveToSensitive() {
		conf.setProperty(Defaults.CONF_VANISH_STATS_ENABLED,
                ReportLevel.INSENSITIVE.name());
		conf.setProperty(Defaults.CONF_VANISH_STATS_BULK_UPLOAD_SIZE, 1024);
		try {
			collector.setParams(collector.getRenewedParamsFromConfig(conf));
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("Failed to refresh configuration");
		}

		TestStat my_stat = new TestStat("insensitive_value",
				                        ReportLevel.INSENSITIVE);
		try {
			collector.report(my_stat);
		} catch (IllegalServerStateVanishException e) {
			e.printStackTrace();
			fail("Failed to report");
		}
		
		my_stat = new TestStat("sensitive_value", ReportLevel.SENSITIVE);
		try {
			collector.report(my_stat);
		} catch (IllegalServerStateVanishException e) {
			e.printStackTrace();
			fail("Failed to report");
		}
		
		// Turn the level to sensitive.
		
		conf.setProperty(Defaults.CONF_VANISH_STATS_ENABLED,
                ReportLevel.SENSITIVE.name());
		try {
			collector.setParams(collector.getRenewedParamsFromConfig(conf));
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("Failed to refresh configuration");
		}
		
		my_stat = new TestStat("sensitive_value2", ReportLevel.SENSITIVE);
		try {
			collector.report(my_stat);
		} catch (IllegalServerStateVanishException e) {
			e.printStackTrace();
			fail("Failed to report");
		}
		
		String logged_entries = collector.getOutput();
		assertNotNull(logged_entries);
		
		assertEquals("ts=10;insensitive_value\n" +
				     "ts=10;sensitive_value2\n",
			         logged_entries);
	}
}
