/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.system;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.impl.InputStreamWrapper;
import edu.washington.cs.vanish.internal.impl.OutputStreamWrapper;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.system.VanishInterface;
import edu.washington.cs.vanish.util.Defaults;
import junit.framework.TestCase;

// TODO(roxana): We don't test prepushes now... Must do that.
/**
 * 
 * @author roxana
 *
 */
public abstract class AbstractTestForVanishInterface extends TestCase {
	protected static final VanishConfiguration CONF =
		new VanishConfiguration() {
			@Override
			public long convertFromUnitToMillis(int value, TimeUnit unit) {
				return value;
			}
	};
	
	protected static final int kEpochLength = 200;  // ms
	
	static {
		CONF.setProperty(Defaults.CONF_VANISH_SYSTEM_PREPUSH_NUMVDOS, 0);
		CONF.setProperty(Defaults.CONF_VANISH_SYSTEM_REFRESH_SLACKTIME_MIN, 0);
		CONF.setProperty(Defaults.CONF_VANISH_SYSTEM_REFRESH_PERIOD_H,
				kEpochLength);
		
		CONF.setProperty(Defaults.CONF_VANISH_SYSTEM_REFRESH_SAVESTATE_MIN,
				         -1);
		    // Treated by the fake time module as milliseconds actually
		    // instead of hours.
	}
	
	protected VanishInterface vanish;
	
	public AbstractTestForVanishInterface(VanishInterface vanish) {
		this.vanish = vanish;
	}
	
	@Override
	protected void setUp() {
		try {
			vanish.init();
		} catch (VanishException e) {
			e.printStackTrace();
			fail("Failed to init vanish.");
		}
	}
	
	@Override
	protected void tearDown() {
		try {
			vanish.stop();
		} catch (VanishException e) {
			e.printStackTrace();
			fail("Failed to stop vanish.");
		}
	}
	
	// Helper functions:
	
	protected abstract void checkPutOutcome(byte[] encrypted_data,
			                                long timeout_h);
	
	protected byte[] put(String value, int num_shares, int threshold,
			             int timeout_ms) {
		ByteArrayInputStream is = new ByteArrayInputStream(value.getBytes());
		ByteArrayOutputStream os = new ByteArrayOutputStream();
		InputStreamWrapper clear_data_is = new InputStreamWrapper(is);
		OutputStreamWrapper enc_data_os  = new OutputStreamWrapper(os);

		try {
			vanish.encapsulate(clear_data_is, enc_data_os,
					new VDOParams(num_shares, threshold, timeout_ms,
						      VDOParams.kEncryptionAlgo,
						      VDOParams.kEncryptionKeyLength,
						      VDOParams.kEncryptionMode,
						      new String[] {"test"}));
		} catch (VanishException e) {
			e.printStackTrace();
			fail("Failed to put into vanish.");
		}
		
		try {
			clear_data_is.close();
		} catch (IOException e) {
			e.printStackTrace();
			fail("Failed to close the data InputStream");
		}
		
		byte[] enc_data = os.toByteArray();
		
		try {
			enc_data_os.close();
		} catch (IOException e) {
			e.printStackTrace();
			fail("Failed to close the data OutputStream");
		}
		
		return enc_data;
	}
	
	protected void get(byte[] encrypted_data, String expected_return,
			           boolean exception_expected) {
		ByteArrayInputStream is = new ByteArrayInputStream(encrypted_data);
		ByteArrayOutputStream os = new ByteArrayOutputStream();
		InputStreamWrapper clear_data_is = new InputStreamWrapper(is);
		OutputStreamWrapper enc_data_os  = new OutputStreamWrapper(os);

		try {
			vanish.decapsulate(clear_data_is, enc_data_os);
			if (exception_expected) {
				fail("Succeeded in getting back value when we shouldn't have");
			}
			
			byte[] b = os.toByteArray();
			assertEquals(expected_return, new String(b));
		} catch (VanishException e) {
			if (! exception_expected) {  // error.
				e.printStackTrace();
				fail("Failed to get the object.");
			} else {  // the exception was expected.
				assertEquals(expected_return, e.getMessage());
			}
		}
		
		try {
			clear_data_is.close();
		} catch (IOException e) {
			e.printStackTrace();
			fail("Failed to close the data InputStream");
		}
		
		try {
			enc_data_os.close();
		} catch (IOException e) {
			e.printStackTrace();
			fail("Failed to close the data OutputStream");
		}
	}
	
	// Test cases:
	
	public void testEncapsulateThenDecapsulate() {  // 1-epoch timeout
		byte[] encrypted_data = put("my_value", 10, 7, kEpochLength);
		get(encrypted_data, "my_value", false);
		checkPutOutcome(encrypted_data, kEpochLength);
	}

	public void testEncapsulate9HourTimeout() {  // still 1-epoch timeout
		byte[] encrypted_data = put("my_value", 10, 7, kEpochLength + 20);
		get(encrypted_data, "my_value", false);
		checkPutOutcome(encrypted_data, kEpochLength + 20);
	}
	
	public void testEncapsulate15HourTimeout() {  // still 1-epoch timeout
		byte[] encrypted_data = put("my_value", 10, 7, 2 * kEpochLength - 20);
		get(encrypted_data, "my_value", false);
		checkPutOutcome(encrypted_data, 2 * kEpochLength - 20);
	}
	
	public void testEncapsulate16HourTimeout() {  // 2-epoch timeout
		byte[] encrypted_data = put("my_value", 10, 7, 2 * kEpochLength + 20);  
		get(encrypted_data, "my_value", false);
		checkPutOutcome(encrypted_data, 2 * kEpochLength + 20);
	}

	public void testEncapsulate24HourTimeout() {  // 3-epoch timeout
		byte[] encrypted_data = put("my_value", 10, 7, 3 * kEpochLength + 20);
		get(encrypted_data, "my_value", false);
		checkPutOutcome(encrypted_data, 3 * kEpochLength + 20);
	}
}
