/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.system.impl;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Timer;

import junit.framework.TestCase;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.internal.metadata.impl.BasicMetadataImpl;
import edu.washington.cs.vanish.system.VanishInterface;
import edu.washington.cs.vanish.system.VanishRefreshTask;
/**
 * 
 * @author roxana
 *
 */
public class ReadOnlyVanishRefreshTaskTest extends TestCase {
	@Override
	protected void tearDown() { }

	// Test cases:

	public void testRefreshTests16hTimeout() {
		long creation_time = 0; 
		TestVanishRefreshTask task = new TestVanishRefreshTask(
				creation_time, creation_time + 16, 8);
		assertEquals(creation_time + 8,  task.getNextScheduledRefreshTime());
		assertEquals(true,  task.shouldStillBeRefreshingAtTime(creation_time + 0));
		
		assertEquals(false, task.shouldRefreshNow(creation_time + 7));
		assertEquals(true, task.shouldStillBeRefreshingAtTime(creation_time + 7));
		
		assertEquals(true,  task.shouldRefreshNow(creation_time + 8));
		assertEquals(true, task.shouldStillBeRefreshingAtTime(creation_time + 8));

		assertEquals(false, task.shouldRefreshNow(creation_time + 9));
		assertEquals(false, task.shouldStillBeRefreshingAtTime(creation_time + 9));

		assertEquals(false, task.shouldRefreshNow(creation_time + 25));
		assertEquals(false, task.shouldStillBeRefreshingAtTime(creation_time + 25));
	}

	public void testRefreshTests8hTimeout() {
		long creation_time = 0;
		TestVanishRefreshTask task = new TestVanishRefreshTask(
				creation_time, creation_time + 8, 8);

		assertEquals(false,  task.shouldRefreshNow(creation_time + 0));
		assertEquals(true, task.shouldStillBeRefreshingAtTime(creation_time + 0));
		
		assertEquals(false, task.shouldRefreshNow(creation_time + 1));
		assertEquals(false, task.shouldStillBeRefreshingAtTime(creation_time + 1));
	}

	public void testRefreshTestsNoRefreshes() {
		long creation_time = 0;
		TestVanishRefreshTask task = new TestVanishRefreshTask(
				creation_time, creation_time + 8, 9);
		assertEquals(false,  task.shouldRefreshNow(creation_time + 0));
		assertEquals(false, task.shouldStillBeRefreshingAtTime(creation_time + 0));
	}
	
	// Running the actual task.
	
	public void testRunRefreshesOnceFor16hTimeout() {
		Timer t = new Timer();
		
		long creation_date = System.currentTimeMillis();
		TestVanishRefreshTask task = new TestVanishRefreshTask(
				creation_date, creation_date + 170, 80);  // all in ms to be
		                                                  // fast
		t.schedule(task, 80, 80);
		
		try { Thread.sleep(200); }
		catch (InterruptedException e) { e.printStackTrace(); }
		
		assertEquals(1, task.num_refreshes);
	}
	
	public void testRunRefreshesTwiceFor24hTimeout() {
		Timer t = new Timer();
		
		long creation_date = System.currentTimeMillis();
		TestVanishRefreshTask task = new TestVanishRefreshTask(
				creation_date, creation_date + 270, 80);
		t.schedule(task, 80, 80);
		
		try { Thread.sleep(500); }
		catch (InterruptedException e) { e.printStackTrace(); }
		
		assertTrue(1 <= task.num_refreshes);
	}
	
	public void testRunRefreshesNoneFor8hTimeout() {
		Timer t = new Timer();
		
		long creation_date = System.currentTimeMillis();
		TestVanishRefreshTask task = new TestVanishRefreshTask(
				creation_date, creation_date + 90, 80);
		t.schedule(task, 80, 80);
		
		try { Thread.sleep(200); }
		catch (InterruptedException e) { e.printStackTrace(); }
		
		assertEquals(0, task.num_refreshes);
	}
	
	// Serialization tests:
	
	public void testSerializationDeserialization() {
		long creation_date = System.currentTimeMillis();
		
		VDOParams vdo_params = null;
		try {
			vdo_params = new VDOParams(2, 1, 80);
		} catch (InvalidArgumentVanishException e1) {
			e1.printStackTrace();
			fail("FAiled to create vdo params");
		}
		
		ReadOnlyVanishRefreshTask task = new ReadOnlyVanishRefreshTask(null,
				new BasicMetadataImpl(vdo_params, 4),
				creation_date, creation_date + 80, 80);

		byte[] serialized_task = null; 
		ObjectOutputStream oos = null;
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			oos = new ObjectOutputStream(baos);
			oos.writeObject(task);
			serialized_task = baos.toByteArray();
		} catch (IOException e) {
			e.printStackTrace();
			fail("Failed to serialize");
		} finally {
			if (oos != null) {
				try { oos.close(); }
				catch (IOException e) { fail("Failed to close OS"); }
			}
		}

		assertNotNull(serialized_task);
		
		ObjectInputStream ois = null;
		try {
			ois = new ObjectInputStream(new ByteArrayInputStream(
					serialized_task));
			ReadOnlyVanishRefreshTask read_task =
				(ReadOnlyVanishRefreshTask)ois.readObject();
			
			assertNotNull(read_task);
			assertEquals(task, read_task);
		} catch (Exception e) {
			e.printStackTrace();
			fail("Failed to read task.");
		} finally {
			try { ois.close(); }
			catch (IOException e) { fail("Failed to close IS"); }
		}
	}
	
	// Helpers:
	
	private static class TestVanishRefreshTask extends VanishRefreshTask {
		private static final long serialVersionUID = -5967901489687014359L;

		private int num_refreshes;
		
		public TestVanishRefreshTask(long creation_date, long timeout_date,  
				                     long running_interval) {
			super(null, creation_date, timeout_date, running_interval);
		}

		@Override
		protected void refresh(VanishInterface vanish_system) {
			this.num_refreshes++;
		}

		@Override
		protected long getCurrentTime() { return System.currentTimeMillis(); }
	}
}