/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.system.impl;

import java.io.OutputStream;
import java.security.SecureRandom;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.MetadataEncapsulatedVanishInternalInterface;
import edu.washington.cs.vanish.internal.MockVanishBackend;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.backend.VanishBackendInterface;
import edu.washington.cs.vanish.internal.crypto.SecretSharingInterface;
import edu.washington.cs.vanish.internal.crypto.SymmetricEncDec;
import edu.washington.cs.vanish.internal.crypto.impl.ShamirSecretSharing;
import edu.washington.cs.vanish.internal.crypto.impl.SymmetricEncDecImpl;
import edu.washington.cs.vanish.internal.impl.MetadataEncapsulatedIndirectKeyVanishImpl;
import edu.washington.cs.vanish.internal.metadata.Metadata;
import edu.washington.cs.vanish.internal.metadata.MetadataFactory;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.internal.metadata.impl.EpochAwareMetadataImpl;
import edu.washington.cs.vanish.internal.metadata.impl.IndirectKeyMetadataImpl;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.system.AbstractTestForVanishInterface;
import edu.washington.cs.vanish.system.VanishInterface;
import edu.washington.cs.vanish.system.VanishRefreshTask;
import edu.washington.cs.vanish.system.VanishRefreshTaskFactory;
import edu.washington.cs.vanish.util.Immutable;
import edu.washington.cs.vanish.util.VanishTime;

/**
 * 
 * @author roxana
 */
public class VanishImplTest extends AbstractTestForVanishInterface {
	private static final Map<String, VanishBackendInterface> kMockBackends =
		new HashMap<String, VanishBackendInterface>();
	static {
		kMockBackends.put("test", new MockVanishBackend());
	}
	
	public VanishImplTest() throws Exception {
		super(new TestVanishImpl(
					new TestMetadataEncapsulatedIndirectKeyVanish(
						kMockBackends,
						new SymmetricEncDecImpl(new SecureRandom(),
								CONF,
								VanishLogger.getLogger("test")),
						new ShamirSecretSharing(new SecureRandom(),
								new VanishConfiguration(),
						        VanishLogger.getLogger("test"))),
					new TestMetadataFactory(),
					new TestVanishRefreshTaskFactory(),
					new VanishTime(),
					CONF));
	}

	@Override 
	protected void tearDown() {
		TestMetadataEncapsulatedIndirectKeyVanish.num_refreshes_so_far = 0;
		try {
			this.vanish.stop();
		} catch (VanishException e) {
			e.printStackTrace();
			fail("Failed to stop Vanish");
		}
	}

	@Override
	protected void checkPutOutcome(byte[] encrypted_data, long timeout_h) {		
		try { Thread.sleep(timeout_h + 1000); }
		catch (InterruptedException e) { }
		
		long num_refreshes = timeout_h / kEpochLength - 1;
		
		assertEquals(num_refreshes,
			TestMetadataEncapsulatedIndirectKeyVanish.num_refreshes_so_far);
	}
	
	static class TestMetadataFactory implements MetadataFactory {
		public Metadata createMetadata(VDOParams params, long seed) {
			return new EpochAwareMetadataImpl(
					new IndirectKeyMetadataImpl(params, seed), kEpochLength);
		}
	}
}

class TestVanishImpl extends VanishImpl {
	public TestVanishImpl(
		MetadataEncapsulatedVanishInternalInterface vanish_core,
		MetadataFactory metadata_factory,
		VanishRefreshTaskFactory refresher_factory,
		VanishTime time,
		VanishConfiguration conf)
	throws InvalidArgumentVanishException {
		super(vanish_core, metadata_factory, refresher_factory, new Random(),
			  time, conf, VanishLogger.getLogger("test"));
	}
}

class TestMetadataEncapsulatedIndirectKeyVanish
extends MetadataEncapsulatedIndirectKeyVanishImpl {
	static int num_refreshes_so_far;

	public TestMetadataEncapsulatedIndirectKeyVanish(
			Map<String, VanishBackendInterface> backend_module,
			SymmetricEncDec crypto_module,
			SecretSharingInterface sss_module) throws VanishException {
		super(backend_module, crypto_module, sss_module,
			  VanishLogger.getLogger("test"));
	}
	
	@Override
	public void refreshReadOnly(@Immutable Metadata metadata)
	throws VanishException {
		++num_refreshes_so_far;
		super.refreshReadOnly(metadata);
	}
}

class TestVanishRefreshTaskFactory implements VanishRefreshTaskFactory {
	public VanishRefreshTask createRefreshTask(
			VanishInterface vanish_system_ptr, OutputStream encrypted_data_os,
			Metadata metadata,
			long creation_date,
			long timeout_date,
			long running_interval) {
		return new TestReadOnlyVanishRefreshTask(vanish_system_ptr, metadata,
				creation_date, timeout_date, running_interval);
	}
}

class TestReadOnlyVanishRefreshTask extends ReadOnlyVanishRefreshTask {
	private static final long serialVersionUID = -6576585734930541464L;

	protected static int num_refreshes = 0;

	public TestReadOnlyVanishRefreshTask(
			VanishInterface vanish_system_ptr,
			Metadata metadata,
			long creation_date, long timeout_date, long running_interval) {
		super(vanish_system_ptr, metadata, creation_date, timeout_date,
			  running_interval);
	}

	@Override
	protected void refresh(VanishInterface vanish_system) {
		++num_refreshes;
		super.refresh(vanish_system);
	}
}
