/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

/**
 * 
 */
package edu.washington.cs.vanish.system.impl;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

import junit.framework.TestCase;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.impl.InputStreamWrapper;
import edu.washington.cs.vanish.internal.impl.OutputStreamWrapper;
import edu.washington.cs.vanish.internal.metadata.Metadata;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.internal.metadata.impl.BasicMetadataImpl;
import edu.washington.cs.vanish.util.Constants;


/**
 * @author levya
 * 
 */
public class VanishMockImplTest extends TestCase {
	private VDOParams wrapVDOParams(int num_shares, int threshold) {
		VDOParams params = null;
		try {
			params = new VDOParams(num_shares, threshold, 8);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("Failed to create vdo params");
		}
		
		return params;
	}

	public void testDecapsulate() throws Exception {
		long seed = System.currentTimeMillis();
		VanishMockImpl vanish = new VanishMockImpl();
		vanish.seedTtls.put(seed, 8 * Constants.HOURS);

		ByteArrayOutputStream clear_data_os = new ByteArrayOutputStream();
		ByteArrayInputStream byteStream = new ByteArrayInputStream(
				generateEncryptedByteArrayFrom(
						new BasicMetadataImpl(wrapVDOParams(10, 5), seed),
						"Hello World"));
		vanish.decapsulate(new InputStreamWrapper(byteStream),
				           new OutputStreamWrapper(clear_data_os));

		assertEquals("Hello World", clear_data_os.toString());
	}
	
	public void testDecapsulateExpire() throws Exception {
		try {
			long seed = System.currentTimeMillis();
			VanishMockImpl vanish = new VanishMockImpl();
			vanish.seedTtls.put(seed, 0L);
			
			ByteArrayOutputStream clear_data_os = new ByteArrayOutputStream();
			ByteArrayInputStream byteStream = new ByteArrayInputStream(
					generateEncryptedByteArrayFrom(
							new BasicMetadataImpl(wrapVDOParams(10, 5), seed),
							"Hello World"));
			vanish.decapsulate(new InputStreamWrapper(byteStream),
					           new OutputStreamWrapper(clear_data_os));
			fail("Should not be able to decrypt expired message");
		} catch (VanishException e) {
			assertEquals("Could not decrypt message", e.getMessage());
		}
	}

	private byte[] generateEncryptedByteArrayFrom(
			BasicMetadataImpl basicMetadataImpl, String msg) throws IOException {
		ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
		ObjectOutputStream objStream = new ObjectOutputStream(byteStream);
		objStream.writeObject(basicMetadataImpl);
		byteStream.write((VanishMockImpl.ENCRYPTION_STRING + msg).getBytes());
		return byteStream.toByteArray();
	}

	public void testEncapsulate() throws Exception {
		VanishMockImpl vanish = new VanishMockImpl();
		ByteArrayOutputStream enc_data_os = new ByteArrayOutputStream();
		vanish.encapsulate(
				new InputStreamWrapper(
						new ByteArrayInputStream("Hello World".getBytes())),
				new OutputStreamWrapper(enc_data_os),
				wrapVDOParams(10, 5));
		ByteArrayInputStream stream = new ByteArrayInputStream(enc_data_os
				.toByteArray());
		ObjectInputStream objectStream = new ObjectInputStream(stream);
		Metadata metadata = (Metadata) objectStream.readObject();

		assertEquals(10, metadata.getVDOParams().getNumShares());
		assertEquals(5, metadata.getVDOParams().getThreshold());

		StringBuilder sb = new StringBuilder();
		int b;
		while ((b = stream.read()) >= 0) {
			sb.append((char) b);
		}
		assertEquals(VanishMockImpl.ENCRYPTION_STRING + "Hello World", sb
				.toString());
	}

	public void testRefreshDeeply() throws Exception {
		VanishMockImpl vanish = new VanishMockImpl();

		int oldSeed = 1234;
		ByteArrayInputStream byteStream = new ByteArrayInputStream(
				generateEncryptedByteArrayFrom(
						new BasicMetadataImpl(wrapVDOParams(10, 5), oldSeed),
						VanishMockImpl.ENCRYPTION_STRING + "Hello World"));
		ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
		Metadata metadata = vanish
				.refreshDeeply(new InputStreamWrapper(byteStream),
						new OutputStreamWrapper(outputStream),
						new VDOParams(20, 10, 100));
		
		assertEquals(20, metadata.getVDOParams().getNumShares());
		assertEquals(10, metadata.getVDOParams().getThreshold());
		assertNotSame(oldSeed, metadata.getLocationSeed());
		
		assertEquals(100 * Constants.HOURS,
				     (long)vanish.seedTtls.get(metadata.getLocationSeed()));
	}

	public void testRefreshReadOnly() throws Exception {
		VanishMockImpl vanish = new VanishMockImpl();
		vanish.seedTtls.put(1234L, 0L);
		
		vanish.refreshReadOnly(new BasicMetadataImpl(wrapVDOParams(10, 5),
				1234L));
		assertEquals(8 * Constants.HOURS, (long)vanish.seedTtls.get(1234L));
	}
}
