/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.system.impl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Vector;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.metadata.Metadata;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.internal.metadata.impl.BasicMetadataImpl;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.util.Defaults;
import edu.washington.cs.vanish.util.VanishTime;

import junit.framework.TestCase;

/**
 * 
 * @author roxana
 */
public class VanishRefreshTimerTest extends TestCase {
	private static final VanishConfiguration CONF = new VanishConfiguration() {
		@Override
		public long convertFromUnitToMillis(int value, TimeUnit unit) {
			return value;
		}
	};
	
	private static final int kEpochLength = 300;
	static {
		CONF.setProperty(Defaults.CONF_VANISH_SYSTEM_PREPUSH_NUMVDOS, 0);
		CONF.setProperty(Defaults.CONF_VANISH_SYSTEM_REFRESH_SLACKTIME_MIN, 0);
		CONF.setProperty(Defaults.CONF_VANISH_SYSTEM_REFRESH_PERIOD_H,
				         kEpochLength);
		CONF.setProperty(Defaults.CONF_VANISH_SYSTEM_REFRESH_SAVESTATE_MIN,
				-1);  // invalidate for default tests.
	}
	
	private TestVanishRefreshTaskFactory refresh_task_factory;
		
	public VanishRefreshTimerTest() {
		refresh_task_factory = new TestVanishRefreshTaskFactory();
	}
	

	@Override
	protected void tearDown() {
		TestReadOnlyVanishRefreshTask.num_refreshes = 0;
	}
	
	private VDOParams instantiateVDOParams(int num_shares,
			int threshold, int timeout_h) {
		VDOParams params = null;
		try {
			params = new VDOParams(num_shares, threshold, timeout_h);
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("Failed to instantiate vdo params");
		}
		return params;
	}
	
	private VanishRefreshTimer createAndInitializeTimer() {
		VanishRefreshTimer timer = null;
		try {
			timer = new VanishRefreshTimer(new VanishMockImpl(),
					refresh_task_factory, new VanishTime(), CONF,
					VanishLogger.getLogger("test"));
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("Failed to create timer");
		}
		
		File f = timer.getParams().kSaveStateFile;
		if (f != null) f.delete();

		// Check that timer doesn't accept anything before init.
		Metadata meta = new BasicMetadataImpl(instantiateVDOParams(2, 1, 30),
				4);
		try {
			timer.scheduleRefreshesIfNeeded(meta, null);
			fail("Timer scheduled task without having been initialized.");
		} catch (VanishException e) {
			assertEquals("Refresh timer is not init.", e.getMessage());
		}
		
		timer.init();
		
		return timer;
	}
	
	private void stopTimer(VanishRefreshTimer timer) {
		timer.stop();
	}
	
	// Tests:
	
	// Timer-related tests:

	public void testDoesntRefreshFor8HourTimeout() {
		VanishRefreshTimer timer = createAndInitializeTimer();

		Metadata meta = new BasicMetadataImpl(
				instantiateVDOParams(2, 1, kEpochLength), 4);
		try {
			timer.scheduleRefreshesIfNeeded(meta, null);
		} catch (VanishException e) {
			e.printStackTrace();
			fail("Failed to schedule.");
		}
		
		try { Thread.sleep(2 * kEpochLength); }
		catch (InterruptedException e) { }
		
		assertEquals(0, TestReadOnlyVanishRefreshTask.num_refreshes);
		
		stopTimer(timer);
	}
	
	public void testRefreshesOnceFor16HourTimeout() {
		VanishRefreshTimer timer = createAndInitializeTimer();
		Metadata meta = new BasicMetadataImpl(
				instantiateVDOParams(2, 1, 2 * kEpochLength + 20), 4);
		try {
			timer.scheduleRefreshesIfNeeded(meta, null);
		} catch (VanishException e) {
			e.printStackTrace();
			fail("Failed to schedule.");
		}
		
		try { Thread.sleep(kEpochLength + 10); }
		catch (InterruptedException e) { }
		
		assertEquals(1, TestReadOnlyVanishRefreshTask.num_refreshes);
		
		try { Thread.sleep(kEpochLength + 10); }
		catch (InterruptedException e) { }
		
		assertEquals(1, TestReadOnlyVanishRefreshTask.num_refreshes);
		
		stopTimer(timer);
	}
	
	// Fault-tolerance tests:
	
	public void testLoadsFromDiskCorrectly() {		
		Vector<ReadOnlyVanishRefreshTask> task_vector =
			new Vector<ReadOnlyVanishRefreshTask>();

		long now = System.currentTimeMillis();
		task_vector.add(new TestReadOnlyVanishRefreshTask(
				null,
				new BasicMetadataImpl(
						instantiateVDOParams(2, 1, 5 * kEpochLength), 0),
				now,
				now + 5 * kEpochLength + 20L,
				kEpochLength));

		VanishRefreshTimer timer = null;
		try {
			timer = new VanishRefreshTimer(new VanishMockImpl(),
					refresh_task_factory, new VanishTime(), CONF,
					VanishLogger.getLogger("test"));
		} catch (InvalidArgumentVanishException e) {
			e.printStackTrace();
			fail("Failed to create timer");
		}

		File f = timer.getParams().kSaveStateFile;
		File parent = f.getParentFile();
		if (!parent.exists()) parent.mkdirs();

		ObjectOutputStream oos = null;
		try {
			oos = new ObjectOutputStream(new FileOutputStream(f));
			oos.writeObject(task_vector);
		} catch (IOException e1) {
			e1.printStackTrace();
			fail("Failed to write state.");
		} finally {
			if (oos != null) {
				try { oos.close(); }
				catch (IOException e) { fail("Failed to close OS"); }
			}
		}

		timer.init();

		try { Thread.sleep(3000); }
		catch (InterruptedException e) { }

		f.delete();

		assertEquals(4, TestReadOnlyVanishRefreshTask.num_refreshes);

		stopTimer(timer);
	}

	// Must be last, as it overrides the CONF value.
	
	@SuppressWarnings({ "unchecked", "deprecation" })
	public void testSerializesToDiskCorrectly() {
		CONF.setProperty(Defaults.CONF_VANISH_SYSTEM_REFRESH_SAVESTATE_MIN,
				         kEpochLength - 10);
		    // The above are treated as ms by the fake time.
	
		Metadata meta = new BasicMetadataImpl(
				instantiateVDOParams(2, 1, 10 * kEpochLength), 0);
		
		VanishRefreshTimer timer = createAndInitializeTimer();
		try {
			timer.scheduleRefreshesIfNeeded(meta, null);
		} catch (VanishException e) {
			e.printStackTrace();
			fail("Failed to schedule.");
		}
		
		try { Thread.sleep(2 * kEpochLength); }
		catch (InterruptedException e) { }

		File f = timer.getParams().kSaveStateFile;
		stopTimer(timer);
		
		try { Thread.sleep(2 * kEpochLength); }
		catch (InterruptedException e) { }
		
		// Check the serialized state.
		assertEquals(true, f.exists());
		
		ObjectInputStream ois = null;
		try {
			ois = new ObjectInputStream(new FileInputStream(f));
			Vector<ReadOnlyVanishRefreshTask> tasks =
				(Vector<ReadOnlyVanishRefreshTask>)ois.readObject();
			assertEquals(1, tasks.size());

			ReadOnlyVanishRefreshTask task = tasks.get(0);
			assertEquals(meta, task.getMetadata());
		} catch (Exception e) {
			e.printStackTrace();
			fail("Failed to read saved state.");
		} finally {
			if (ois != null) {
				try { ois.close(); }
				catch (IOException e) { fail("Failed to close file."); }
			}
		}
		
		f.delete();
	}
}
