/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.apps;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.Map;

import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.impl.InputStreamWrapper;
import edu.washington.cs.vanish.internal.impl.OutputStreamWrapper;
import edu.washington.cs.vanish.internal.metadata.Metadata;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.internal.metadata.impl.BasicMetadataImpl;
import edu.washington.cs.vanish.util.Constants;
import edu.washington.cs.vanish.util.FileUtil;

/**
 * Mock class of {@link VanshClientInterface}. Keeps track of timeouts
 * in-memory.
 * 
 * @author alevy
 * 
 */
public class VanishMockClientImpl implements VanishClientInterface {

	public static final String ENCRYPTION_STRING = "This data is encrypted. You know because I just said so.";

	public static final long HOUR_IN_MILLIS = 3600000;

	protected Map<Long, Long> seedTtls = new HashMap<Long, Long>();

	public void decapsulate(InputStream enc_data_is,
			                OutputStream clear_data_os)
	throws VanishException {
		try {
			InputStreamWrapper encDataInputStream = new InputStreamWrapper(
					enc_data_is);
			Metadata metadata = (Metadata) encDataInputStream
					.getObjectInputStream().readObject();
			Long ttl = seedTtls.get(metadata.getLocationSeed());
			if (ttl == null
					|| System.currentTimeMillis() > metadata.getLocationSeed()
							+ ttl) {
				throw new VanishException("Could not decrypt message");
			}

			for (int i = 0; i < ENCRYPTION_STRING.length(); ++i) {
				encDataInputStream.read();
			}

			FileUtil.writeInputOutput(encDataInputStream, clear_data_os);

			encDataInputStream.close();
			clear_data_os.close();
		} catch (IOException e) {
			throw new VanishException(e.getMessage());
		} catch (ClassNotFoundException e) {
			throw new VanishException(e.getMessage());
		}
	}

	public void encapsulate(InputStream clear_data_is,
			                OutputStream enc_data_os,
			                VDOParams params)
	throws VanishException {
		Metadata metadata = new BasicMetadataImpl(params,
				                                  System.currentTimeMillis());

		seedTtls.put(metadata.getLocationSeed(),
				     params.getTimeout() * Constants.HOURS);
		try {
			OutputStreamWrapper encDataOutputStream = new OutputStreamWrapper(
					enc_data_os);
			encDataOutputStream.getObjectOutputStream().writeObject(metadata);
			encDataOutputStream.write(ENCRYPTION_STRING.getBytes());
			FileUtil.writeInputOutput(clear_data_is, encDataOutputStream);
			clear_data_is.close();
			encDataOutputStream.close();
		} catch (IOException e) {
			throw new VanishException(e.getMessage());
		}
	}

	public void refreshDeeply(InputStream enc_data_is,
			                  OutputStream refreshed_enc_data_os,
			                  VDOParams params)
	throws VanishException {
		try {
			InputStreamWrapper encDataInputStream = new InputStreamWrapper(
					enc_data_is);
			OutputStreamWrapper refreshedEncDataOutputStream = new OutputStreamWrapper(
					refreshed_enc_data_os);
			encDataInputStream.getObjectInputStream().readObject();
			Metadata newMetadata = new BasicMetadataImpl(params,
					System.currentTimeMillis());
			seedTtls.put(newMetadata.getLocationSeed(),
					     params.getTimeout() * Constants.HOURS);
			refreshedEncDataOutputStream.getObjectOutputStream().writeObject(
					newMetadata);
			FileUtil.writeInputOutput(encDataInputStream,
					refreshedEncDataOutputStream);
			encDataInputStream.close();
			refreshedEncDataOutputStream.close();
		} catch (IOException e) {
			throw new VanishException(e.getMessage());
		} catch (ClassNotFoundException e) {
			throw new VanishException(e.getMessage());
		}
	}

	public void refreshReadOnly(InputStream enc_data_is) throws VanishException {
		try {
			InputStreamWrapper encDataInputStream = new InputStreamWrapper(
					enc_data_is);
			Metadata metadata = (Metadata) encDataInputStream
					.getObjectInputStream().readObject();
			seedTtls.put(metadata.getLocationSeed(), 8 * HOUR_IN_MILLIS);
		} catch (IOException e) {
			throw new VanishException(e.getMessage());
		} catch (ClassNotFoundException e) {
			throw new VanishException(e.getMessage());
		}

	}

	public void init() throws VanishException {

	}

	public boolean isInitialized() {
		return true;
	}

	public void stop() throws VanishException { }

	public void setVanishServiceURL(URL url) throws UnknownHostException { }

	public void decapsulate(String enc_file_name, String clear_file_name)
			throws VanishException {
		throw new UnsupportedOperationException("VanishMockClientImple does not support local operations");
	}

	public void encapsulate(String clear_data_file_name,
			                String enc_data_file_name,
			                VDOParams params)
	throws VanishException {
		throw new UnsupportedOperationException(
				"VanishMockClientImple does not support local operations");
	}

	public void refreshDeeply(String enc_data_file_name,
			                  String refreshed_enc_data_file_name,
			                  VDOParams params)
	throws VanishException {
		throw new UnsupportedOperationException(
				"VanishMockClientImple does not support local operations");
	}

	public void refreshReadOnly(String enc_data_file_name)
			throws VanishException {
		throw new UnsupportedOperationException(
				"VanishMockClientImple does not support local operations");
	}
}
