/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.apps;

import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;

import org.apache.commons.configuration.ConfigurationException;
import org.apache.xmlrpc.XmlRpcException;
import org.apache.xmlrpc.client.XmlRpcClient;
import org.apache.xmlrpc.client.XmlRpcClientConfigImpl;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.logging.LogStat;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.service.LocalVanishServiceInterface;
import edu.washington.cs.vanish.service.RemoteVanishServiceInterface;
import edu.washington.cs.vanish.util.Constants;
import edu.washington.cs.vanish.util.FileUtil;
import edu.washington.cs.vanish.util.IOVanishException;

/**
 * 
 * @author roxana
 */
public class VanishClientImpl implements VanishClientInterface, Constants {
	// Configuration parameters:
	private final long kRpcTimeout; 
	private final int kNumRPCTries;
	
	private final VanishLogger LOG;
	
	// State:
	
	private URL sds_url;
	private Boolean is_server_local;

	private XmlRpcClient client;
	
	// Used for stats:
	
	private final Map<String, Object> client_map;

	/**
	 * Empty constructor. Any subclass must have one of these to be composable
	 * with the VanishClientCmdLine.
	 */
	public VanishClientImpl(VanishConfiguration CONF, VanishLogger log) {		
		kRpcTimeout = CONF.getTimeInMillis(
				WebappDefaults.CONF_VANISH_APPS_RPCTIMEOUT_S,
				WebappDefaults.CONF_VANISH_APPS_RPCTIMEOUT_S_DEFAULT,
				VanishConfiguration.TimeUnit.SECONDS);
		kNumRPCTries = CONF.getInt(
				WebappDefaults.CONF_VANISH_APPS_NUMRPCTRIES, 
				WebappDefaults.CONF_VANISH_APPS_NUMRPCTRIES_DEFAULT);
		
		this.LOG = log;
		
		client_map = new HashMap<String, Object>();
		loadClientMap(CONF);
	}
	
	private void loadClientMap(VanishConfiguration CONF) {
		long installation_id =
			CONF.getLong(WebappDefaults.CONF_VANISH_APPS_INSTALLATION_ID, 0);
		if (installation_id == 0) {  // First install; generate id randomly
			installation_id = new Random().nextLong();
			CONF.setProperty(WebappDefaults.CONF_VANISH_APPS_INSTALLATION_ID,
					         installation_id);
			try { CONF.save(); }
			catch (ConfigurationException e) { e.printStackTrace(); }
		}
		
		client_map.put("vid", installation_id);
		client_map.put("ref", "shell");
	}

	// Initializable interface:

	// @Override
	public void init() throws InvalidArgumentVanishException {
		assert (!isInitialized());

		// Perform some special configs.
		System.setProperty("sun.net.client.defaultConnectTimeout",
				"" + kRpcTimeout);
		System.setProperty("sun.net.client.defaultReadTimeout",
				"" + kRpcTimeout);

		try {
			initConnectionToVanishServer();
		} catch (UnknownHostException e) {
			throw new InvalidArgumentVanishException(
					"Failed to init vanish server connection", e);
		}
	}

	// @Override
	public boolean isInitialized() { return client != null; }

	// @Override
	public void stop() { }

	private void initConnectionToVanishServer() throws UnknownHostException {
		XmlRpcClientConfigImpl config = new XmlRpcClientConfigImpl();
		config.setEnabledForExtensions(true);
		config.setServerURL(sds_url);
		int timeout = getTimeout();
		if (timeout > 0) {
			config.setReplyTimeout(timeout);
			config.setConnectionTimeout(timeout);
		}
		client = new XmlRpcClient();
		client.setConfig(config);

		setIsServerLocal();
	}

	private void setIsServerLocal() throws UnknownHostException {
		is_server_local = false;
		InetAddress sds_service_addr = InetAddress.getByName(
				sds_url.getHost());
		if (InetAddress.getLocalHost().equals(sds_service_addr) ||
			sds_service_addr.isLoopbackAddress()) {
			is_server_local = true;
		}
	}

	private boolean isServerLocal() { return is_server_local; }

	protected int getTimeout() { return -1; }

	// VanishClientInterface:

	public void setVanishServiceURL(URL url) throws UnknownHostException {
		this.sds_url = url;
		initConnectionToVanishServer(); // re-initialize.
	}

	public void encapsulate(InputStream clear_data_is,
			                OutputStream enc_data_os,
			                VDOParams params)
	throws VanishException {
		final LogStat logstat = LOG.logStat("VanishClientImpl.encapsulate");
		try {
			byte[] result = executeWithReturn(
			  RemoteVanishServiceInterface.kVanishServiceName + ".encapsulate",
			  new Object[] { client_map, readStream(clear_data_is),
					         params.toMap() });
			enc_data_os.write(result);
		} catch (IOException e) {
			throw new IOVanishException("Failed to write result to OS");
		}
		logstat.end();
	}

	public void encapsulate(String clear_data_file_name,
			                String enc_data_file_name,
			                VDOParams params)
	throws VanishException {
		final LogStat logstat = LOG.logStat("VanishClientImpl.encapsulate");
		if (!isServerLocal()) {
			try {
				encapsulate(new FileInputStream(clear_data_file_name),
						    new FileOutputStream(enc_data_file_name),
						    params);
			} catch (IOException e) {
				throw new IOVanishException("Failed to open I/O streams", e);
			}
		} else {
			executeWithoutReturn(
			  LocalVanishServiceInterface.kVanishServiceName + ".encapsulate",
			  new Object[] { client_map,
					         clear_data_file_name, enc_data_file_name,
					         params.toMap() });
		}
		logstat.end();
	}

	public void decapsulate(InputStream enc_data_is, 
			                OutputStream clear_data_os)
	throws VanishException {
		final LogStat logstat = LOG.logStat("VanishClientImpl.decapsulate");
		try {
			byte[] result = executeWithReturn(
			  RemoteVanishServiceInterface.kVanishServiceName + ".decapsulate",
			  new Object[] { client_map, readStream(enc_data_is) });
			clear_data_os.write(result);
		} catch (IOException e) {
			throw new IOVanishException("Failed to write result", e);
		}
		logstat.end();
	}

	public void decapsulate(String enc_file_name, String clear_file_name)
	throws VanishException {
		final LogStat logstat = LOG.logStat("VanishClientImpl.decapsulate");
		if (! isServerLocal()) {
			try {
				decapsulate(new FileInputStream(enc_file_name),
						    new FileOutputStream(clear_file_name));
			} catch (IOException e) {
				throw new IOVanishException("Failed to open I/O streams", e);
			}
		} else {
			executeWithoutReturn(
			  LocalVanishServiceInterface.kVanishServiceName + ".decapsulate",
			  new Object[] { client_map, enc_file_name, clear_file_name });
		}
		logstat.end();
	}

	public void refreshDeeply(InputStream enc_data_is,
                              OutputStream refreshed_enc_data_os,
                              VDOParams params)
	throws VanishException {
		final LogStat logstat = LOG.logStat("VanishClientImpl.refreshDeeply");
		try {
			byte[] result = executeWithReturn(
			  RemoteVanishServiceInterface.kVanishServiceName + ".refreshDeeply",
			  new Object[] { client_map, readStream(enc_data_is),
					         params.toMap() });
			refreshed_enc_data_os.write(result);
		} catch (IOException e) {
			throw new IOVanishException("Failed to write output", e);
		}
		logstat.end();
	}

	public void refreshDeeply(String enc_data_file_name,
			                  String refreshed_enc_data_file_name,
			                  VDOParams params)
	throws VanishException {
		final LogStat logstat = LOG.logStat("VanishClientImpl.refreshDeeply");
		if (!isServerLocal()) {
			try {
				refreshDeeply(new FileInputStream(enc_data_file_name),
						new FileOutputStream(refreshed_enc_data_file_name),
						params);
			} catch (IOException e) {
				throw new VanishException(e.getMessage());
			}
		} else {
			executeWithoutReturn(
			  LocalVanishServiceInterface.kVanishServiceName + ".refreshDeeply",
			  new Object[] { client_map,
					         enc_data_file_name, refreshed_enc_data_file_name,
				             params.toMap() });
		}
		logstat.end();
	}

	public void refreshReadOnly(InputStream enc_data_is)
	throws VanishException {
		final LogStat logstat = LOG.logStat(
				"VanishClientImpl.refreshReadOnly");
		executeWithoutReturn(
		  RemoteVanishServiceInterface.kVanishServiceName + ".refreshReadOnly",
		  new Object[] { client_map, readStream(enc_data_is) });
		logstat.end();
	}

	public void refreshReadOnly(String enc_data_file_name)
	throws VanishException {
		final LogStat logstat = LOG.logStat(
				"VanishClientImpl.refreshReadOnly");
		if (!isServerLocal()) {
			try {
				refreshReadOnly(new FileInputStream(enc_data_file_name));
			} catch (IOException e) {
				throw new IOVanishException("Failed to open input stream", e);
			}
		} else {
			executeWithoutReturn(
			  LocalVanishServiceInterface.kVanishServiceName + ".refreshReadOnly",
			  new Object[] { client_map, enc_data_file_name });
		}
		logstat.end();
	}

	// Helper functions:

	// TODO(roxana): Some redundant code here -- eliminate.
	
	private byte[] executeWithReturn(String operation, Object[] params)
	throws VanishException {
		XmlRpcException last_exception = null;
		for (int i = 0; i < kNumRPCTries; ++i) {
			try {
				return (byte[])client.execute(operation, params);
			} catch (XmlRpcException e) {
				last_exception = e;
				continue;
			}
		}
		throw new VanishException("Failed after retries", last_exception);
	}

	private void executeWithoutReturn(String operation, Object[] params)
	throws VanishException {
		XmlRpcException last_exception = null;
		for (int i = 0; i < kNumRPCTries; ++i) {
			try {
				client.execute(operation, params);
				return; // done.
			} catch (XmlRpcException e) {
				last_exception = e;
				continue;
			}
		}
		throw new VanishException("Failed after retries", last_exception);
	}

	private byte[] readStream(InputStream in) throws IOVanishException {
		ByteArrayOutputStream output = new ByteArrayOutputStream();
		FileUtil.writeInputOutput(in, output);
		return output.toByteArray();
	}
}
