/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.apps;

import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.net.UnknownHostException;

import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.util.Initializable;
/**
 * 
 * @author roxana
 *
 */
public interface VanishClientInterface extends Initializable {
	// Initialization:
	public void setVanishServiceURL(URL url) throws UnknownHostException;

	// Encapsulate:
	/**
	 * Encapsulates a cleartext message to a VDO. Implementing classes will
	 * output self-containing VDOs - that is - all information needed to
	 * decapsulate the message is output to {@link enc_data_os}.
	 * 
	 * @param clear_data_is
	 *            an {@link InputStream} from which the cleartext data should be
	 *            read.
	 * @param enc_data_os
	 *            an {@link OutputStream} to which the
	 * @param num_shares
	 *            the total number of shares to break the key into.
	 * @param threshold
	 *            the number of shares needed to restore the key.
	 * @param timeout_h
	 *            the timeout of the object in hours.
	 * @param flag
	 *            a flag that supports future extensions of this interface;
	 *            use (byte)0 for default behavior.
	 */
	public void encapsulate(InputStream clear_data_is,
			                OutputStream enc_data_os,
			                VDOParams params)
	throws VanishException;

	/**
	 * Encapsulates a cleartext message to a VDO. Implementing classes will
	 * output self-containing VDOs - that is - all information needed to
	 * decapsulate the message is output to the file <i>enc_data_file_name</i>.
	 * 
	 * @param clear_data_file_name
	 *            the URI of the file from which the cleartext data should be
	 *            read.
	 * @param enc_data_os
	 *            the URI of the file to which the encrypted data should be
	 *            written.
	 * @param params
	 *            the operation's parameters.
	 */
	public void encapsulate(String clear_data_file_name,
			                String enc_data_file_name,
			                VDOParams params)
	throws VanishException;

	// Decapsulate:
	/**
	 * Decapsulates a VDO.
	 * 
	 * @param enc_data_is
	 *            an {@link InputStream} from which to read the VDO.
	 * @param clear_data_os
	 *            an {@link OutputStream} to which the cleartext is written.
	 */
	public void decapsulate(InputStream enc_data_is,
			                OutputStream clear_data_os)
	throws VanishException;
	/**
	 * Decapsulates a VDO.
	 * 
	 * @param enc_data_file_name
	 *            the URI of the file from which to read the VDO.
	 * @param clear_data_file_name
	 *            the URI of the file to which the cleartext is written.
	 */
	public void decapsulate(String enc_file_name, String clear_file_name)
	throws VanishException;

	// Refresh:
	/**
	 * Refreshes a VDO, extending its life.
	 * 
	 * @param enc_data_is
	 *            an {@link InputStream} from which to read the VDO.
	 */
	public void refreshReadOnly(final InputStream enc_data_is)
	throws VanishException;

	/**
	 * Refreshes a VDO, extending its life.
	 * 
	 * @param enc_data_file_name
	 *            the URI of the file from which to read the VDO.
	 */
	public void refreshReadOnly(String enc_data_file_name)
	throws VanishException;

	/**
	 * Re-encapsulates the VDO with a new key.
	 * 
	 * @param enc_data_is
	 *            an {@link InputStream} from which the old VDO is read.
	 * @param refreshed_enc_data_os
	 *            an {@link OutputStream} to which the new VDO will be output.
	 * @param params
	 *            the operation's parameters.
	 */
	public void refreshDeeply(InputStream enc_data_is,
			                  OutputStream refreshed_enc_data_os,
			                  VDOParams params)
	throws VanishException;

	/**
	 * Re-encapsulates the VDO with a new key.
	 * 
	 * @param enc_data_file_name
	 *            the URI of the file from which the old VDO is read.
	 * @param refreshed_enc_data_file_name
	 *            the URI of the file to which the new VDO will be output.
	 */
	public void refreshDeeply(String enc_data_file_name,
			                  String refreshed_enc_data_file_name,
			                  VDOParams params)
	throws VanishException;
}
