/*
 * This code is subject to an academic license contained in the LICENSE-vanish.txt
 * file that can be found in the root directory of Vanish.
 */

package edu.washington.cs.vanish.apps;

import java.lang.reflect.Constructor;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.UnknownHostException;

import org.apache.commons.configuration.Configuration;

import edu.washington.cs.vanish.conf.VanishConfiguration;
import edu.washington.cs.vanish.internal.InvalidArgumentVanishException;
import edu.washington.cs.vanish.internal.VanishException;
import edu.washington.cs.vanish.internal.metadata.VDOParams;
import edu.washington.cs.vanish.logging.VanishLogger;
import edu.washington.cs.vanish.util.CmdLineParser;
import edu.washington.cs.vanish.util.Initializable;
import edu.washington.cs.vanish.util.VanishUtils;

/**
 * TODO(roxana): Revise cmdline parsing. Forward most cmd-line args to the
 * client implementation instead of parsing them here.
 * 
 * @author roxana
 */
public class VanishClientRunner implements Initializable {
	private static final VanishLogger LOG = VanishLogger.getLogger("client");

	// Cmdline params for the vanish operation:
	private VanishOperationArgs vanish_op_args;

	// Underlying client implementation:
	private final VanishClientInterface base_client;

	public VanishClientRunner(VanishConfiguration conf,
			                  VanishOperationArgs vanish_op_args)
	throws UnknownHostException, InvalidArgumentVanishException,
	       VanishException, MalformedURLException {
		this.vanish_op_args = vanish_op_args;
		
		String client_class_name = conf.getString(
				WebappDefaults.CONF_VANISH_APPS_CLIENTCLASS,
				WebappDefaults.CONF_VANISH_APPS_CLIENTCLASS_DEFAULT);
		try {
			Class<? extends VanishClientInterface> client_class =
				Class.forName(client_class_name).asSubclass(
						VanishClientInterface.class);
			Constructor<? extends VanishClientInterface> constructor =
				client_class.getConstructor(VanishConfiguration.class,
						                    VanishLogger.class);
			base_client = constructor.newInstance(conf, LOG);
		} catch (Exception e) {
			throw new InvalidArgumentVanishException(
				"Failed to instantiate client " + client_class_name, e);
		}
		
		URL vanish_server_url = new URL(
				conf.getString(WebappDefaults.CONF_VANISH_APPS_VANISHURL,
						       WebappDefaults.CONF_VANISH_APPS_VANISHURL_DEFAULT));

		base_client.setVanishServiceURL(vanish_server_url);
	}
	
	protected VDOParams instantiateVDOParams()
	throws InvalidArgumentVanishException { 
		return new VDOParams();
	}

	// Initializable interface:

	// @Override
	public void init() throws VanishException { base_client.init(); }

	// @Override
	public boolean isInitialized() { return base_client.isInitialized(); }

	// @Override
	public void stop() throws VanishException { base_client.stop(); }

	// Execution-related functions:
	
	// @Override
	public void execute() throws VanishException {
		String[] ops = vanish_op_args.operation.split("[+]");
		for (String operation : ops) {
			if (operation.equals("encapsulate")) {
				base_client.encapsulate(vanish_op_args.input,
						vanish_op_args.output, vanish_op_args.params);

			} else if (operation.equals("decapsulate")) {
				base_client.decapsulate(vanish_op_args.input,
						vanish_op_args.output);
				
			} else if (operation.equals("refreshReadOnly")) {
				base_client.refreshReadOnly(vanish_op_args.input);
				
			} else if (operation.equals("refreshDeeply")) {
				base_client.refreshDeeply(vanish_op_args.input,
						vanish_op_args.output, vanish_op_args.params);
			} else {
				System.err.println("Invalid Vanish operation " + operation);
				throw new InvalidArgumentVanishException("Invalid operation");
			}
		}
	}

	protected void doWork() throws VanishException {
		init();
		execute();
		stop();
	}

	@SuppressWarnings("deprecation")
	public static void main(String[] args) {
		VanishConfiguration conf = VanishUtils.getConfig(
				WebappDefaults.WEBAPPS_CONFIG_FILE);
		ClientCmdLineParser parser = null;
		try {
			parser = new ClientCmdLineParser(conf);
		} catch (InvalidArgumentVanishException e1) {
			System.exit(-1);
		}
		
		if (! parser.parse(args)) System.exit(-1);

		try {
			VanishClientRunner client_runner =
				new VanishClientRunner(conf, parser.vanish_op_args);
			client_runner.doWork();
			System.exit(0);
		} catch (Exception e) {
			e.printStackTrace();
		}
		System.exit(-1);  // very important to exit with != 0; plugins
                          // depend on it.
	}
}

class VanishOperationArgs {
	public String input     = null;
	public String output    = null;
	public String operation = null;
	
	public VDOParams params;

	public int id = 0;  // a client ID
	
	public VanishOperationArgs(VDOParams params_instance) {
		this.params = params_instance;
	} 
	
	public VanishOperationArgs(VanishOperationArgs other) {
		if (other == null) return;
		this.input       = other.input;
		this.output      = other.output;
		this.operation   = other.operation;
		
		this.params      = other.params;

		this.id          = other.id;
	}
	
	public void validate() throws InvalidArgumentVanishException {
		if (operation == null) {
			throw new InvalidArgumentVanishException("Invalid null operation");
		}
		params.validate();
	}
}

class ClientCmdLineParser extends CmdLineParser {
	protected VanishOperationArgs vanish_op_args;
	
	public ClientCmdLineParser(Configuration conf)
	throws InvalidArgumentVanishException {
		super(conf);
		vanish_op_args = new VanishOperationArgs(new VDOParams());
	}

	@Override
	protected int parseOption(String arg, String[] args, int i)
	throws InvalidArgumentVanishException {
		if (arg.equals("-f")) {
			if (i < args.length) {
				vanish_op_args.input = args[i++];
				if (vanish_op_args.input == null ||
					vanish_op_args.input.equals("")) {
					throw new InvalidArgumentVanishException(
							"-f requires a filename");
				}
			} else {
				throw new InvalidArgumentVanishException(
						"-f requires a filename");
			}
		} else if (arg.equals("-o")) {
			if (i < args.length) {
				vanish_op_args.output = args[i++];
				if (vanish_op_args.output == null ||
					vanish_op_args.output.equals("")) {
					throw new InvalidArgumentVanishException(
							"-o requires a filename");
				}
			} else {
				throw new InvalidArgumentVanishException(
						"-o requires a filename");
			}
		} else if (arg.equals("-n")) {
			if (i < args.length) {
				try {
					vanish_op_args.params.setThreshold(Integer
							.parseInt(args[i++]));
				} catch (Exception e) {
					throw new InvalidArgumentVanishException(
							"-n requires a number");
				}
			} else {
				throw new InvalidArgumentVanishException(
						"-n requires a number");
			}
		} else if (arg.equals("-t")) {
			if (i < args.length) {
				try {
					vanish_op_args.params.setNumShares(Integer
							.parseInt(args[i++]));
				} catch (Exception e) {
					throw new InvalidArgumentVanishException(
						"-t requires a number");
				}
			} else {
				throw new InvalidArgumentVanishException(
						"-t requires a number");
			}
		} else if (arg.equals("-u")) {
			if (i < args.length) {
				String url = args[i++];
				conf.setProperty(WebappDefaults.CONF_VANISH_APPS_VANISHURL, url);
			} else {
				throw new InvalidArgumentVanishException(
						"-u requires a valid URL");
			}
		} else if (arg.equals("-x")) {
			if (i < args.length) {
				try {
					vanish_op_args.params.setTimeoutInHours(Integer
							.parseInt(args[i++]));
				} catch (Exception e) {
					throw new InvalidArgumentVanishException(
							"-x requires a number (hours)");
				}
			} else {
				throw new InvalidArgumentVanishException(
						"-x requires a number (hours)");
			}
		} else if (arg.equals("-c")) {
			if (i < args.length) {
				String base_client_class = args[i++];
				conf.setProperty(WebappDefaults.CONF_VANISH_APPS_CLIENTCLASS,
						         base_client_class);
			} else {
				throw new InvalidArgumentVanishException(
						"-c requires a class name");
			}
		} else {
			// TODO(roxana): VanishClient implementations should also
			// have options. This is the chance to ask them to parse their
			// options.
			throw new InvalidArgumentVanishException(
					"Invalid option: " + args[i]);
		}

		return i;
	}

	@Override
	public String usage() {
		return "USAGE: java " + this.getClass().getName() + " " +
		"-f <infile> -o <outfile>\n" +
		"\t[-t <total_keys>]\n" +
		"\t[-n <needed_keys>]\n" +
		"\t[-u <vanish server URL> " + "(default "+ 
		WebappDefaults.CONF_VANISH_APPS_VANISHURL_DEFAULT + ")]\n" +
		"\t[-x <timeout in hours>]\n" +
		"\t[-c <vanish client class> " +
		"(default edu.washington.cs.vanish.apps.VanishClientImpl)]\n" +
		"encapsulate | decapsulate | refreshReadOnly | refreshDeeply\n";
	}

	@Override
	public void validate() throws InvalidArgumentVanishException {
		if (vanish_op_args.input == null) {
			throw new InvalidArgumentVanishException(
					"invalid null input, use -f <infile>");
		}
		vanish_op_args.validate();
	}

	@Override
	protected void parseCommand(String command)
	throws InvalidArgumentVanishException {
		this.vanish_op_args.operation = command;
	}
}
